/*
 * common.c:
 *      common.c contains functions that are common to all
 *      parts of XFolder. As opposed to helpers.c, these
 *      functions work with XFolder only, while the helper
 *      functions will work with any PM application.
 *
 *      These funcs are forward-declared in common.h, which
 *      is #include'd in all XFolder source files.
 *
 *      Copyright (C) 1997-98 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of the XFolder main distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

// #define INCL_DOSFILEMGR
// #define INCL_DOSMODULEMGR
// #define INCL_DOSMEMMGR
// #define INCL_DOSPROCESS
// #define INCL_DOSSEMAPHORES
// #define INCL_DOSEXCEPTIONS
// #define INCL_DOSERRORS
// #define INCL_DOSMISC
// #define INCL_GPILOGCOLORTABLE
// #define INCL_GPIPRIMITIVES
#define INCL_BASE
#define INCL_SPL
#define INCL_SPLDOSPRINT
#define INCL_SPLERRORS
#define INCL_WIN
#define INCL_WINWORKPLACE
#define INCL_GPILOGCOLORTABLE
#include <os2.h>

#include "xfldr.ih"
#include "xfobj.h"

#include "common.h"

#include  <wpdesk.h>        // WPDesktop
#include  <wpclsmgr.h>

HMODULE         hmodNLS = NULLHANDLE;
PVOID           pNLSStringsGlobal = NULL;
PGLOBALSETTINGS pGlobalSettings = NULL;

CHAR            szStatusBarFont[100];
CHAR            szSBTextNoneSel[CCHMAXMNEMONICS],
                // szSBText1Sel[CCHMAXMNEMONICS],
                szSBTextMultiSel[CCHMAXMNEMONICS];
ULONG           ulStatusBarHeight;

CHAR            szHelpLibrary[CCHMAXPATH] = "";
CHAR            szMessageFile[CCHMAXPATH] = "";
CHAR            szLanguageCode[10] = "";

ULONG           ulCurHelpPanel = 0;      // holds help panel for dialog

HFILE   hfDebug = NULLHANDLE;

/* The following functions are used from several XFolder instance or
   class methods and have therefore been withdrawn from the actual
   method code. To mark them as "common" routines, their function
   names are introduced with "cmn". */

/*
 * cmnQueryXFolderBasePath:
 *          this routine retrieves the XFolder path from OS2.INI and
 *          copies it to pszPath (null-terminated); returns number of
 *          copied chars, 0 upon errors
 */

ULONG cmnQueryXFolderBasePath(PSZ pszPath, ULONG ulBufsize)
{
    ULONG ulCopied = PrfQueryProfileString(HINI_USERPROFILE,
               INIAPP, INIKEY_XFOLDERPATH,
               NULL,                    // no default: returns 0 if not found
               (PVOID)pszPath,
               ulBufsize);
    if (ulCopied) {
        pszPath[ulCopied] = '\0';
        // strcpy(pszPath+ulCopied-1, "\\bin");
    }
    #ifdef DEBUG_LANGCODES
        _Pmpf(( "cmnQueryXFolderBasePath: %s", pszPath ));
    #endif
    return (ulCopied);
}

/*
 * cmnQueryLanguageCode:
 *           returns PSZ to three-digit language code (e.g. "001")
 */

PSZ cmnQueryLanguageCode(VOID)
{
    if (szLanguageCode[0] == '\0')
        PrfQueryProfileString(HINI_USERPROFILE,
               INIAPP, INIKEY_LANGUAGECODE,
               DEFAULT_LANGUAGECODE,
               (PVOID)szLanguageCode,
               sizeof(szLanguageCode));

    szLanguageCode[3] = '\0';
    #ifdef DEBUG_LANGCODES
        _Pmpf(( "cmnQueryLanguageCode: %s", szLanguageCode ));
    #endif
    return (szLanguageCode);
}

/*
 * cmnSetLanguageCode:
 *           changes XFolder's language to three-digit language code in
 *           pszLanguage (e.g. "001")
 */

BOOL cmnSetLanguageCode(PSZ pszLanguage)
{
    strcpy(szLanguageCode, pszLanguage);

    return (PrfWriteProfileString(HINI_USERPROFILE, INIAPP, INIKEY_LANGUAGECODE,
        szLanguageCode));
}

/*
 * cmnQueryHelpLibrary:
 *           returns PSZ to full help library path in XFolder directory
 */

PSZ cmnQueryHelpLibrary(VOID)
{
    ULONG ulCopied = PrfQueryProfileString(HINI_USERPROFILE,
               INIAPP, INIKEY_XFOLDERPATH,
               NULL,                    // no default: returns 0 if not found
               (PVOID)szHelpLibrary,
               sizeof(szHelpLibrary));
    if (ulCopied) {
        // INI entry found: append "\" if neccessary
        sprintf(szHelpLibrary+ulCopied-1,
                "\\help\\xfldr%s.hlp",
                cmnQueryLanguageCode());
        #ifdef DEBUG_LANGCODES
            _Pmpf(( "cmnQueryHelpLibrary: %s", szHelpLibrary ));
        #endif
        return (szHelpLibrary);
    } else
        return (NULL);
}

/*
 * cmnDisplayFocusHelp:
 *      this is used from all kinds of settings dlg
 *      procs to display a help panel according to
 *      the dlg item which currently has the focus;
 *      if no matching help panel is found,
 *      ulPanelIfNotFound is displayed instead
 */

BOOL cmnDisplayFocusHelp(ULONG ulPanelIfNotFound)
{
    BOOL brc = TRUE;
    PSZ pszHelpLibrary = cmnQueryHelpLibrary();

    HWND hwndFocus = WinQueryFocus(HWND_DESKTOP);
    if (hwndFocus) {
        WPObject *pHelpSomSelf = _wpclsQueryActiveDesktop(_WPDesktop);
                            // (XFldSystem*)WinQueryWindowULong(hwndDlg, QWL_USER);
        if (pHelpSomSelf) {
            USHORT idItem = WinQueryWindowUShort(hwndFocus, QWS_ID);
            #ifdef DEBUG_LANGCODES
                _Pmpf(( "cmnDisplayFocusHelp: lib %s, idItem: %d ulPanel: %d", pszHelpLibrary, idItem, idItem+ID_XSH_FIRST-ID_XSDI_FIRST ));
            #endif
            if (!_wpDisplayHelp(pHelpSomSelf, (idItem-ID_XSDI_FIRST+ID_XSH_FIRST),
                pszHelpLibrary))
            {
                #ifdef DEBUG_LANGCODES
                    _Pmpf(( "  not found, displaying superpanel %d", ulPanelIfNotFound ));
                #endif
                if (!_wpDisplayHelp(pHelpSomSelf, ulPanelIfNotFound,
                    pszHelpLibrary))
                {
                    cmnMessageBoxMsg(HWND_DESKTOP, 104, 134, MB_OK);
                    brc = FALSE;
                }
            }
        }
    }
    return (brc);
}

/*
 * cmnQueryMessageFile:
 *           returns PSZ to full message file path in XFolder directory
 */

PSZ cmnQueryMessageFile(VOID)
{
    ULONG ulCopied = PrfQueryProfileString(HINI_USERPROFILE,
               INIAPP, INIKEY_XFOLDERPATH,
               NULL,                    // no default: returns 0 if not found
               (PVOID)szMessageFile,
               sizeof(szMessageFile));
    if (ulCopied) {
        // INI entry found: append "\" if neccessary
        sprintf(szMessageFile+ulCopied-1,
                "\\help\\xfldr%s.msg",
                cmnQueryLanguageCode());
        #ifdef DEBUG_LANGCODES
            _Pmpf(( "cmnQueryMessageFile: %s", szMessageFile ));
        #endif
        return (szMessageFile);
    } else
        return (NULL);
}

/*
 * cmnLoadString:
 *      pretty similar to WinLoadString, but allocates
 *      neccessary memory as well. *ppsz is a pointer
 *      to a PSZ; if this PSZ is != NULL, whatever it
 *      points to will be free()d, so this better be a
 *      valid pointer
 */

void cmnLoadString(HAB habDesktop, HMODULE hmodResource, ULONG ulID, PSZ *ppsz)
{
    CHAR    szBuf[200];
    if (*ppsz)
        free(*ppsz);
    if (WinLoadString(habDesktop, hmodResource, ulID , sizeof(szBuf), szBuf))
        *ppsz = strcpy(malloc(strlen(szBuf)+1), szBuf);
    else
        *ppsz = "Error: String resource not found";
}

/*
 * cmnQueryNLSModuleHandle:
 *           returns the module handle of the language-dependent XFolder
 *           national language support DLL.
 *           On the first call, the DLL is loaded into memory and all
 *           XFolder strings are loaded from the DLL; if EnforceReload
 *           is TRUE, we will enforce loading a new resource DLL and
 *           check its version too (XFolder Internal Settings page)
 */

HMODULE cmnQueryNLSModuleHandle(BOOL fEnforceReload)
{
    CHAR    szResourceModuleName[CCHMAXPATH];
    ULONG   ulCopied;
    HMODULE hmodOldResource = hmodNLS;

    // load resource DLL if it's not loaded yet or a reload is enforced
    if ((hmodNLS == NULLHANDLE) || (fEnforceReload))
    {
        PNLSSTRINGS pNLSStrings = cmnQueryNLSStrings();

        // get the XFolder path first
        ulCopied = cmnQueryXFolderBasePath(szResourceModuleName,
                            sizeof(szResourceModuleName));
        if (ulCopied)
        {
            // now compose module name from language code
            strcat(szResourceModuleName, "\\bin\\xfldr");
            strcat(szResourceModuleName, cmnQueryLanguageCode());
            strcat(szResourceModuleName, ".dll");

            // try to load the module
            if (DosLoadModule(NULL,
                               0,
                               szResourceModuleName,
                               &hmodNLS))
            {
                DebugBox("XFolder: Couldn't Find Resource DLL",
                    szResourceModuleName);
            } else { // module loaded alright!
                HAB habDesktop = WinQueryAnchorBlock(HWND_DESKTOP);

                if (habDesktop)
                {
                    if (fEnforceReload) {
                        // if fEnforceReload == TRUE, we will load a test string from
                        // the module to see if it has at least the same version level as
                        // the main XFolder DLL; if not, we show a warning */
                        CHAR   szTest[30];
                        LONG   lLength;
                        lLength = WinLoadString(habDesktop, hmodNLS, ID_XSSI_XFOLDERVERSION,
                                sizeof(szTest), szTest);
                        if (    (lLength == 0)
                            ||  (strcmp(szTest, XFOLDER_VERSION) < 0)
                           )
                        {
                            cmnSetHelpPanel(-1);
                            if (lLength == 0) {
                                DebugBox("XFolder", "The requested file is not an XFolder National Language Support DLL.");
                            }
                            else if (WinDlgBox(HWND_DESKTOP,       // parent is desktop
                                         HWND_DESKTOP,             // owner is desktop
                                         (PFNWP)fnwpDlgGeneric,    // dialog procedure, defd. at bottom
                                         hmodNLS,                  // load from resource file
                                         ID_XFD_WRONGVERSION,      // dialog resource id
                                         (PVOID)NULL)              // no dialog parameters
                               == DID_CANCEL)
                            {
                                DebugBox("XFolder", "The new National Language Support DLL was not loaded.");
                                DosFreeModule(hmodNLS);
                                hmodNLS = hmodOldResource;
                                return NULLHANDLE;
                            }
                        }
                    }

                    // now let's load strings
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_NOTDEFINED, &(pNLSStrings->pszNotDefined));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_PRODUCTINFO, &(pNLSStrings->pszProductInfo));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_REFRESHNOW, &(pNLSStrings->pszRefreshNow));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SNAPTOGRID, &(pNLSStrings->pszSnapToGrid));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_FLDRCONTENT, &(pNLSStrings->pszFldrContent));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_COPYFILENAME, &(pNLSStrings->pszCopyFilename));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_BORED       , &(pNLSStrings->pszBored));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_FLDREMPTY   , &(pNLSStrings->pszFldrEmpty));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SELECTSOME  , &(pNLSStrings->pszSelectSome));

                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_QUICKSTATUS    , &(pNLSStrings->pszQuickStatus));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_NAME     , &(pNLSStrings->pszSortByName));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_TYPE     , &(pNLSStrings->pszSortByType));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_CLASS    , &(pNLSStrings->pszSortByClass));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_REALNAME , &(pNLSStrings->pszSortByRealName));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_SIZE     , &(pNLSStrings->pszSortBySize));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_WRITEDATE, &(pNLSStrings->pszSortByWriteDate));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_ACCESSDATE, &(pNLSStrings->pszSortByAccessDate));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_CREATIONDATE, &(pNLSStrings->pszSortByCreationDate));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_EXT       , &(pNLSStrings->pszSortByExt));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_FOLDERSFIRST, &(pNLSStrings->pszSortFoldersFirst));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_CTRL       , &(pNLSStrings->pszCtrl));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_Alt        , &(pNLSStrings->pszAlt));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_SHIFT      , &(pNLSStrings->pszShift));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_BACKSPACE  , &(pNLSStrings->pszBackspace));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_TAB        , &(pNLSStrings->pszTab));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_BACKTABTAB , &(pNLSStrings->pszBacktab));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_ENTER      , &(pNLSStrings->pszEnter));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_ESC        , &(pNLSStrings->pszEsc));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_SPACE      , &(pNLSStrings->pszSpace));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_PAGEUP     , &(pNLSStrings->pszPageup));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_PAGEDOWN   , &(pNLSStrings->pszPagedown));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_END        , &(pNLSStrings->pszEnd));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_HOME       , &(pNLSStrings->pszHome));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_LEFT       , &(pNLSStrings->pszLeft));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_UP         , &(pNLSStrings->pszUp));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_RIGHT      , &(pNLSStrings->pszRight));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_DOWN       , &(pNLSStrings->pszDown));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_PRINTSCRN  , &(pNLSStrings->pszPrintscrn));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_INSERT     , &(pNLSStrings->pszInsert));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_DELETE     , &(pNLSStrings->pszDelete));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_SCRLLOCK   , &(pNLSStrings->pszScrlLock));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_NUMLOCK    , &(pNLSStrings->pszNumLock));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_WINLEFT    , &(pNLSStrings->pszWinLeft));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_WINRIGHT   , &(pNLSStrings->pszWinRight));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_KEY_WINMENU    , &(pNLSStrings->pszWinMenu));

                    cmnLoadString(habDesktop, hmodNLS, ID_SDSI_FLUSHING       , &(pNLSStrings->pszSDFlushing));
                    cmnLoadString(habDesktop, hmodNLS, ID_SDSI_CAD            , &(pNLSStrings->pszSDCAD));
                    cmnLoadString(habDesktop, hmodNLS, ID_SDSI_REBOOTING      , &(pNLSStrings->pszSDRebooting));
                    cmnLoadString(habDesktop, hmodNLS, ID_SDSI_CLOSING        , &(pNLSStrings->pszSDClosing));
                    cmnLoadString(habDesktop, hmodNLS, ID_SDSI_SHUTDOWN       , &(pNLSStrings->pszShutdown));
                    cmnLoadString(habDesktop, hmodNLS, ID_SDSI_RESTARTWPS     , &(pNLSStrings->pszRestartWPS));
                    cmnLoadString(habDesktop, hmodNLS, ID_SDSI_RESTARTINGWPS  , &(pNLSStrings->pszSDRestartingWPS));
                    cmnLoadString(habDesktop, hmodNLS, ID_SDSI_SAVINGDESKTOP  , &(pNLSStrings->pszSDSavingDesktop));
                    cmnLoadString(habDesktop, hmodNLS, ID_SDSI_SAVINGPROFILES , &(pNLSStrings->pszSDSavingProfiles));

                    cmnLoadString(habDesktop, hmodNLS, ID_SDSI_STARTING       , &(pNLSStrings->pszStarting));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_POPULATING     , &(pNLSStrings->pszPopulating));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_1GENERIC       , &(pNLSStrings->psz1Generic));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_2REMOVEITEMS   , &(pNLSStrings->psz2RemoveItems));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_25ADDITEMS     , &(pNLSStrings->psz25AddItems));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_26CONFIGITEMS  , &(pNLSStrings->psz26ConfigFolderMenus));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_27STATUSBAR    , &(pNLSStrings->psz27StatusBar));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_3SNAPTOGRID    , &(pNLSStrings->psz3SnapToGrid));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_4ACCELERATORS  , &(pNLSStrings->psz4Accelerators));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_5INTERNALS     , &(pNLSStrings->psz5Internals));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_FILEOPS        , &(pNLSStrings->pszFileOps));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SORT           , &(pNLSStrings->pszSort));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SV_ALWAYSSORT  , &(pNLSStrings->pszAlwaysSort));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_INTERNALS      , &(pNLSStrings->pszInternals));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SB_CLASSMNEMONICS   , &(pNLSStrings->pszSBClassMnemonics));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SB_CLASSNOTSUPPORTED, &(pNLSStrings->pszSBClassNotSupported));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_WPSCLASSES     , &(pNLSStrings->pszWpsClasses));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_WPSCLASSLOADED , &(pNLSStrings->pszWpsClassLoaded));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_WPSCLASSLOADINGFAILED, &(pNLSStrings->pszWpsClassLoadingFailed));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_WPSCLASSREPLACEDBY, &(pNLSStrings->pszWpsClassReplacedBy));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_WPSCLASSORPHANS   , &(pNLSStrings->pszWpsClassOrphans));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_WPSCLASSORPHANSINFO, &(pNLSStrings->pszWpsClassOrphansInfo));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_SCHEDULER      , &(pNLSStrings->pszScheduler));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_MEMORY         , &(pNLSStrings->pszMemory));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_ERRORS         , &(pNLSStrings->pszErrors));
                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_WPS            , &(pNLSStrings->pszWPS));

                    cmnLoadString(habDesktop, hmodNLS, ID_XSSI_PROCESSCONTENT , &(pNLSStrings->pszProcessContent));

                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_SETTINGS       , &(pNLSStrings->pszSettings));
                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_SETTINGSNOTEBOOK,&(pNLSStrings->pszSettingsNotebook));
                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_ATTRIBUTES     , &(pNLSStrings->pszAttributes));
                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_ATTR_ARCHIVED  , &(pNLSStrings->pszAttrArchived));
                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_ATTR_SYSTEM    , &(pNLSStrings->pszAttrSystem));
                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_ATTR_HIDDEN    , &(pNLSStrings->pszAttrHidden));
                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_ATTR_READONLY  , &(pNLSStrings->pszAttrReadOnly));

                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_FLDRSETTINGS   , &(pNLSStrings->pszFldrSettings));
                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_SMALLICONS     , &(pNLSStrings->pszSmallIcons  ));
                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_FLOWED         , &(pNLSStrings->pszFlowed));
                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_NONFLOWED      , &(pNLSStrings->pszNonFlowed));
                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_NOGRID         , &(pNLSStrings->pszNoGrid));

                    cmnLoadString(habDesktop, hmodNLS, ID_XFSI_SHOWSTATUSBAR  , &(pNLSStrings->pszShowStatusBar));
                }

                // after all this, unload the old resource module
                DosFreeModule(hmodOldResource);

            } // end else

        }
        else
            DebugBox("XFolder: Fatal Error",
                    "XFolder could not find its path in OS2.INI. "
                    "Please re-install XFolder, which should solve this "
                    "problem.");
    }

    return (hmodNLS);
}

/*
 * cmnQueryNLSStrings:
 *           returns pointer to metaclass NLSSTRINGS structure which contains
 *           all the language-dependent XFolder strings from the resource
 *           files
 */

PNLSSTRINGS cmnQueryNLSStrings(VOID)
{
    if (pNLSStringsGlobal == NULL) {
        pNLSStringsGlobal = malloc(sizeof(NLSSTRINGS));
        memset(pNLSStringsGlobal, 0, sizeof(NLSSTRINGS));
    }
    return (pNLSStringsGlobal);
}

/*
 * cmnQueryStatusBarSetting:
 *          returns a PSZ to a certain status bar setting, which
 *          may be:
 *              SBS_STATUSBARFONT       font (e.g. "9.WarpSans")
 *              SBS_TEXTNONESEL         mnemonics for no-object mode
 *              SBS_TEXTMULTISEL        mnemonics for multi-object mode
 *          Note that there is no key for querying the mnemonics for
 *          one-object mode, because this is now handled by the
 *          WPS classes to provide different data depending on the
 *          class of the selected object.
 */

PSZ cmnQueryStatusBarSetting(USHORT usSetting)
{
    switch (usSetting) {
        case SBS_STATUSBARFONT:
                return (szStatusBarFont);
        case SBS_TEXTNONESEL:
                return (szSBTextNoneSel);
        case SBS_TEXTMULTISEL:
                return (szSBTextMultiSel);
        default:
                return (NULL);
    }
}

/*
 * cmnSetStatusBarSetting:
 *          sets usSetting to pszSetting. If pszSetting == NULL, the
 *          default value will be loaded from the XFolder NLS DLL.
 *          usSetting works just like in cmnQueryStatusBarSetting.
 */

BOOL cmnSetStatusBarSetting(USHORT usSetting, PSZ pszSetting)
{
    HAB habDesktop = WinQueryAnchorBlock(HWND_DESKTOP);
    HMODULE hmodResource = cmnQueryNLSModuleHandle(FALSE);

    switch (usSetting) {
        case SBS_STATUSBARFONT: {
                CHAR szDummy[CCHMAXMNEMONICS];
                if (pszSetting) {
                    strcpy(szStatusBarFont, pszSetting);
                    PrfWriteProfileString(HINI_USERPROFILE, INIAPP, INIKEY_STATUSBARFONT,
                             szStatusBarFont);
                } else
                    strcpy(szStatusBarFont, "8.Helv");
                sscanf(szStatusBarFont, "%d.%s", &(ulStatusBarHeight), &szDummy);
                ulStatusBarHeight += 15;
        break; }
        case SBS_TEXTNONESEL: {
                if (pszSetting) {
                    strcpy(szSBTextNoneSel, pszSetting);
                    PrfWriteProfileString(HINI_USERPROFILE, INIAPP, INIKEY_SBTEXTNONESEL,
                             szSBTextNoneSel);
                } else
                    WinLoadString(habDesktop, hmodResource, ID_XSSI_SBTEXTNONESEL,
                        sizeof(szSBTextNoneSel), szSBTextNoneSel);
        break; }
        case SBS_TEXTMULTISEL: {
                if (pszSetting) {
                    strcpy(szSBTextMultiSel, pszSetting);
                    PrfWriteProfileString(HINI_USERPROFILE, INIAPP, INIKEY_SBTEXTMULTISEL,
                             szSBTextMultiSel);
                } else
                    WinLoadString(habDesktop, hmodResource, ID_XSSI_SBTEXTMULTISEL,
                        sizeof(szSBTextMultiSel), szSBTextMultiSel);
        break; }
        default:
                return (FALSE);
    }
    return (TRUE);
}

/*
 * cmnQueryStatusBarHeight:
 *           returns the height of the status bars according to the
 *           current settings in pixels
 */

ULONG cmnQueryStatusBarHeight(VOID)
{
    return (ulStatusBarHeight);
}

/*
 * cmnLoadGlobalSettings:
 *           this loads the Global Settings from the INI files; should
 *           not be called directly, because this is done automatically
 *           by cmnQueryGlobalSettings, if neccessary
 */

PVOID cmnLoadGlobalSettings(VOID)
{
    ULONG       ulCopied1;

    if (pGlobalSettings == NULL) {
        pGlobalSettings = malloc(sizeof(GLOBALSETTINGS));
        memset(pGlobalSettings, 0, sizeof(GLOBALSETTINGS));
    }

    // first set default settings for each settings page;
    // we only load the "real" settings from OS2.INI afterwards
    // because the user might have updated XFolder, and the
    // settings struct in the INIs might be too short
    cmnSetDefaultSettings(SP_1GENERIC              );
    cmnSetDefaultSettings(SP_2REMOVEITEMS          );
    cmnSetDefaultSettings(SP_25ADDITEMS            );
    cmnSetDefaultSettings(SP_26CONFIGITEMS         );
    cmnSetDefaultSettings(SP_27STATUSBAR           );
    cmnSetDefaultSettings(SP_3SNAPTOGRID           );
    cmnSetDefaultSettings(SP_4ACCELERATORS         );
    cmnSetDefaultSettings(SP_5INTERNALS            );
    cmnSetDefaultSettings(SP_DTP1                  );
    cmnSetDefaultSettings(SP_DTP2                  );
    cmnSetDefaultSettings(SP_FLDRSORT              );
    cmnSetDefaultSettings(SP_FILEOPS               );

    // set global settings w/out settings page
    pGlobalSettings->ShowBootupStatus = 0;
    pGlobalSettings->WpsShowClassInfo = 1;

    PrfQueryProfileString(HINI_USERPROFILE, INIAPP, INIKEY_STATUSBARFONT,
             "8.Helv", &(szStatusBarFont), sizeof(szStatusBarFont));
    sscanf(szStatusBarFont, "%d.*%s", &(ulStatusBarHeight));
    ulStatusBarHeight += 15;

    PrfQueryProfileString(HINI_USERPROFILE, INIAPP, INIKEY_SBTEXTNONESEL,
                NULL, &(szSBTextNoneSel), sizeof(szSBTextNoneSel));
    PrfQueryProfileString(HINI_USERPROFILE, INIAPP, INIKEY_SBTEXTMULTISEL,
                NULL, &(szSBTextMultiSel), sizeof(szSBTextMultiSel));

    ulCopied1 = sizeof(GLOBALSETTINGS);
    // get global XFolder settings from OS2.INI
    PrfQueryProfileData(HINI_USERPROFILE, INIAPP, INIKEY_GLOBALSETTINGS,
                pGlobalSettings, &ulCopied1);

    return (pGlobalSettings);
}

/*
 * cmnQueryGlobalSettings:
 *           returns pointer to metaclass GLOBALSETTINGS structure which
 *           contains the XFolder Global Settings valid for all classes
 */

PGLOBALSETTINGS cmnQueryGlobalSettings(VOID)
{
    if (pGlobalSettings == NULL) {
        cmnLoadGlobalSettings();
    }

    return (pGlobalSettings);
}

/*
 * cmnStoreGlobalSettings:
 *           stores the current Global Settings back into the INI files;
 *           returns TRUE if successful
 */

BOOL cmnStoreGlobalSettings(VOID)
{
    cmnStoreAccelerators();

    return (PrfWriteProfileData(HINI_USERPROFILE, INIAPP, INIKEY_GLOBALSETTINGS,
        pGlobalSettings, sizeof(GLOBALSETTINGS)));
}

/*
 * cmnSetDefaultSettings:
 *           resets those Global Settings which correspond to usSettingsPage
 *           in the System notebook to the default values; usSettingsPage
 *           may have 1 thru the number of XFolder Settings pages in the
 *           System notebook
 */

BOOL cmnSetDefaultSettings(USHORT usSettingsPage)
{
    // HMODULE hmodResource = cmnQueryNLSModuleHandle(FALSE);

    switch(usSettingsPage) {
        case SP_1GENERIC: {
            pGlobalSettings->ShowInternals = 1;
            pGlobalSettings->ReplIcons = 1;
            pGlobalSettings->FullPath = 1;
            pGlobalSettings->KeepTitle = 1;
            pGlobalSettings->MaxPathChars = 25;
            pGlobalSettings->TreeViewAutoScroll = 1;
        break; }

        case SP_2REMOVEITEMS: {
            pGlobalSettings->DefaultMenuItems = 0;
            pGlobalSettings->RemoveLockInPlace = 0;
            pGlobalSettings->RemoveCheckDisk = 0;
            pGlobalSettings->RemoveFormatDisk = 0;
            pGlobalSettings->RemoveView = 0;
            pGlobalSettings->RemoveInsert = 0;
        break; }

        case SP_25ADDITEMS: {
            pGlobalSettings->ExtendSettings = 1;
            pGlobalSettings->CopyFilename = 1;
            pGlobalSettings->Move4Refresh = (DosIsWarp4() ? 1 : 0);
            pGlobalSettings->SelectSome = 1;
            pGlobalSettings->FolderContent = 1;
            pGlobalSettings->FCShowIcons = 0;
        break; }

        case SP_26CONFIGITEMS: {
            pGlobalSettings->Cascade = 0;
            pGlobalSettings->RemoveX = 1;
            pGlobalSettings->AppdParam = 1;
            pGlobalSettings->OpenSettings = BM_INDETERMINATE;
            pGlobalSettings->TplPosition = 1;
        break; }

        case SP_27STATUSBAR: {
            // HAB habDesktop = WinQueryAnchorBlock(HWND_DESKTOP);
            pGlobalSettings->StatusBar = 0;
            pGlobalSettings->SBStyle = (DosIsWarp4() ? SBSTYLE_WARP4MENU : SBSTYLE_WARP3RAISED);
            pGlobalSettings->SBForViews = SBV_ICON | SBV_DETAILS;
            pGlobalSettings->lSBBgndColor = WinQuerySysColor(HWND_DESKTOP, SYSCLR_INACTIVEBORDER, 0);
            pGlobalSettings->lSBTextColor = WinQuerySysColor(HWND_DESKTOP, SYSCLR_OUTPUTTEXT, 0);
            cmnSetStatusBarSetting(SBS_TEXTNONESEL, NULL);
            // cmnSetStatusBarSetting(SBS_TEXT1SEL, NULL);
            cmnSetStatusBarSetting(SBS_TEXTMULTISEL, NULL);
        break; }

        case SP_3SNAPTOGRID: {
            pGlobalSettings->SnapToGrid = 1;
            pGlobalSettings->GridX = 15;
            pGlobalSettings->GridY = 10;
            pGlobalSettings->GridCX = 20;
            pGlobalSettings->GridCY = 35;
        break; }

        case SP_4ACCELERATORS: {
            pGlobalSettings->Accelerators = 1;
        break; }

        case SP_5INTERNALS: {
            pGlobalSettings->VarMenuOffset = 300;
            pGlobalSettings->NoWorkerThread = 0;
            pGlobalSettings->NoSubclassing = 0;
            pGlobalSettings->ShowXFolderAnim = 1;
        break; }

        case SP_DTP1: { // shutdown page
            pGlobalSettings->XShutdown =
                XSD_RESTARTWPS | XSD_WPS_CLOSEWINDOWS | XSD_CONFIRM | XSD_REBOOT | XSD_ANIMATE;
        break; }

        case SP_DTP2: {
            pGlobalSettings->ShowStartupProgress = 1;
            pGlobalSettings->StartupDelay = 1000;
        break; }

        case SP_FLDRSORT: {
            pGlobalSettings->ReplaceSort = 0;
            pGlobalSettings->DefaultSort = SV_NAME;
            pGlobalSettings->AlwaysSort = 0;
        break; }

        case SP_FILEOPS: {
            pGlobalSettings->ReplConfirms = 1;
            pGlobalSettings->CleanupINIs = 1;
        break; }
    }

    return (TRUE);
}

/*
 * cmnOpenDebugFile:
 *      this opens or creates C:\XFLDTRAP.LOG and writes
 *      a debug header into it (date and time); returns
 *      the handle of the file, so additional data can
 *      be written; after this, DosClose must be called
 */

void cmnOpenDebugFile(VOID)
{
    ULONG   ulAction;
    ULONG   cbWritten;
    // ULONG   ulBootDrive = 0;

    APIRET  rc;
    CHAR    szFileName[CCHMAXPATH];
    CHAR    szBuffer[1000];
            // szBuffer2[100];
    DATETIME DT;
    ULONG     ibActual;

    // ULONG ulCopied;

    if (hfDebug)
        return;

    DosBeep( 250, 30);
    DosBeep( 500, 30);
    DosBeep(1000, 30);
    DosBeep(2000, 30);
    DosBeep(4000, 30);
    DosBeep(2000, 30);
    DosBeep(1000, 30);
    DosBeep( 500, 30);
    DosBeep( 250, 30);

    sprintf(szFileName, "%c:\\%s", DosQueryBootDrive(), XFOLDER_CRASHLOG);
    rc = DosOpen(szFileName, &hfDebug, &ulAction, 0,
                 FILE_NORMAL,
                 OPEN_ACTION_CREATE_IF_NEW | OPEN_ACTION_OPEN_IF_EXISTS,
                 OPEN_ACCESS_WRITEONLY | OPEN_SHARE_DENYWRITE,
                 (PEAOP2)NULL);

    DosSetFilePtr(hfDebug, 0, FILE_END, &ibActual);

    if (!rc) {
        DosGetDateTime(&DT);
        sprintf(szBuffer, "\n\nXFolder debug message -- Date: %02d/%02d/%04d, Time: %02d:%02d:%02d\n",
            DT.month, DT.day, DT.year,
            DT.hours, DT.minutes, DT.seconds);
        strcat(szBuffer, "---------------------------------------------------------\n"
                         "\nAn internal error occurred within XFolder.\n"
                         "Please contact the author so that this error may be removed\n"
                         "in future XFolder versions. The author's e-mail address may\n"
                         "be obtained from the XFolder Online Reference. Please supply\n"
                         "this file (?:\\XFLDTRAP.LOG) with your e-mail and describe as\n"
                         "exactly as possible the conditions under which the error\n"
                         "occured.\n"
                         "\nRunning XFolder version: " XFOLDER_VERSION "\n");

        DosWrite(hfDebug, szBuffer, strlen(szBuffer), &cbWritten);
    }
    return;
}

/*
 * cmnWriteAndCloseDebugFile:
 *
 */

void cmnWriteAndCloseDebugFile(PSZ pszBuffer)
{
    ULONG cbWritten;
    DosWrite(hfDebug, pszBuffer, strlen(pszBuffer), &cbWritten);
    DosClose(hfDebug);
    hfDebug = NULLHANDLE;
}

/*
 * cmnError:
 *
 */

APIRET cmnError(PSZ psz, APIRET rc)
{
    if (rc)
        _Pmpf(( "%s%d", psz, rc ));
    return (rc);
}

/*
 * fnwpAutoSizeStatic:
 *      dlg proc for the subclassed static text control in msg box;
 *      automatically resizes the msg box window when text changes
 */

PFNWP pfnwpOrigStatic = NULL;

MRESULT EXPENTRY fnwpAutoSizeStatic(HWND hwndStatic, ULONG msg, MPARAM mp1, MPARAM mp2)
{
    MRESULT mrc = NULL;
    switch (msg) {

        /*
         * WM_SETLONGTEXT:
         *      this arrives here when the window text changes,
         *      especially when fnwpMessageBox sets the window
         *      text; we will now reposition all the controls.
         *      mp1 is a PSZ to the new text.
         */

        case WM_SETLONGTEXT: {
            CHAR szFont[300];
            PSZ p = (PSZ)WinQueryWindowULong(hwndStatic, QWL_USER);
            if (p)
                free(p);
            p = strdup((PSZ)mp1);
            WinSetWindowULong(hwndStatic, QWL_USER, (ULONG)p);
            // WinPostMsg(hwndStatic, WM_UPDATE, MPNULL, MPNULL);
            PrfQueryProfileString(HINI_USER, INIAPP, INIKEY_DLGFONT, "9.WarpSans",
                szFont, sizeof(szFont)-1);
            WinSetPresParam(hwndStatic, PP_FONTNAMESIZE,
                 (ULONG)strlen(szFont) + 1, (PVOID)szFont);
            // this will also update the display

            /* PSZ pwp = (PWNDPARAMS)mp1;
            _Pmpf(( "WM_SETWINDOWPARAMS" ));
            mrc = (*pfnwpOrigStatic)(hwndStatic, msg, mp1, mp2);
            if (pwp) {
            } */
        break; }

        /*
         * WM_PRESPARAMCHANGED:
         *      if a font has been dropped on the window, store
         *      the information and enforce control repositioning
         *      also
         */

        case WM_PRESPARAMCHANGED: {
            // _Pmpf(( "PRESPARAMCHANGED" ));
            switch ((ULONG)mp1) {
                case PP_FONTNAMESIZE: {
                    ULONG attrFound; //, abValue[32];
                    CHAR  szFont[100];
                    HWND  hwndDlg = WinQueryWindow(hwndStatic, QW_PARENT);
                    WinQueryPresParam(hwndStatic,
                                PP_FONTNAMESIZE,
                                0,
                                &attrFound,
                                (ULONG)sizeof(szFont),
                                (PVOID)&szFont,
                                0);
                    PrfWriteProfileString(HINI_USER, INIAPP, INIKEY_DLGFONT,
                        szFont);

                    // now also change the buttons
                    WinSetPresParam(WinWindowFromID(hwndDlg, 1),
                        PP_FONTNAMESIZE,
                        (ULONG)strlen(szFont) + 1, (PVOID)szFont);
                    WinSetPresParam(WinWindowFromID(hwndDlg, 2),
                        PP_FONTNAMESIZE,
                        (ULONG)strlen(szFont) + 1, (PVOID)szFont);
                    WinSetPresParam(WinWindowFromID(hwndDlg, 3),
                        PP_FONTNAMESIZE,
                        (ULONG)strlen(szFont) + 1, (PVOID)szFont);

                    WinPostMsg(hwndStatic, WM_UPDATE, MPNULL, MPNULL);
                break; }
            }
        break; }

        /*
         * WM_UPDATE:
         *      this actually does all the calculations to reposition
         *      all the msg box controls
         */

        case WM_UPDATE: {
            RECTL   rcl, rcl2;
            SWP     swp;
            HWND    hwndIcon, hwndDlg;
            PSZ     pszText;

            HPS hps = WinGetPS(hwndStatic);

            // _Pmpf(( "WM_UPDATE" ));

            pszText = (PSZ)WinQueryWindowULong(hwndStatic, QWL_USER);
            if (pszText) {
                WinQueryWindowRect(hwndStatic, &rcl);         // get window dimensions
                // rcl.yTop = 1000;
                memcpy(&rcl2, &rcl, sizeof(rcl));

                WinDrawFormattedText(hps, &rcl, pszText, DT_LEFT | DT_TOP | DT_QUERYEXTENT);
                WinEndPaint(hps);

                if ((rcl.yTop - rcl.yBottom) < 40)
                    rcl.yBottom -= 40 - (rcl.yTop - rcl.yBottom);

                // printfRtl("  Original rect:", &rcl2);
                // printfRtl("  New rect:     ", &rcl);

                // reposition the text
                WinQueryWindowPos(hwndStatic, &swp);
                swp.cy -= rcl.yBottom;
                WinSetWindowPos(hwndStatic, 0,
                    swp.x, swp.y, swp.cx, swp.cy,
                    SWP_SIZE);

                // reposition the icon
                hwndIcon = WinWindowFromID(WinQueryWindow(hwndStatic, QW_PARENT),
                                ID_XFDI_GENERICDLGICON);
                WinQueryWindowPos(hwndIcon, &swp);
                swp.y -= rcl.yBottom;
                WinSetWindowPos(hwndIcon, 0,
                    swp.x, swp.y, swp.cx, swp.cy,
                    SWP_MOVE);

                // resize the dlg frame window
                hwndDlg = WinQueryWindow(hwndStatic, QW_PARENT);
                WinQueryWindowPos(hwndDlg, &swp);
                swp.cy -= rcl.yBottom;
                swp.y  += (rcl.yBottom / 2);
                WinInvalidateRect(hwndDlg, NULL, FALSE);
                WinSetWindowPos(hwndDlg, 0,
                    swp.x, swp.y, swp.cx, swp.cy,
                    SWP_SIZE | SWP_MOVE);
            }
        break; }

        /*
         * WM_PAINT:
         *      draw the formatted text
         */

        case WM_PAINT: {
            RECTL   rcl;
            PSZ pszText = (PSZ)WinQueryWindowULong(hwndStatic, QWL_USER);
            HPS hps = WinBeginPaint(hwndStatic, NULLHANDLE, &rcl);
            if (pszText) {
                WinQueryWindowRect(hwndStatic, &rcl);         // get window dimensions
                // switch to RGB mode
                GpiCreateLogColorTable(hps, 0, LCOLF_RGB, 0, 0, NULL);
                // set "static text" color
                GpiSetColor(hps,
                        WinQuerySysColor(HWND_DESKTOP, SYSCLR_WINDOWSTATICTEXT, 0));
                WinDrawFormattedText(hps, &rcl, pszText, DT_LEFT | DT_TOP);
            }
            WinEndPaint(hps);
        break; }

        case WM_DESTROY: {
            PSZ pszText = (PSZ)WinQueryWindowULong(hwndStatic, QWL_USER);
            free(pszText);
            mrc = (*pfnwpOrigStatic)(hwndStatic, msg, mp1, mp2);
        break; }
        default:
            mrc = (*pfnwpOrigStatic)(hwndStatic, msg, mp1, mp2);
    }
    return (mrc);
}

/*
 * fnwpMessageBox:
 *      dlg proc for cmnMessageBox below
 */

MRESULT EXPENTRY fnwpMessageBox(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2)
{
    MRESULT mrc = NULL;
    switch (msg) {
        case WM_INITDLG: {
            // CHAR szFont[CCHMAXPATH];
            HWND hwndStatic = WinWindowFromID(hwndDlg,
                                ID_XFDI_GENERICDLGTEXT);

            /* PrfQueryProfileString(HINI_USER, INIAPP, INIKEY_DLGFONT,
                    "9.WarpSans", szFont, sizeof(szFont)-1);
            WinSetPresParam(hwndStatic, PP_FONTNAMESIZE,
                 (ULONG)strlen(szFont)+1, (PVOID)szFont); */

            // set string to 0
            WinSetWindowULong(hwndStatic, QWL_USER, (ULONG)NULL);
            pfnwpOrigStatic = WinSubclassWindow(hwndStatic,
                    fnwpAutoSizeStatic);
            mrc = fnwpDlgGeneric(hwndDlg, msg, mp1, mp2);
        break; }

        default:
            mrc = fnwpDlgGeneric(hwndDlg, msg, mp1, mp2);
        break;
    }
    return (mrc);
}

/*
 * cmnMessageBox:
 *      this is a new message box function; it works
 *      pretty similar to WinMessageBox, but displays
 *      a user-defined icon instead. It also adjusts
 *      itself according to font changes and such.
 */

ULONG cmnMessageBox(HWND hwndOwner, PSZ pszTitle, PSZ pszMessage, ULONG flStyle)
{
    HAB     hab = WinQueryAnchorBlock(hwndOwner);
    HMODULE hmod = NLS_MODULE;
    HWND    hwndDlg = NULLHANDLE, hwndStatic = NULLHANDLE;
    PSZ     pszButton = NULL;
    ULONG   ulrcDlg, ulrc = DID_CANCEL;

    REGREC2               RegRec2;
    ULONG                 ulExcpt;
    APIRET                rc;

    PULONG pulWorkplaceFunc2 = cmnGetWorkplaceFunc2();

    *pulWorkplaceFunc2 = 9500;
    // set our extended exception handler
    RegRec2.pfnHandler = (PFN)ExcptHandlerQuiet;
    if ( rc = DosSetExceptionHandler((PEXCEPTIONREGISTRATIONRECORD)&RegRec2) )
        DosBeep(100, 1000);
    // store a known thread state
    ulExcpt = setjmp(RegRec2.jmpThread);
    if (ulExcpt)
    {
        DosBeep(100, 1000);
    } else {
        ULONG flButtons = flStyle & 0xF;        // low nibble contains MB_YESNO etc.

        hwndDlg = WinLoadDlg(HWND_DESKTOP, hwndOwner,
                        fnwpMessageBox,
                        hmod,
                        ID_XFD_GENERICDLG,
                        NULL);

        hwndStatic = WinWindowFromID(hwndDlg, ID_XFDI_GENERICDLGTEXT);

        *pulWorkplaceFunc2 = 9510;
        // now work on the three buttons of the dlg template:
        // give them proper titles or hide them
        if (flButtons == MB_YESNO) {
            cmnLoadString(hab, hmod, ID_XSSI_DLG_YES, &pszButton);
            WinSetDlgItemText(hwndDlg, 1, pszButton);
            cmnLoadString(hab, hmod, ID_XSSI_DLG_NO, &pszButton);
            WinSetDlgItemText(hwndDlg, 2, pszButton);
            WinShowDlgItem(hwndDlg, 3, FALSE);
            free(pszButton);
        }
        else if (flButtons == MB_OK) {
            cmnLoadString(hab, hmod, ID_XSSI_DLG_OK, &pszButton);
            WinSetDlgItemText(hwndDlg, 1, pszButton);
            WinShowDlgItem(hwndDlg, 2, FALSE);
            WinShowDlgItem(hwndDlg, 3, FALSE);
            free(pszButton);
        }
        else if (flButtons == MB_CANCEL) {
            cmnLoadString(hab, hmod, ID_XSSI_DLG_CANCEL, &pszButton);
            WinSetDlgItemText(hwndDlg, 1, pszButton);
            WinShowDlgItem(hwndDlg, 2, FALSE);
            WinShowDlgItem(hwndDlg, 3, FALSE);
            free(pszButton);
        }
        else if (flButtons == MB_OKCANCEL) {
            cmnLoadString(hab, hmod, ID_XSSI_DLG_OK, &pszButton);
            WinSetDlgItemText(hwndDlg, 1, pszButton);
            cmnLoadString(hab, hmod, ID_XSSI_DLG_CANCEL, &pszButton);
            WinSetDlgItemText(hwndDlg, 2, pszButton);
            WinShowDlgItem(hwndDlg, 3, FALSE);
            free(pszButton);
        }
        else if (flButtons == MB_ABORTRETRYIGNORE) {
            cmnLoadString(hab, hmod, ID_XSSI_DLG_ABORT, &pszButton);
            WinSetDlgItemText(hwndDlg, 1, pszButton);
            cmnLoadString(hab, hmod, ID_XSSI_DLG_RETRY, &pszButton);
            WinSetDlgItemText(hwndDlg, 2, pszButton);
            cmnLoadString(hab, hmod, ID_XSSI_DLG_IGNORE, &pszButton);
            WinSetDlgItemText(hwndDlg, 3, pszButton);
            free(pszButton);
        }

        if (flStyle & MB_DEFBUTTON2)
            WinSetWindowBits(WinWindowFromID(hwndDlg, 2), QWL_STYLE,
                    BS_DEFAULT,
                    1);
        else if (flStyle & MB_DEFBUTTON3)
            WinSetWindowBits(WinWindowFromID(hwndDlg, 3), QWL_STYLE,
                    BS_DEFAULT,
                    1);
        else
            WinSetWindowBits(WinWindowFromID(hwndDlg, 1), QWL_STYLE,
                    BS_DEFAULT,
                    1);

        *pulWorkplaceFunc2 = 9520;
        WinSetWindowText(hwndDlg, pszTitle);
        WinSendMsg(hwndStatic, WM_SETLONGTEXT, pszMessage, MPNULL);

        *pulWorkplaceFunc2 = 9530;
        WinCenterWindow(hwndDlg);
        if (flStyle & MB_SYSTEMMODAL)
            WinSetSysModalWindow(HWND_DESKTOP, hwndDlg);
        ulrcDlg = WinProcessDlg(hwndDlg);
        WinDestroyWindow(hwndDlg);

        *pulWorkplaceFunc2 = 9540;
        if (flButtons == MB_YESNO)
            switch (ulrcDlg) {
                case 1:     ulrc = MBID_YES; break;
                default:    ulrc = MBID_NO;  break;
            }
        else if (flButtons == MB_OK)
            ulrc = MBID_OK;
        else if (flButtons == MB_CANCEL)
            ulrc = MBID_CANCEL;
        else if (flButtons == MB_OKCANCEL)
            switch (ulrcDlg) {
                case 1:     ulrc = MBID_OK; break;
                default:    ulrc = MBID_CANCEL;  break;
            }
        else if (flButtons == MB_ABORTRETRYIGNORE)
            switch (ulrcDlg) {
                case 1:     ulrc = MBID_ABORT; break;
                case 3:     ulrc = MBID_IGNORE; break;
                default:    ulrc = MBID_RETRY;  break;
            }
    }

    *pulWorkplaceFunc2 = 9590;
    DosUnsetExceptionHandler((PEXCEPTIONREGISTRATIONRECORD)&RegRec2);
    *pulWorkplaceFunc2 = 9591;

    return (ulrc);
}

/*
 * cmnGetMessage:
 *      like DosGetMessage, but automatically finds the
 *      (NLS) XFolder message file.
 *      Parameters:
 *          PCHAR *pTable     array[ulTable] of PSZs with
 *                            substitutes for "%1" etc. in
 *                            the msg string
 *          ULONG ulTable     size of that array
 *          PSZ pszBuf        buffer to hold message string
 *          ULONG cbBuf       size of pszBuf
 *          ULONG ulMsgNumber msg number to retrieve
 *
 */

APIRET cmnGetMessage(PCHAR *pTable, ULONG ulTable, PSZ pszBuf, ULONG cbBuf,
                        ULONG ulMsgNumber)
{
    PSZ     pszMessageFile = cmnQueryMessageFile();
    ULONG   ulReturned;
    BOOL    fCont = FALSE;
    APIRET  arc;
    arc = DosGetMessage(pTable, ulTable,
                        pszBuf, cbBuf,
                        ulMsgNumber, pszMessageFile, &ulReturned);
    pszBuf[ulReturned] = '\0';

    // remove trailing newlines
    do {
        PSZ p = pszBuf + strlen(pszBuf) - 1;
        if (    (*p == '\n')
             || (*p == '\r')
           )
        {
            *p = '\0';
            fCont = TRUE;
        } else
            fCont = FALSE;
    } while (fCont);

    return (arc);
}

/*
 * cmnMessageBoxMsg:
 *      calls cmnMessageBox, but this one accepts ULONG indexes
 *      into the XFolder message file (XFLDRxxx.MSG) instead
 *      of real PSZs
 */

ULONG cmnMessageBoxMsg(HWND hwndOwner, ULONG ulTitle, ULONG ulMessage, ULONG flStyle)
{
    CHAR    szTitle[200], szMessage[2000];

    cmnGetMessage(NULL, 0,
            szTitle, sizeof(szTitle)-1,
            ulTitle);
    cmnGetMessage(NULL, 0,
            szMessage, sizeof(szMessage)-1,
            ulMessage);

    return (cmnMessageBox(hwndOwner, szTitle, szMessage, flStyle));
}

/*
 * cmnMessageBoxMsgExt:
 *      like CmnMessageBoxMsg, but with string substitution
 *      (see DosGetMessage docs for more); substitution only
 *      takes place for the message specified with ulMessage,
 *      not for the title
 */

ULONG cmnMessageBoxMsgExt(HWND hwndOwner, ULONG ulTitle,
        PCHAR *pTable, ULONG ulTable, ULONG ulMessage, ULONG flStyle)
{
    CHAR    szTitle[200], szMessage[2000];

    cmnGetMessage(NULL, 0,
            szTitle, sizeof(szTitle)-1,
            ulTitle);
    cmnGetMessage(pTable, ulTable,
            szMessage, sizeof(szMessage)-1,
            ulMessage);

    return (cmnMessageBox(hwndOwner, szTitle, szMessage, flStyle));
}

/*
 * cmnSetHelpPanel:
 *      sets help panel before calling fnwpDlgGeneric
 */

VOID cmnSetHelpPanel(ULONG ulHelpPanel)
{
    ulCurHelpPanel = ulHelpPanel;
}

/*
 * cmnAnimate:
 *      proceeds to the next animation step
 */

VOID cmnAnimate(PANIMATION pa)
{
    if (pa)
    {
        pa->usAniCurrent++;
        if (pa->usAniCurrent >= pa->usAniCount)
            pa->usAniCurrent = 0;

        if (pa->hwndIcon) {
            WinSendMsg(pa->hwndIcon,
                    SM_SETHANDLE,
                    (MPARAM)pa->ahptrAniIcons[pa->usAniCurrent],
                    MPNULL);
        }
    }
}

/*
 *  fnwpDlgGeneric:
 *          this is the dlg procedure for XFolder dlg boxes;
 *          it can process WM_HELP messages
 */

MRESULT EXPENTRY fnwpDlgGeneric(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
    HOBJECT hobjRef = 0;
    HWND    hwndItem;
    MRESULT mrc = NULL;

    switch (msg) {
        case WM_INITDLG:
            if (hwndItem = WinWindowFromID(hwnd, ID_XFDI_XFLDVERSION)) {
                CHAR    szTemp[100];
                WinQueryWindowText(hwndItem, sizeof(szTemp), szTemp);
                strcpy(strstr(szTemp, "%a"), XFOLDER_VERSION);
                WinSetWindowText(hwndItem, szTemp);
            }
            mrc = WinDefDlgProc(hwnd, msg, mp1, mp2);
        break;

        case WM_ANIMATE: {
            PANIMATION pa = (PANIMATION)mp1;
            if (pa)
            {
                pa->usAniCurrent = 0;
                WinSetWindowULong(hwnd, QWL_USER, (ULONG)pa);
                WinStartTimer(WinQueryAnchorBlock(hwnd), hwnd, 1, pa->ulDelay);
            }
        break; }

        case WM_TIMER: {
            PANIMATION pa = (PANIMATION)WinQueryWindowULong(hwnd, QWL_USER);
            cmnAnimate(pa);
        break; }

        case WM_HELP: {
            // HMODULE hmodResource = cmnQueryNLSModuleHandle(FALSE);
            /* WM_HELP is received by this function when F1 or a "help" button
               is pressed in a dialog window. */
            if (ulCurHelpPanel > 0) {
                WPObject    *pHelpSomSelf = _wpclsQueryActiveDesktop(_WPDesktop);
                /* ulCurHelpPanel is set by instance methods before creating a
                   dialog box in order to link help topics to the displayed
                   dialog box. Possible values are:
                        0: open online reference ("<XFOLDER_REF>", INF book)
                      > 0: open help topic in xfldr.hlp
                       -1: ignore WM_HELP */
                if (pHelpSomSelf) {
                    PSZ pszHelpLibrary;
                    BOOL fProcessed = FALSE;
                    if (pszHelpLibrary = cmnQueryHelpLibrary())
                        // path found: display help panel
                        if (_wpDisplayHelp(pHelpSomSelf, ulCurHelpPanel, pszHelpLibrary))
                            fProcessed = TRUE;

                    if (!fProcessed)
                        cmnMessageBoxMsg(HWND_DESKTOP, 104, 134, MB_OK);
                }
            } else if (ulCurHelpPanel == 0)
            { // open online reference
                ulCurHelpPanel = -1; // ignore further WM_HELP messages: this one suffices
                hobjRef = WinQueryObject("<XFOLDER_REF>");
                if (hobjRef) {
                    WinOpenObject(hobjRef, OPEN_DEFAULT, TRUE);
                } else {
                    cmnMessageBoxMsg(HWND_DESKTOP, 104, 137, MB_OK);

                    /* WinDlgBox(HWND_DESKTOP,
                              hwnd,                     // destroy with first msg box
                              (PFNWP)fnwpDlgGeneric,
                              hmodResource, // load from resource file
                              ID_XFD_NOREF,             // dialog resource id
                              (PVOID)NULL);             // no dialog parameters
                              */
                }
            } // end else; if ulCurHelpPanel is < 0, nothing happens
            mrc = NULL;
        break; } // end case WM_HELP

        default:
            mrc = WinDefDlgProc(hwnd, msg, mp1, mp2);
        break;
    }

    return (mrc);
}

/*
 * fnwpRebootExt:
 *      dlg proc for the "Extended Reboot" options
 */

typedef struct REBOOTLISTITEM_STRUCT {
    struct REBOOTLISTITEM_STRUCT *pNext, *pPrevious;
    ULONG                        ulSize;
    CHAR        szItemName[100];
    CHAR        szCommand[CCHMAXPATH];
} REBOOTLISTITEM, *PREBOOTLISTITEM;

typedef struct _REBOOTWINDATA
{
    PREBOOTLISTITEM pliReboot;
    PANIMATION      pAnimation;
    SHORT           sSelected;
    PREBOOTLISTITEM pliSelected;
    USHORT          usItemCount;
} REBOOTWINDATA, *PREBOOTWINDATA;

MRESULT EXPENTRY fnwpRebootExt(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2)
{
    MRESULT mrc = (MPARAM)NULL;

    switch (msg) {
        case WM_INITDLG: {
            ULONG       ulKeyLength;
            PSZ         p, pINI;

            // create window data in QWL_USER
            PREBOOTWINDATA pData = malloc(sizeof(REBOOTWINDATA));
            memset(pData, 0, sizeof(REBOOTWINDATA));

            // set animation
            pData->pAnimation = malloc( sizeof(ANIMATION) + (7 * sizeof(HPOINTER)) );

            pData->pAnimation->cb = sizeof(ANIMATION) + (7 * sizeof(HPOINTER));
            pData->pAnimation->hwndIcon = WinWindowFromID(hwndDlg, ID_SDDI_ICON);
            pData->pAnimation->usAniCount = 8;
            pData->pAnimation->ulDelay = 150;
            cmnLoadShutdownAnimation(&(pData->pAnimation->ahptrAniIcons[0]),
                    _xfclsQueryModuleHandle(_XFldObject));
            WinPostMsg(hwndDlg, WM_ANIMATE, (MPARAM)pData->pAnimation, MPNULL);


            pData->usItemCount = 0;

            // get existing items from INI
            if (PrfQueryProfileSize(HINI_USER,
                        INIAPP, INIKEY_BOOTMGR,
                        &ulKeyLength))
            {
                // _Pmpf(( "Size: %d", ulKeyLength ));
                // items exist: evaluate
                pINI = malloc(ulKeyLength);
                if (pINI)
                {
                    PrfQueryProfileData(HINI_USER,
                                INIAPP, INIKEY_BOOTMGR,
                                pINI,
                                &ulKeyLength);
                    p = pINI;
                    // _Pmpf(( "%s", p ));
                    while (strlen(p)) {
                        PREBOOTLISTITEM pliNew = malloc(sizeof(REBOOTLISTITEM));
                        strcpy(pliNew->szItemName, p);
                        lstAppendItem((PLISTITEM*)&(pData->pliReboot), NULL,
                                (PLISTITEM)pliNew, 0, 0);

                        WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                        LM_INSERTITEM,
                                        (MPARAM)LIT_END,
                                        (MPARAM)p);
                        p += (strlen(p)+1);

                        if (strlen(p)) {
                            strcpy(pliNew->szCommand, p);
                            p += (strlen(p)+1);
                        }
                        pData->usItemCount++;
                    }
                    free(pINI);
                }
            }

            WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_ITEMNAME,
                EM_SETTEXTLIMIT, (MPARAM)(100-1), MPNULL);
            WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_COMMAND,
                EM_SETTEXTLIMIT, (MPARAM)(CCHMAXPATH-1), MPNULL);

            WinSetWindowULong(hwndDlg, QWL_USER, (ULONG)pData);

            WinPostMsg(hwndDlg, WM_UPDATE, MPNULL, MPNULL);
        break; }

        case WM_CONTROL: {
            switch (SHORT1FROMMP(mp1)) {
                case ID_XSDI_XRB_LISTBOX: {
                    if (SHORT2FROMMP(mp1) == LN_SELECT)
                        WinSendMsg(hwndDlg, WM_UPDATE, MPNULL, MPNULL);
                break; }

                case ID_XSDI_XRB_ITEMNAME: {
                    if (SHORT2FROMMP(mp1) == EN_KILLFOCUS) {
                        PREBOOTWINDATA pData =
                                (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
                        // _Pmpf(( "WM_CONTROL ID_XSDI_XRB_ITEMNAME EN_KILLFOCUS" ));
                        if (pData) {
                            if (pData->pliSelected) {
                                WinQueryDlgItemText(hwndDlg, ID_XSDI_XRB_ITEMNAME,
                                    sizeof(pData->pliSelected->szItemName)-1,
                                    pData->pliSelected->szItemName);
                                WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                    LM_SETITEMTEXT,
                                    (MPARAM)pData->sSelected,
                                    (MPARAM)(pData->pliSelected->szItemName));
                            }
                        }
                    }
                break; }

                case ID_XSDI_XRB_COMMAND: {
                    if (SHORT2FROMMP(mp1) == EN_KILLFOCUS) {
                        PREBOOTWINDATA pData =
                                (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
                        // _Pmpf(( "WM_CONTROL ID_XSDI_XRB_COMMAND EN_KILLFOCUS" ));
                        if (pData) {
                            if (pData->pliSelected)
                                WinQueryDlgItemText(hwndDlg, ID_XSDI_XRB_COMMAND,
                                    sizeof(pData->pliSelected->szCommand)-1,
                                    pData->pliSelected->szCommand);
                        }
                    }
                break; }

                default:
                    mrc = fnwpDlgGeneric(hwndDlg, msg, mp1, mp2);
            }
        break; }

        case WM_UPDATE: {
            PREBOOTWINDATA pData =
                    (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
            // _Pmpf(( "WM_CONTROL ID_XSDI_XRB_LISTBOX LN_SELECT" ));
            if (pData) {
                pData->pliSelected = NULL;
                pData->sSelected = (USHORT)WinSendDlgItemMsg(hwndDlg,
                        ID_XSDI_XRB_LISTBOX,
                        LM_QUERYSELECTION,
                        (MPARAM)LIT_CURSOR,
                        MPNULL);
                // _Pmpf(( "  Selected: %d", pData->sSelected ));
                if (pData->sSelected != LIT_NONE) {
                    pData->pliSelected = (PREBOOTLISTITEM)lstItemFromIndex(
                            (PLISTITEM)(pData->pliReboot),
                            pData->sSelected, 0, 0);
                }

                if (pData->pliSelected) {
                    WinSetDlgItemText(hwndDlg, ID_XSDI_XRB_ITEMNAME,
                            pData->pliSelected->szItemName);
                    WinSetDlgItemText(hwndDlg, ID_XSDI_XRB_COMMAND,
                            pData->pliSelected->szCommand);
                } else {
                    WinSetDlgItemText(hwndDlg, ID_XSDI_XRB_ITEMNAME,
                            "");
                    WinSetDlgItemText(hwndDlg, ID_XSDI_XRB_COMMAND,
                            "");
                }
                WinEnableDlgItem(hwndDlg, ID_XSDI_XRB_ITEMNAME,
                            (pData->pliSelected != NULL));
                WinEnableDlgItem(hwndDlg, ID_XSDI_XRB_COMMAND ,
                            (pData->pliSelected != NULL));
                WinEnableDlgItem(hwndDlg, ID_XSDI_XRB_UP      ,
                            (   (pData->pliSelected != NULL)
                             && (pData->usItemCount > 1)
                             && (pData->sSelected > 0)
                            ));
                WinEnableDlgItem(hwndDlg, ID_XSDI_XRB_DOWN    ,
                            (   (pData->pliSelected != NULL)
                             && (pData->usItemCount > 1)
                             && (pData->sSelected < (pData->usItemCount-1))
                            ));

                WinEnableDlgItem(hwndDlg, ID_XSDI_XRB_DELETE,
                            (   (pData->usItemCount > 0)
                             && (pData->pliSelected)
                            )
                        );
            }
        break; }

        case WM_COMMAND: {
            switch (SHORT1FROMMP(mp1)) {

                /*
                 * ID_XSDI_XRB_NEW:
                 *      create new item
                 */

                case ID_XSDI_XRB_NEW: {
                    PREBOOTWINDATA pData =
                            (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
                    PREBOOTLISTITEM pliNew = malloc(sizeof(REBOOTLISTITEM));
                    // _Pmpf(( "WM_COMMAND ID_XSDI_XRB_NEW BN_CLICKED" ));
                    strcpy(pliNew->szItemName, "???");
                    strcpy(pliNew->szCommand, "???");
                    lstAppendItem((PLISTITEM*)&(pData->pliReboot), NULL,
                            (PLISTITEM)pliNew, 0, 0);

                    pData->usItemCount++;
                    WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                    LM_INSERTITEM,
                                    (MPARAM)LIT_END,
                                    (MPARAM)pliNew->szItemName);
                    WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                    LM_SELECTITEM,
                                    (MPARAM)(lstCountItems(
                                            (PLISTITEM)(pData->pliReboot),
                                            0, 0) - 1),
                                    (MPARAM)TRUE);
                    WinSetDlgItemFocus(hwndDlg, ID_XSDI_XRB_ITEMNAME);
                    WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_ITEMNAME,
                            EM_SETSEL,
                            MPFROM2SHORT(0, 1000), // select all
                            MPNULL);
                break; }

                /*
                 * ID_XSDI_XRB_DELETE:
                 *      delete delected item
                 */

                case ID_XSDI_XRB_DELETE: {
                    PREBOOTWINDATA pData =
                            (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
                    // _Pmpf(( "WM_COMMAND ID_XSDI_XRB_DELETE BN_CLICKED" ));
                    if (pData) {
                        if (pData->pliSelected) {
                            lstRemoveItem((PLISTITEM*)&(pData->pliReboot), NULL,
                                    (PLISTITEM)pData->pliSelected, 0, 0);
                            WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                    LM_DELETEITEM,
                                    (MPARAM)pData->sSelected,
                                    MPNULL);
                        }
                        WinPostMsg(hwndDlg, WM_UPDATE, MPNULL, MPNULL);
                    }
                    WinSetDlgItemFocus(hwndDlg, ID_XSDI_XRB_LISTBOX);
                break; }

                /*
                 * ID_XSDI_XRB_UP:
                 *      move selected item up
                 */

                case ID_XSDI_XRB_UP: {
                    PREBOOTWINDATA pData =
                            (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
                    if (pData) {
                        // _Pmpf(( "WM_COMMAND ID_XSDI_XRB_UP BN_CLICKED" ));
                        if (pData->pliSelected) {
                            PREBOOTLISTITEM pliNew = malloc(sizeof(REBOOTLISTITEM));
                            *pliNew = *(pData->pliSelected);
                            // remove selected
                            lstRemoveItem((PLISTITEM*)&(pData->pliReboot), NULL,
                                    (PLISTITEM)pData->pliSelected, 0, 0);
                            WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                    LM_DELETEITEM,
                                    (MPARAM)pData->sSelected,
                                    MPNULL);
                            // insert item again
                            lstInsertItemAt((PLISTITEM*)&(pData->pliReboot), NULL,
                                    (PLISTITEM)pliNew,
                                    (pData->sSelected-1),
                                    0, 0);
                            WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                            LM_INSERTITEM,
                                            (MPARAM)(pData->sSelected-1),
                                            (MPARAM)pliNew->szItemName);
                            WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                            LM_SELECTITEM,
                                            (MPARAM)(pData->sSelected-1),
                                            (MPARAM)TRUE); // select flag
                        }
                        WinPostMsg(hwndDlg, WM_UPDATE, MPNULL, MPNULL);
                    }
                    WinSetDlgItemFocus(hwndDlg, ID_XSDI_XRB_LISTBOX);
                break; }

                /*
                 * ID_XSDI_XRB_DOWN:
                 *      move selected item down
                 */

                case ID_XSDI_XRB_DOWN: {
                    PREBOOTWINDATA pData =
                            (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
                    if (pData) {
                        // _Pmpf(( "WM_COMMAND ID_XSDI_XRB_DOWN BN_CLICKED" ));
                        if (pData->pliSelected) {
                            PREBOOTLISTITEM pliNew = malloc(sizeof(REBOOTLISTITEM));
                            *pliNew = *(pData->pliSelected);
                            // remove selected
                            // _Pmpf(( "  Removing index %d", pData->sSelected ));
                            lstRemoveItem((PLISTITEM*)&(pData->pliReboot), NULL,
                                    (PLISTITEM)pData->pliSelected, 0, 0);
                            WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                    LM_DELETEITEM,
                                    (MPARAM)pData->sSelected,
                                    MPNULL);
                            // insert item again
                            lstInsertItemAt((PLISTITEM*)&(pData->pliReboot), NULL,
                                    (PLISTITEM)pliNew,
                                    (pData->sSelected+1),
                                    0, 0);
                            WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                            LM_INSERTITEM,
                                            (MPARAM)(pData->sSelected+1),
                                            (MPARAM)pliNew->szItemName);
                            WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                            LM_SELECTITEM,
                                            (MPARAM)(pData->sSelected+1),
                                            (MPARAM)TRUE); // select flag
                        }
                        WinPostMsg(hwndDlg, WM_UPDATE, MPNULL, MPNULL);
                    }
                    WinSetDlgItemFocus(hwndDlg, ID_XSDI_XRB_LISTBOX);
                break; }

                /*
                 * DID_OK:
                 *      store data in INI and dismiss dlg
                 */

                case DID_OK: {
                    PREBOOTWINDATA pData =
                            (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
                    PSZ     pINI, p;
                    BOOL    fValid = TRUE;

                    // _Pmpf(( "WM_COMMAND DID_OK BN_CLICKED" ));
                    // store data in INI
                    if (pData->pliReboot) {
                        pINI = malloc(
                                    sizeof(REBOOTLISTITEM)
                                  * lstCountItems((PLISTITEM)(pData->pliReboot), 0, 0)
                               );
                        memset(pINI, 0,
                                    sizeof(REBOOTLISTITEM)
                                  * lstCountItems((PLISTITEM)(pData->pliReboot), 0, 0));

                        if (pINI) {
                            PREBOOTLISTITEM pli = pData->pliReboot;
                            USHORT          usCurrent = 0;
                            p = pINI;
                            while (pli) {
                                if (    (strlen(pli->szItemName) == 0)
                                     || (strlen(pli->szCommand) == 0)
                                   )
                                {
                                    WinAlarm(HWND_DESKTOP, WA_ERROR);
                                    WinSendDlgItemMsg(hwndDlg, ID_XSDI_XRB_LISTBOX,
                                                    LM_SELECTITEM,
                                                    (MPARAM)usCurrent,
                                                    (MPARAM)TRUE);
                                    fValid = FALSE;
                                    break;
                                }
                                strcpy(p, pli->szItemName);
                                p += (strlen(p)+1);
                                strcpy(p, pli->szCommand);
                                p += (strlen(p)+1);

                                pli = pli->pNext;
                                usCurrent++;
                            }

                            PrfWriteProfileData(HINI_USER,
                                        INIAPP, INIKEY_BOOTMGR,
                                        pINI,
                                        (p - pINI + 2));

                            free (pINI);
                        }
                    } // end if (pData->pliReboot)
                    else
                        PrfWriteProfileData(HINI_USER,
                                    INIAPP, INIKEY_BOOTMGR,
                                    NULL, 0);

                    // dismiss dlg
                    if (fValid)
                        mrc = fnwpDlgGeneric(hwndDlg, msg, mp1, mp2);
                break; }

                default: { // includes DID_CANCEL
                    mrc = fnwpDlgGeneric(hwndDlg, msg, mp1, mp2);
                break; }
            }
        break; }

        case WM_ANIMATE: {
            PREBOOTWINDATA pData =
                    (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
            if (pData)
                if (pData->pAnimation) {
                    pData->pAnimation->usAniCurrent = 0;
                    WinStartTimer(WinQueryAnchorBlock(hwndDlg), hwndDlg, 1,
                            pData->pAnimation->ulDelay);
                }
        break; }

        case WM_TIMER: {
            PREBOOTWINDATA pData =
                    (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
            PANIMATION pa = NULL;
            if (pData)
                if (pData->pAnimation)
                    pa = pData->pAnimation;

            cmnAnimate(pa);
        break; }

        /*
         * WM_DESTROY:
         *      clean up allocated memory
         */

        case WM_DESTROY: {
            PREBOOTWINDATA pData = (PREBOOTWINDATA)WinQueryWindowULong(hwndDlg, QWL_USER);
            // _Pmpf(( "WM_DESTROY" ));
            cmnFreeShutdownAnimation(&(pData->pAnimation->ahptrAniIcons[0]));
            free(pData->pAnimation);
            lstClear((PLISTITEM*)&(pData->pliReboot), NULL, 0, 0);
            mrc = fnwpDlgGeneric(hwndDlg, msg, mp1, mp2);
        break; }

        default:
            mrc = fnwpDlgGeneric(hwndDlg, msg, mp1, mp2);
        break;
    }
    return (mrc);
}

/*
 * cmnQuerySystemSound:
 *      this gets a system sound from the MMPM.INI file.
 *      Parameters:
 *      usIndex: index of sound to query:
 *          MMSOUND_WARNING         0
 *          MMSOUND_INFORMATION     1
 *          MMSOUND_LOCKUP          10
 *          MMSOUND_ALARMCLOCK      11
 *          MMSOUND_PRINTERROR      12
 *          MMSOUND_ERROR           2
 *          MMSOUND_ANIMATEOPEN     3
 *          MMSOUND_ANIMATECLOSE    4
 *          MMSOUND_DRAG            5
 *          MMSOUND_DROP            6
 *          MMSOUND_SYSTEMSTARTUP   7
 *          MMSOUND_SHUTDOWN        8
 *          MMSOUND_SHREDDER        9
 *          MMSOUND_XFLD_SHUTDOWN   555
 *          MMSOUND_XFLD_RESTARTWPS 556
 *          MMSOUND_XFLD_SORT       557
 *      pszDescr: address of buffer to where to return the
 *                sound description as in the "Sound" object;
 *                may be NULL
 *      pszFile:  address of buffer to where to return the
 *                sound file; may be NULL
 *      pulVolume: address of ULONG to where to return the
 *                 volume (0-100); if the "Global volume"
 *                 flag is set, the global volume will be
 *                 returned, otherwise the individual sound
 *                 volume
 *      The string buffers are recommended to be at least
 *      CCHMAXPATH in size.
 */

BOOL cmnQuerySystemSound(USHORT usIndex, PSZ pszDescr, PSZ pszFile, PULONG pulVolume)
{
    // We inline this func because it's only used once.
    HAB     habDesktop = WinQueryAnchorBlock(HWND_DESKTOP);
    HINI    hiniMMPM;
    CHAR    szMMPM[CCHMAXPATH];
    BOOL    rc = FALSE;

    sprintf(szMMPM, "%c:\\MMOS2\\MMPM.INI", DosQueryBootDrive());

    // _Pmpf(( "cmnQuerySystemSound: %s", szMMPM ));
    hiniMMPM = PrfOpenProfile(habDesktop, szMMPM);
    if (hiniMMPM) {
        CHAR szData[1000], szData2[100];
        CHAR szKey[10];
        sprintf(szKey, "%d", usIndex);
        // _Pmpf(( "  file opened; szKey: %s", szKey ));
        PrfQueryProfileString(hiniMMPM,
                MMINIKEY_SOUNDSETTINGS, "EnableSounds",
                "TRUE",
                szData2, sizeof(szData2));
        if (strcmp(szData2, "TRUE") == 0)
            // sounds enabled at all
            if (PrfQueryProfileString(hiniMMPM,
                    MMINIKEY_SYSSOUNDS, szKey,
                    ".",
                    szData, sizeof(szData)-1) > 3)
                {
                    PSZ p1 = szData, p2;

                    // _Pmpf(( "  %s", p1 ));
                    p2 = strchr(p1, '#');
                    if (pszFile) {
                        strncpy(pszFile, p1, p2-p1);
                        pszFile[p2-p1] = '\0';
                    }
                    p1 = p2+1;

                    // _Pmpf(( "  %s", p1 ));
                    p2 = strchr(p1, '#');
                    if (pszDescr) {
                        strncpy(pszDescr, p1, p2-p1);
                        pszDescr[p2-p1] = '\0';
                    }
                    p1 = p2+1;

                    if (pulVolume) {
                        PrfQueryProfileString(hiniMMPM,
                                MMINIKEY_SOUNDSETTINGS, "ApplyVolumeToAll",
                                "FALSE",
                                szData2, sizeof(szData2));
                        if (strcmp(szData2, "FALSE") == 0) {
                            // individual volume settings per sound
                            p2 = strchr(p1, '#');
                            sscanf(p1, "%d", pulVolume);
                        } else {
                            // global volume setting for all sounds
                            PrfQueryProfileString(hiniMMPM,
                                    MMINIKEY_SOUNDSETTINGS, "Volume",
                                    "100",
                                    szData2, sizeof(szData2));
                            sscanf(szData2, "%d", pulVolume);
                        }
                    }

                    rc = TRUE;
                }

        PrfCloseProfile(hiniMMPM);
    }
    return (rc);
}

/*
 * cmnSetSystemSound:
 *      reversely, this SETS a system sound in MMPM.INI.
 */

BOOL cmnSetSystemSound(USHORT usIndex, PSZ pszDescr, PSZ pszFile, BYTE bVolume)
{
    HAB     habDesktop = WinQueryAnchorBlock(HWND_DESKTOP);
    HINI    hiniMMPM;
    CHAR    szMMPM[CCHMAXPATH];

    sprintf(szMMPM, "%c\\MMOS2\\MMPM.INI", DosQueryBootDrive());

    hiniMMPM = PrfOpenProfile(habDesktop, szMMPM);
    if (hiniMMPM) {
        CHAR szData[1000];
        CHAR szKey[10];
        sprintf(szKey, "%d", usIndex);
        sprintf(szData, "%s#%s#%d", pszFile, pszDescr, bVolume);
        PrfWriteProfileString(hiniMMPM,
                MMINIKEY_SYSSOUNDS, szKey,
                szData);
        PrfCloseProfile(hiniMMPM);
        return (TRUE);
    }
    return (FALSE);
}

/*
 * cmnQueryPendingSpoolJobs:
 *      returns the number of pending print jobs
 */

ULONG cmnQueryPendingSpoolJobs(VOID)
{
    // BOOL    rcPending = FALSE;
    ULONG   ulTotalJobCount = 0;

    SPLERR splerr;
    USHORT jobCount ;
    ULONG  cbBuf ;
    ULONG  cTotal;
    ULONG  cReturned ;
    ULONG  cbNeeded ;
    ULONG  ulLevel ;
    ULONG  i,j ;
    PSZ    pszComputerName ;
    PBYTE  pBuf = NULL;
    PPRQINFO3 prq ;
    PPRJINFO2 prj2 ;

    ulLevel = 4L;
    pszComputerName = (PSZ)NULL ;
    splerr = SplEnumQueue(pszComputerName, ulLevel, pBuf, 0L, // cbBuf
                          &cReturned, &cTotal,
                          &cbNeeded, NULL);
    if ( splerr == ERROR_MORE_DATA || splerr == NERR_BufTooSmall )
    {
        if (!DosAllocMem( (PPVOID)&pBuf, cbNeeded,
                          PAG_READ | PAG_WRITE | PAG_COMMIT) )
        {
            cbBuf = cbNeeded ;
            splerr = SplEnumQueue(pszComputerName, ulLevel, pBuf, cbBuf,
                                    &cReturned, &cTotal,
                                    &cbNeeded, NULL);
            if (splerr == NO_ERROR)
            {
                // Set pointer to point to the beginning of the buffer.
                prq = (PPRQINFO3)pBuf ;

                // cReturned has the count of the number of PRQINFO3 structures.
                for (i=0;i < cReturned ; i++)
                {
                    // Save the count of jobs. There are this many PRJINFO2
                    // structures following the PRQINFO3 structure.
                    jobCount = prq->cJobs;
                    // _Pmpf(( "Job count in this queue is %d",jobCount ));

                    // Increment the pointer past the PRQINFO3  structure.
                    prq++;

                    // Set a pointer to point to the first PRJINFO2 structure.
                    prj2=(PPRJINFO2)prq;
                    for (j=0;j < jobCount ; j++)
                    {
                        // Increment the pointer to point to the next structure.
                        prj2++;
                        // increase the job count, which we'll return
                        ulTotalJobCount++;

                    } // endfor jobCount

                    // After doing all the job structures, prj2 points to the next
                    // queue structure. Set the pointer for a PRQINFO3 structure.
                    prq = (PPRQINFO3)prj2;
                } //endfor cReturned
            } // endif NO_ERROR
            DosFreeMem(pBuf) ;
        }
    } // end if Q level given

    return (ulTotalJobCount);
}
