/*
 * common.h:
 *      header file for common.c.
 *      This prototypes functions that are common to all
 *      parts of XFolder. As opposed to helpers.c, these
 *      functions work with XFolder only, while the helper
 *      functions will work with any PM application.
 *
 *      This file also declares all kinds of structures, id's,
 *      flags, strings, commands etc. which are used by
 *      all XFolder parts, especially xfldr.c, xfdesk.c and
 *      common.c.
 *
 *      Additional declarations are passed through from the
 *      XFolder .IDL files into the xf*.h files.
 *
 *      As opposed to the declarations in dlgids.h, these
 *      declarations are NOT used by the NLS resource DLLs.
 *      The declarations have been separated to avoid
 *      unneccessary recompiles.
 *
 *      NOTES about the #include policy of XFolder:
 *      1)  This file already #include's a lot of standard
 *          include files (see below). It does NOT however
 *          #include stdlib.h and string.h, because these
 *          are already #include'd by the SOM-generated
 *          header files.
 *      2)  This file also #include's XFolder include files
 *          which are common to all XFolder parts (see below).
 *
 *      Copyright (C) 1997-98 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of the XFolder main distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

/********************************************************************
 *                                                                  *
 *   Debugging flags                                                *
 *                                                                  *
 ********************************************************************/

/*
 *  The following #define's determine whether additional debugging
 *  info will be compiled into xfldr.dll. These flags fall into two
 *  categories:
 *  1)  if you uncomment the DONTDEBUGATALL line,
 *      no debugging will be coded at all;
 *  2)  otherwise, the following lines allow
 *      debugging of certain XFolder code only.
 */

#define DONTDEBUGATALL

#ifndef DONTDEBUGATALL

    // if the following is commented out, no PMPRINTF will be
    // used at all. XFolder uses Dennis Bareis' PMPRINTF
    // package to do this.
    #define _PMPRINTF_

    // the following replaces the SOMMethodDebug macros with
    // a PMPRINTF version (makes the computer real slow)
    // #define DEBUG_SOMMETHODS

    // the following is used for some miscellaneous debugs
    #define DEBUG_PRINTF

    // the following adds debugging context menus to folders
    // #define DEBUG_CONTEXT

    // the following allows debug mode for XShutdown, which
    // will be enabled by holding down SHIFT while selecting
    // "Shutdown..." from the desktop context menu. In addition,
    // you'll get some PMPRINTF info and beeps
    #define DEBUG_SHUTDOWN

    // the following beeps when thread priorities change
    #define DEBUG_PRIORITY

    // the following printf's each added awake object
    // #define DEBUG_AWAKEOBJECTS

    // the following gives information on ordered folder content
    // #define DEBUG_ORDEREDLIST

    // the following displays internal status bar data
    #define DEBUG_STATUSBARS

    // the following will printf all kinds of settings
    // notebook information
    // #define DEBUG_SETTINGS

    // the following will printf WM_CONTROL for WPS cnrs
    // #define DEBUG_CNRCNTRL

    // the following will printf wpAddToContent
    // #define DEBUG_CNRCONTENT

    // the following will printf WM_CONTROL for WPS cnrs
    #define DEBUG_SORT

    // the following will printf language code queries etc.
    // #define DEBUG_LANGCODES

    // the following will printf folder hotkey info
    // #define DEBUG_KEYS

    // the following displays XFolder memory usage in the
    // "Object internals" of the Desktop settings; this
    // will produce additional debug code for malloc(),
    // so this better only be used for debugging
    #define DEBUG_MEMORY

    // the following beeps when the Worker thread cleans up
    // the default heap
    // #define DEBUG_MEMORYBEEP

    // the following displays a lot of infos about menu
    // processing (msgs), esp. for folder content menus
    // #define DEBUG_MENUS

    // debug startup folder processing
    // #define DEBUG_STARTUP

    // debug title clash dialog
    // #define DEBUG_TITLECLASH
#endif

/********************************************************************
 *                                                                  *
 *   #include's                                                     *
 *                                                                  *
 ********************************************************************/

// standard #include's
#include <setjmp.h>         // needed for exception handlers
#include <assert.h>         // needed for exception handlers
#include <stdio.h>          // C standard lib for printf and such

// XFolder #include's
#include "dlgids.h"
#include "helpers.h"

// the following reacts to the _PMPRINTF_ macro def'd above
#include "PMPRINTF.H"

#ifdef   SOMMethodDebug
         #undef  SOMMethodDebug
#endif

// SOMMethodDebug is the macro defined for all those
// xxxMethodDebug thingies created by the SOM compiler.
// If you have uncommended DEBUG_SOMMETHODS above, this
// will make sure that _Pmpf is used for that. In order
// for this to work, you'll also need _PMPRINTF_.
#ifdef DEBUG_SOMMETHODS
    #define  SOMMethodDebug(c,m) _Pmpf(("%s::%s", c,m))
#else
    #define  SOMMethodDebug(c,m) ;
#endif

/********************************************************************
 *                                                                  *
 *   Global #define's                                               *
 *                                                                  *
 ********************************************************************/

/* XFolder app and key names for OS2.INI */
#define INIAPP                  "XFolder"
#define INIKEY_DEFAULTTITLE     "DefaultTitle"
#define INIKEY_GLOBALSETTINGS   "GlobalSettings"
#define INIKEY_XFOLDERPATH      "XFolderPath"
#define INIKEY_ACCELERATORS     "Accelerators"
#define INIKEY_LANGUAGECODE     "Language"
#define INIKEY_JUSTINSTALLED    "JustInstalled"
#define INIKEY_DONTDOSTARTUP    "DontDoStartup"
#define INIKEY_FAVORITEFOLDERS  "FavoriteFolders"
#define INIKEY_QUICKOPENFOLDERS "QuickOpenFolders"

#define INIKEY_STARTUPWNDPOS    "StartupWndPos"
#define INIKEY_NAMECLASHWNDPOS  "NameClashWndPos"
#define INIKEY_NAMECLASHFOCUS   "NameClashLastFocus"

#define INIKEY_STATUSBARFONT    "SB_Font"
#define INIKEY_SBTEXTNONESEL    "SB_NoneSelected"
#define INIKEY_SBTEXT_WPOBJECT  "SB_WPObject"
#define INIKEY_SBTEXT_WPPROGRAM "SB_WPProgram"
#define INIKEY_SBTEXT_WPDATAFILE "SB_WPDataFile"
#define INIKEY_SBTEXT_WPDISK    "SB_WPDisk"
#define INIKEY_SBTEXT_WPFOLDER  "SB_WPFolder"
#define INIKEY_SBTEXTMULTISEL   "SB_MultiSelected"
#define INIKEY_SB_LASTCLASS     "SB_LastClass"
#define INIKEY_DLGFONT          "DialogFont"

#define INIKEY_BOOTMGR          "RebootTo"
#define INIKEY_AUTOCLOSE        "AutoClose"

#define DEFAULT_LANGUAGECODE    "001"

#define WPINIKEY_FOLDERPOS      "PM_Workplace:FolderPos"

/* keys for accessing sounds in MMPM.INI */
#define MMINIKEY_SYSSOUNDS      "MMPM2_AlarmSounds"
#define MMINIKEY_SOUNDSETTINGS  "MMPM2_AlarmSoundsData"

// system sound indexes in MMPM.INI
#define MMSOUND_WARNING         0
#define MMSOUND_INFORMATION     1
#define MMSOUND_LOCKUP          10
#define MMSOUND_ALARMCLOCK      11
#define MMSOUND_PRINTERROR      12
#define MMSOUND_ERROR           2
#define MMSOUND_ANIMATEOPEN     3
#define MMSOUND_ANIMATECLOSE    4
#define MMSOUND_DRAG            5
#define MMSOUND_DROP            6
#define MMSOUND_SYSTEMSTARTUP   7
#define MMSOUND_SHUTDOWN        8
#define MMSOUND_SHREDDER        9

// new XFolder system sounds
#define MMSOUND_XFLD_SHUTDOWN   555
#define MMSOUND_XFLD_RESTARTWPS 556
#define MMSOUND_XFLD_CTXTOPEN   558
#define MMSOUND_XFLD_CTXTSELECT 559
#define MMSOUND_XFLD_CNRDBLCLK  560

// ID's of XFolder object windows
#define ID_XFOLDEROBJECT        0x1234
#define ID_WORKEROBJECT         0x1235
#define ID_SOUNDOBJECT          0x1236

// XFolder settings page numbers; used in
// fnwpSettingsDlgCommon to tell the pages
// apart
#define SP_1GENERIC             1
#define SP_2REMOVEITEMS         2
#define SP_25ADDITEMS           3
#define SP_26CONFIGITEMS        4
#define SP_27STATUSBAR          5
#define SP_3SNAPTOGRID          6
#define SP_4ACCELERATORS        7
#define SP_5INTERNALS           8
#define SP_DTP1                 9
#define SP_DTP2                 10
#define SP_28STATUSBAR2         11
#define SP_FILEOPS              12

// the following is NOT used in fnwpSettingsDlgCommon
#define SP_FLDRSORT             20

// the following is used in fnwpConfigCommon
#define SP_SCHEDULER            1
#define SP_MEMORY               2
#define SP_HPFS                 3
#define SP_FAT                  4
#define SP_ERRORS               5
#define SP_WPS                  6

/*
 * Help panels
 *
 */

#define ID_XMH_VARIABLE         2
#define ID_XMH_CONFIGFOLDER     3
#define ID_XMH_REFRESH          4
#define ID_XMH_SNAPTOGRID       5
#define ID_XMH_COPYFILENAME     6
#define ID_XMH_XSHUTDOWN        7
#define ID_XMH_RESTARTWPS       8

// the following defines the panel ID for the
// first settings page in the System notebook;
// the other help panels follow thereafter */
// help panel IDs for XFolder context menu items

#define ID_XSH_SETTINGS1            9
#define ID_XSH_SETTINGS_DTP1        16
#define ID_XSH_SETTINGS_FLDRSORT    18
#define ID_XSH_SETTINGS_FLDR1       19
#define ID_XSH_SETTINGS_SB2         20
#define ID_XSH_SETTINGS_WPSCLASSES  21
#define ID_XSH_SETTINGS_OBJINTERNALS 22
#define ID_XSH_SETTINGS_KERNEL1     23
#define ID_XSH_SETTINGS_KERNEL2     24
#define ID_XMH_STARTUPSHUTDOWN      25
#define ID_XMH_FOLDERCONTENT        26
#define ID_XSH_SETTINGS_HPFS        27
#define ID_XSH_SETTINGS_FAT         27              // same page as HPFS
#define ID_XSH_SETTINGS_CFGM        28              // "Config folder menu items"
#define ID_XSH_SETTINGS_FILEOPS     29              // "Config folder menu items"
#define ID_XSH_SETTINGS_WPS         30
#define ID_XSH_SETTINGS_ERRORS      31

// the following defines the panel ID for the
// first dlg item (ID_XSDI_FIRST == ID_XSDI_REPLICONS);
// the other help panels follow thereafter */
#define ID_XSH_FIRST            61            // single settings dlg items

// help panel IDs for various dlg boxes
#define ID_XFH_BORED            40
#define ID_XFH_NOCONFIG         (ID_XFH_BORED+1)
#define ID_XFH_NOOBJECT         (ID_XFH_BORED+2)
#define ID_XFH_LIMITREACHED     (ID_XFH_BORED+3)
#define ID_XFH_NOTEXTCLIP       (ID_XFH_BORED+4)
#define ID_XFH_REBOOTEXT        (ID_XFH_BORED+5)
#define ID_XFH_AUTOCLOSEDETAILS (ID_XFH_BORED+6)
#define ID_XFH_SELECTCLASS      (ID_XFH_BORED+7)
#define ID_XFH_REGISTERCLASS    (ID_XFH_BORED+8)
#define ID_XFH_TITLECLASH       (ID_XFH_BORED+9)

#define ID_XFH_SELECTSOME       87

// common value for indicating that a Global Setting
// is to be used instead of an instance's one
#define SET_DEFAULT             255

// markers for XFolder variable menu items:
// these flags are used in the global linked list
// of menu items which are inserted into folder
// context menus to determine what we need to do
// if a menu item gets selected
#define OC_PROGRAM              2     // program object
#define OC_FOLDER               3     // folder: insert submenu
#define OC_ACTION               4     // not used
#define OC_TEMPLATE             5     // template: create new object
#define OC_OTHER                1     // all others: simply open

#define OC_CONTENT              7     // this item is part of the "folder content" menu
#define OC_CONTENTFOLDER        8     // the same, but it's a folder

#define OC_SEPARATOR            10    // program object, but separator only

#define CLIPBOARDKEY "%**C"        /* code in program object's parameter list for
                                      inserting clipboard data */

// status bar styles
#define SBSTYLE_WARP3RAISED     1
#define SBSTYLE_WARP3SUNKEN     2
#define SBSTYLE_WARP4RECT       3
#define SBSTYLE_WARP4MENU       4

#define SBS_STATUSBARFONT       1
#define SBS_TEXTNONESEL         2
#define SBS_TEXTMULTISEL        4

#define SBV_ICON                1
#define SBV_TREE                2
#define SBV_DETAILS             4

// max length of status bar mnemonics
#define CCHMAXMNEMONICS         256

// auto-close actions
#define ACL_CTRL_C              1
#define ACL_WMCLOSE             2
#define ACL_KILLSESSION         3
#define ACL_SKIP                4

// shutdown settings bits: composed by the
// Shutdown settings pages, stored in
// pGlobalSettings->XShutdown,
// passed to xfInitiateShutdown
#define XSD_DTM_SYSTEMSETUP     0x00001
#define XSD_DTM_SHUTDOWN        0x00002
#define XSD_DTM_LOCKUP          0x00004

#define XSD_ENABLED             0x00010
#define XSD_CONFIRM             0x00020
#define XSD_REBOOT              0x00040
#define XSD_RESTARTWPS          0x00100
#define XSD_DEBUG               0x01000
#define XSD_AUTOCLOSEVIO        0x02000
#define XSD_WPS_CLOSEWINDOWS    0x04000
#define XSD_LOG                 0x08000
#define XSD_ANIMATE             0x10000
#define XSD_APMPOWEROFF         0x20000

// folder object id's
#define XFOLDER_CONFIGID        "<XFOLDER_CONFIG>"
#define XFOLDER_STARTUPID       "<XFOLDER_STARTUP2>"
#define XFOLDER_SHUTDOWNID      "<XFOLDER_SHUTDOWN2>"

// maximum no of accelerators
#define XFLDACCSLIMIT           40

// miscellaneae
#define DEFAULTTITLELENGTH      50
#define LANGUAGECODELENGTH      30

// log file names
#define XFOLDER_CRASHLOG        "xfldtrap.log"
#define XFOLDER_SHUTDOWNLOG     "xshutdwn.log"

// Worker thread msgs (WOM_xxx)
#define WOM_ADDAWAKEOBJECT          WM_USER+100
#define WOM_REMOVEAWAKEOBJECT       WM_USER+101
#define WOM_SHOWFOLDERDATA          WM_USER+102
#define WOM_REFRESHFOLDERVIEWS      WM_USER+103

#define WOM_TREEVIEWAUTOSCROLL      WM_USER+110
#define WOM_DESKTOPREADY            WM_USER+111
#define WOM_RECREATECONFIGFOLDER    WM_USER+112
#define WOM_WELCOME                 WM_USER+113
#define WOM_DONEWITHSTARTUP         WM_USER+114

#define WOM_INVALIDATEORDEREDCONTENT WM_USER+120
#define WOM_DELETEICONPOSEA         WM_USER+121

#define WOM_PROCESSORDEREDCONTENT   WM_USER+130
#define WOM_WAITFORPROCESSNEXT      WM_USER+131
#define WOM_DESKTOPOPENED           WM_USER+132

#define WOM_UPDATEALLSTATUSBARS     WM_USER+133
#define WOM_CORRECTCNRSORT          WM_USER+134
#define WOM_QUICKOPEN               WM_USER+135

// Quick thread msgs (QM_xxx)
#define QM_PLAYSOUND                WM_USER+1004
#define QM_PLAYSYSTEMSOUND          WM_USER+1005
#define QM_BOOTUPSTATUS             WM_USER+135

// msgs to status bar window (STBM_xxx)
#define STBM_UPDATESTATUSBAR        WM_USER+1000
#define STBM_PROHIBITBROADCASTING     WM_USER+1002

// XFolder object window in Workplace thread (XOM_xxx)
#define XOM_BEGINSTARTUP            WM_USER+1100
#define XOM_ACTIVATESTATUSBAR       WM_USER+1101
#define XOM_EXCEPTIONCAUGHT         WM_USER+1102
#define XOM_LIMITREACHED            WM_USER+1103
#define XOM_POCCALLBACK             WM_USER+1104
#define XOM_QUERYXFOLDERVERSION     WM_USER+1105
#define XOM_EXTERNALSHUTDOWN        WM_USER+1106
#define XOM_BEGINQUICKOPEN          WM_USER+1107
#define XOM_NEXTQUICKOPEN           WM_USER+1108

// flags for recreating config folder (WM_RECREATECONFIGFOLDER)
#define RCF_QUERYACTION         0
#define RCF_RECREATEFOLDERONLY  1
#define RCF_RECREATEALL         2

// common dlg msgs for settings notebook dlg funcs
#define WM_SETTINGS2DLG         WM_USER+200    // set controls
#define WM_DLG2SETTINGS         WM_USER+201    // read controls
#define WM_ENABLEITEMS          WM_USER+202    // enable/disable controls
#define WM_DEFAULT              WM_USER+203    // set default values
#define WM_OPTIMIZE             WM_USER+204    // propose better values (CONFIG.SYS)

// WM_ANIMATE msg to animate an icon
// (see fnwpDlgGeneric in common.c)
#define WM_ANIMATE              WM_USER+1000

// misc
#define WM_UPDATE               WM_USER+1006 // in dlgs
#define WM_SETLONGTEXT          WM_USER+1007 // for cmnMessageBox

// here come some OS/2 default menu items,
// which are either available in Warp 4 only
// or not documented in the Toolkit */
#define ID_WPMI_INSERT                 0x2CB
#define ID_WPMI_SHOWICONVIEW           0x2CC
#define ID_WPMI_SHOWTREEVIEW           0x2CD
#define ID_WPMI_SHOWDETAILSVIEW        0x2CE

#define ID_WPMI_SORTBYNAME             0x1770
#define ID_WPMI_SORTBYEXTENSION        0x9D87
#define ID_WPMI_SORTBYTYPE             0x1771
#define ID_WPMI_SORTBYREALNAME         0x1777
#define ID_WPMI_SORTBYSIZE             0x1778
#define ID_WPMI_SORTBYWRITEDATE        0x1779
#define ID_WPMI_SORTBYACCESSDATE       0x177B
#define ID_WPMI_SORTBYCREATIONDATE     0x177D

#define ID_WPMI_ARRANGEFROMTOP         0x2DE
#define ID_WPMI_ARRANGEFROMLEFT        0x2DF
#define ID_WPMI_ARRANGEFROMRIGHT       0x2E0
#define ID_WPMI_ARRANGEFROMBOTTOM      0x2E1
#define ID_WPMI_ARRANGEPERIMETER       0x2E3
#define ID_WPMI_ARRANGEHORIZONTALLY    0x2E4
#define ID_WPMI_ARRANGEVERTICALLY      0x2E5

#define ID_WPM_LOCKINPLACE             0x2DA
#define ID_WPM_VIEW                    0x68

#define ID_WPMI_FORMATDISK             0x7C
#define ID_WPMI_CHECKDISK              0x80
#define ID_WPMI_REFRESH                0x1F7


/********************************************************************
 *                                                                  *
 *   Global structures                                              *
 *                                                                  *
 ********************************************************************/

/* ANIMATION:
   this must be passed with WM_ANIMATE to describe
   an icon animation in a dlg window */
typedef struct _ANIMATION {
    ULONG       cb;
    HWND        hwndIcon;
    USHORT      usAniCount;
    ULONG       ulDelay;
    USHORT      usAniCurrent;
    HPOINTER    ahptrAniIcons[1];
} ANIMATION, *PANIMATION;

#ifdef SOM_WPObject_h
    /* OBJECTLISTITEM:
       linked list structure for storing a list of
       objects. This is used for
       storing the currently awake WPS objects
       on the system (in the XFolder Worker thread) */
    typedef struct _OBJECTLISTITEM {
        struct _OBJECTLISTITEM
                                *pNext, *pPrevious;
        ULONG                   ulSize;
        WPObject                *pObj;
    } OBJECTLISTITEM, *POBJECTLISTITEM;
#endif

/* GlobalSettings structure for use in memory
   and OS2.INI; we use ULONGs for performance
   reasons. Each setting in here corresponds to
   a setting in the "System" objects and Desktop
   settings notebook.
   Use cmnQueryGlobalSettings() to get a pointer
   to the global settings. */
typedef struct GLOBALSETTINGS_STRUCT
{
    ULONG       ReplIcons,
                Cascade,
                FullPath,
                KeepTitle,
                RemoveX,
                AppdParam,
                Move4Refresh;

    ULONG       MaxPathChars,
                DefaultMenuItems;

    LONG        VarMenuOffset;

    ULONG       SnapToGrid;

    LONG        GridX,
                GridY,
                GridCX,
                GridCY;

    ULONG       Accelerators,
                OpenSettings;
/* XFolder 0.52 */
    ULONG       RemoveLockInPlace,
                RemoveFormatDisk,
                RemoveCheckDisk,
                RemoveView,
                RemoveInsert,

                DebugMode,
                CopyFilename;
    ULONG       XShutdown;
    ULONG       NoWorkerThread,
                NoSubclassing,
                TreeViewAutoScroll,
                ShowStartupProgress;
    ULONG       StartupDelay;

/* XFolder 0.70 */
    ULONG       FolderContent,
                FCShowIcons,
                StatusBar,
                SBStyle;                    // status bar style
    LONG        lSBBgndColor, lSBTextColor; // status bar colors
    ULONG       TplPosition;
    USHORT      LastRebootExt;
    ULONG       SelectSome,
                ReplaceSort,
                AlwaysSort,
                DefaultSort,
                CleanupINIs,

/* XFolder 0.80  */
                ShowBootupStatus,
                WpsShowClassInfo;
    ULONG       SBForViews,                 // SBV_xxx flags
                ReplConfirms,
                ShowXFolderAnim,
                ExtendSettings,
                ShowInternals;

} GLOBALSETTINGS, *PGLOBALSETTINGS;

/* NLSSTRINGS:
   structure which is filled upon WPS IPL with
   all language-dependent XFolder strings. Use
   cmnQueryNLSStrings() to get access to this
   structure. */
typedef struct NLSSTRINGS_STRUCT {
    // strings from resources
    PSZ     pszNotDefined,
            pszProductInfo,
            pszRefreshNow,
            pszSnapToGrid,

            pszFldrContent,
            pszCopyFilename,
            pszBored,
            pszFldrEmpty,

            pszSelectSome,
            pszQuickStatus,

            pszCtrl,
            pszAlt,
            pszShift,

            pszBackspace,
            pszTab,
            pszBacktab,
            pszEnter,
            pszEsc,
            pszSpace,
            pszPageup,
            pszPagedown,
            pszEnd,
            pszHome,
            pszLeft,
            pszUp,
            pszRight,
            pszDown,
            pszPrintscrn,
            pszInsert,
            pszDelete,
            pszScrlLock,
            pszNumLock,

            pszWinLeft,
            pszWinRight,
            pszWinMenu,

            pszSDFlushing,
            pszSDCAD,
            pszSDRebooting,
            pszSDRestartingWPS,
            pszSDClosing,
            pszShutdown,
            pszRestartWPS,
            pszSDSavingDesktop,
            pszSDSavingProfiles,

            pszStarting,

            pszPopulating,

            psz1Generic,
            psz2RemoveItems,
            psz25AddItems,
            psz26ConfigFolderMenus,
            psz27StatusBar,
            psz3SnapToGrid,
            psz4Accelerators,
            psz5Internals,
            pszFileOps,

            pszSortByName,
            pszSortByType,
            pszSortByClass,
            pszSortByRealName,
            pszSortBySize,
            pszSortByWriteDate,
            pszSortByAccessDate,
            pszSortByCreationDate,
            pszSortByExt,
            pszSortFoldersFirst,

            pszSort,
            pszAlwaysSort,

            pszInternals,

            pszSBClassMnemonics,
            pszSBClassNotSupported,

            pszWpsClasses,
            pszWpsClassLoaded,
            pszWpsClassLoadingFailed,
            pszWpsClassReplacedBy,
            pszWpsClassOrphans,
            pszWpsClassOrphansInfo,

            pszScheduler,
            pszMemory,
            pszErrors,
            pszWPS,

            pszProcessContent,

            pszSettings,
            pszSettingsNotebook,
            pszAttributes,
            pszAttrArchived,
            pszAttrSystem,
            pszAttrHidden,
            pszAttrReadOnly,

            pszFldrSettings,
            pszSmallIcons,
            pszFlowed,
            pszNonFlowed,
            pszNoGrid,

            pszShowStatusBar;

} NLSSTRINGS, *PNLSSTRINGS;

// shutdown structure for external calls
typedef struct _SHUTDOWNPARAMS {
    BOOL        optReboot, optConfirm, optDebug, optRestartWPS,
                optWPSCloseWindows, optAutoCloseVIO, optLog,
                optAnimate;
    CHAR        szRebootCommand[CCHMAXPATH];
} SHUTDOWNPARAMS, *PSHUTDOWNPARAMS;

/********************************************************************
 *                                                                  *
 *   Function prototypes                                            *
 *                                                                  *
 ********************************************************************/

/*
 *  The following functions are actually present in
 *  common.c:
 */

ULONG cmnQueryXFolderBasePath(PSZ pszPath, ULONG ulBufsize);
void cmnLoadString(HAB habDesktop, HMODULE hmodResource, ULONG ulID, PSZ *ppsz);
HMODULE cmnQueryNLSModuleHandle(BOOL fEnforceReload);
PNLSSTRINGS cmnQueryNLSStrings(VOID);
PVOID cmnLoadGlobalSettings(VOID);
PGLOBALSETTINGS cmnQueryGlobalSettings(VOID);
BOOL cmnStoreGlobalSettings(VOID);
BOOL cmnSetDefaultSettings(USHORT usSettingsPage);
PSZ cmnQueryHelpLibrary(VOID);
BOOL cmnDisplayFocusHelp(ULONG ulPanelIfNotFound);
PSZ cmnQueryMessageFile(VOID);
PSZ cmnQueryLanguageCode(VOID);
BOOL cmnSetLanguageCode(PSZ pszLanguage);

PSZ cmnQueryStatusBarSetting(USHORT usSetting);
BOOL cmnSetStatusBarSetting(USHORT usSetting, PSZ pszSetting);
ULONG cmnQueryStatusBarHeight(VOID);
void cmnOpenDebugFile(VOID);
void cmnWriteAndCloseDebugFile(PSZ pszBuffer);
void cmnDescribeMutexSem(PSZ pszBuffer, HMTX hmtx);
APIRET cmnError(PSZ psz, APIRET rc);
ULONG cmnMessageBox(HWND hwndOwner, PSZ pszTitle, PSZ pszMessage, ULONG flStyle);
APIRET cmnGetMessage(PCHAR *pTable, ULONG ulTable, PSZ pszBuf, ULONG cbBuf,
                        ULONG ulMsgNumber);
ULONG cmnMessageBoxMsg(HWND hwndOwner, ULONG ulTitle, ULONG ulMessage, ULONG flStyle);
ULONG cmnMessageBoxMsgExt(HWND hwndOwner, ULONG ulTitle,
        PCHAR *pTable, ULONG ulTable, ULONG ulMessage, ULONG flStyle);
VOID cmnSetHelpPanel(ULONG ulHelpPanel);
VOID cmnAnimate(PANIMATION pa);

MRESULT EXPENTRY fnwpDlgGeneric(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpRebootExt(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);

BOOL cmnQuerySystemSound(USHORT usIndex, PSZ pszDescr, PSZ pszFile, PULONG pulVolume);
BOOL cmnSetSystemSound(USHORT usIndex, PSZ pszDescr, PSZ pszFile, BYTE bVolume);

#ifdef SOM_XFolder_h
    BOOL cmnCheckIfPopulated(XFolder *somSelf);
#endif

/*
 *  The following dialog functions are not present in
 *  common.c, but in xfsys.h. We prototype them here
 *  anyway, because this include file is #include'd
 *  by all XFolder modules.
 */

// fwd dcl: dlg procs for settings notebook pages (xfsys.c);
// some of these need to be fwd-declared for xfdesk.c
MRESULT EXPENTRY fnwpSettingsDlg_Desktop1(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsDlg_Desktop2(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);

MRESULT EXPENTRY fnwpSettingsXFolder1(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsXFolderInternals(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsRemoveMenus(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsAddMenus(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsConfigFolderMenus(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsStatusBars1(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsStatusBars2(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsSnapToGrid(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsAccels(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsFileOperations(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);

MRESULT EXPENTRY fnwpConfigScheduler(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpConfigMemory(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpConfigHPFS(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpConfigFAT(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);

MRESULT EXPENTRY fnwpConfigErrors(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpConfigWPS(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);

MRESULT EXPENTRY fnwpSettingsDlg_FldrSort(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSettingsObjDetails(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);

/*
 *  Here come miscellaneous other dialog funcs.
 */

// fwd dcl: wnd procs for subclassed windows (xfldr.c)
MRESULT EXPENTRY fnwpFolderContentMenu(HWND hwndMenu, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpSubclassedFolderFrame(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT EXPENTRY fnwpStatusBar(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);

MRESULT EXPENTRY fnwpSelectSome(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);

// funcs for loading the traffic light animation
#define ANIMAX 8
VOID cmnLoadShutdownAnimation(HPOINTER ahptrAniIcons[ANIMAX], HMODULE hmod);
VOID cmnFreeShutdownAnimation(HPOINTER ahptrAniIcons[ANIMAX]);

// get ulWorkplaceFunc2 (xfldr.c)
PULONG cmnGetWorkplaceFunc2(VOID);

#ifdef SOM_XFolder_h
    // populate a given folder and all of its subfolders
    BOOL cmnPopulateTree(XFolder *somSelf);
#endif

// thread functions (xfdesk.c)
VOID cmnInitializeXFolder(VOID);
BOOL cmnPostWorkerMsg(ULONG msg, MPARAM mp1, MPARAM mp2);
BOOL cmnPostQuickMsg(ULONG msg, MPARAM mp1, MPARAM mp2);
BOOL cmnPlaySystemSound(USHORT usindex);

// post/send msg to XFolder Object window on Workplace thread (xfobj.c)
BOOL cmnPostObjectMsg(ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT cmnSendObjectMsg(ULONG msg, MPARAM mp1, MPARAM mp2);

#ifdef SOM_XFldObject_h
    // copy object filename(s) to clipboard (xfobj.c)
    BOOL cmnCopyObjectFileName(WPObject *somSelf, HWND hwndFrame, BOOL fFullPath);
#endif

// return no. of pending spool (print) jobs (xfdesk.c)
ULONG cmnQueryPendingSpoolJobs(VOID);

// callback function in xfldr.c for Worker thread (in xfdesk.c)
// to update all status bars
MRESULT EXPENTRY fncbUpdateStatusBars(HWND hwndView, ULONG ulActivate,
                MPARAM mpView, MPARAM mpFolder);

// dlg func for "Auto-Close Details" in xfdesk.c
MRESULT EXPENTRY fnwpAutoCloseDetails(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);

// exception handlers for XFolder threads

// replacement EXCEPTIONREGISTRATIONRECORD struct
// for thread exception handling
typedef struct REGREC2_STRUCT {
  PVOID     pNext;
  PFN       pfnHandler;
  jmp_buf   jmpThread;
} REGREC2, *PREGREC2;

// simple exception handler: writes error log and terminates
// PMSHELL.EXE
ULONG _System ExcptHandler(PEXCEPTIONREPORTRECORD         p1,
                                    PEXCEPTIONREGISTRATIONRECORD    p2,
                                    PCONTEXTRECORD                  p3,
                                    PVOID                           pv);

// sophisticated excpt handler: writes error log and jumps
// to a predefined location in the thread, so that it keeps
// running
ULONG _System ExcptHandlerPlus(PEXCEPTIONREPORTRECORD pReportRec,
                                     PREGREC2 pRegRec2,
                                     PCONTEXTRECORD pContextRec,
                                     PVOID pv);

// "quiet" excpt handler: simply suppresses all xcpt's and do
// a longjmp, but no error log
ULONG _System ExcptHandlerQuiet(PEXCEPTIONREPORTRECORD pReportRec,
                                     PREGREC2 pRegRec2,
                                     PCONTEXTRECORD pContextRec,
                                     PVOID pv);

// loading / storing accels (xfsys.c)
void cmnLoadAccelerators(VOID);
void cmnStoreAccelerators(VOID);
void cmnLoadDefaultAccelerators(VOID);

#ifdef SOM_XFolder_h

    /*
     * cmnModifyPopupMenu:
     *      this routine gets called by the wpModifyPopupMenu methods of
     *      both XFolder and XFldDisk; first it removes and inserts menu
     *      items according to the Global / folder settings and then starts
     *      working on folder content and Config Folders
     */

    BOOL cmnModifyPopupMenu(XFolder *somSelf, HWND hwndMenu, HWND hwndCnr, ULONG iPosition);

    /*
     * cmnMenuItemSelected:
     *           checks for whether one of XFolder's menu items
     *           was selected; if so, executes corresponding action
     *           and returns TRUE, otherwise does nothing and
     *           returns FALSE, upon which the calling method
     *           (which should be wpMenuItemSelected) should
     *           call its parent method to process the menu item
     */

    BOOL cmnMenuItemSelected(XFolder *somSelf, HWND hwndFrame, ULONG ulMenuId);

    /*
     * cmnMenuItemHelpSelected:
     *           display help for a context menu item; this routine is
     *           shared by all XFolder classes also, so you'll find
     *           XFldDesktop items in here too
     */

    BOOL cmnMenuItemHelpSelected(XFolder *somSelf, ULONG MenuId);

    /*
     * cmnSetCnrSort:
     *      this resets the cnr of an open folder view to the
     *      settings that XFolder needs to implement proper
     *      extended sorting; the WPS keeps messing with the
     *      container settings, so we need to override this.
     *      Resides in xfldr.c.
     */

    VOID cmnSetCnrSort(XFolder *somSelf, HWND hwndCnr);

    /*
     * cmnXomActivateStatusBar:
     *      this func is called from the XFolder Object window
     *      proc only when XOM_ACTIVATESTATUSBAR is received.
     *      Unfortunately, this extra message is neccessary
     *      because PM will hang if we reformat folder windows
     *      from the worker thread.
     *      Handling this msg has been moved to xfldr.c
     *      although the object window handling is in xfobj.c,
     *      because the neccessary structs are visible here only.
     *      Parameters:
     *          ULONG mp1   0: hide status bar
     *                      1: show status bar
     *                      2: update (reformat) status bar
     *          HWND  mp2:  hwndView (frame) to update
     */

    VOID cmnXomActivateStatusBar(MPARAM mp1, MPARAM mp2);

    /*
     * cmnQueryFrameFromView:
     *           this routine gets the frame window handle of the
     *           specified object view; returns 0 if the specified view is
     *           not currently open.
     */

    HWND cmnQueryFrameFromView(XFolder *somSelf, ULONG ulView);

    /*
     * cmnSubclassFrameWnd:
     *      this routine is used by the wpOpen methods of both
     *      XFolder and XFldDisk to subclass a new frame window
     *      to allow for extra XFolder features.
     *      For this, this routines maintains a global linked list of
     *      subclassed windows, whose first item is stored in psliSubclassed.
     *      Every list item contains (among other things) the original
     *      wnd proc of the subclassed window, which the subclassed wnd
     *      proc (fnwpSubclassedFolderFrame) needs to call the the original folder
     *      wnd proc for a given frame window, because these procs might differ
     *      depending on the class or view type or installed WPS enhancers.
     *      These list items are also used for storing status bar data.
     *      We cannot store this data elsewhere, because QWL_USER in the
     *      wnd data seems to be used by the WPS already.
     *      This func returns the newly created PSUBCLASSEDLISTITEM.
     */

    PSUBCLASSEDLISTITEM cmnSubclassFrameWnd(HWND hwndFrame, XFolder *somSelf, ULONG ulView);
#endif

/*
 * cmnQueryAwakeObjectsCount:
 *      returns the current number of awake WPS objects
 *      (maintained by the Worker thread)
 */

LONG cmnQueryAwakeObjectsCount(VOID);

#ifdef SOM_WPObject_h
    /*
     * cmnCheckObject:
     *      since somIsObj doesn't seem to be working right,
     *      here is a new function which checks if pObject
     *      points to a valid WPS object. This is done by
     *      temporarily installing yet another xcpt handler, so if the
     *      object ain't valid, calling this function doesn't
     *      crash, but returns FALSE only.
     *      Resides in xfdesk.c.
     */

    BOOL cmnCheckObject(WPObject *pObject);
#endif

/*
 * cmnProcessAccel (xfsys.c):
 *      this is called by the subclassed folder frame wnd proc to check
 *      for whether a given WM_CHAR message matches one of the folder
 *      hotkeys; we need to put this in this file because we cannot
 *      access the global XFldAccs array from xfldr.c
 */

BOOL cmnProcessAccel(HWND hwndFrame, MPARAM mp1, MPARAM mp2);

/*
 *  A few MACROS are also declared here:
 */

/*
 * NLS_MODULE:
 *      this macro is just a shortcut to querying the
 *      language-dependent XFolder DLL handle
 */

#define NLS_MODULE cmnQueryNLSModuleHandle(FALSE)

/*
 * DEFAULT_SORT:
 *      returns a useable setting for the default sort criterion
 *      according to instance / global settings; before using this
 *      macro, you need to initialize the following:
 *          PGLOBALSETTINGS     pGlobalSettings
 *          XFolderData         *somThis
 */

#define DEFAULT_SORT ((_DefaultSort == SET_DEFAULT) ? pGlobalSettings->DefaultSort : _DefaultSort)

/*
 * ALWAYS_SORT:
 *      the same for AlwaysSort
 */

#define ALWAYS_SORT ((_AlwaysSort == SET_DEFAULT) ? pGlobalSettings->AlwaysSort : _AlwaysSort)

/*
 * cmnInitiateShutdownExt (xfdesk.c):
 *      just like the XFldDesktop method, but this one
 *      allows setting all the shutdown parameters by
 *      using the SHUTDOWNPARAMS structure. This is used
 *      for calling XShutdown externally, which is done
 *      by sending XOM_EXTERNALSHUTDOWN to the XFolder
 *      object window (see xfobj.c).
 */

BOOL cmnInitiateShutdownExt(PSHUTDOWNPARAMS psdp);


