/*
 * helpers.c:
 *      helpers.c contains helper functions that are independent
 *      of a single application, i.e. you can use them in any
 *      program. Parts of these functions require PM, others
 *      the OS/2 kernel, others are just C helper routines.
 *
 *      These funcs are forward-declared in helpers.h, which
 *      must be #include'd first.
 *
 *      Copyright (C) 1997-98 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of the XFolder main distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#define INCL_DOSDEVIOCTL
#define INCL_DOS
#define INCL_DOSERRORS
#define INCL_WIN
#define INCL_GPI
#include <os2.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "helpers.h"

#include "pmprintf.h"

/********************************************************************
 *                                                                  *
 *   Functions for maintaining linked pointer lists                 *
 *                                                                  *
 ********************************************************************/

/* The following functions are used commonly to maintain linked
   pointer lists: we have functions for appending, counting and
   removing list items, plus one for deleting a whole list.
   All these function names are introduced by "lst".

   Note that all these functions assume the passed pointers to
   be pointers to LISTITEM structures (helper.h). You may, of course,
   use different structures, but the first items in your structure
   must be *pNext, *pPrevious, ulSize like in LISTITEM to make these
   functions work. You will then have to cast your structure pointers
   manually to PLISTITEM or PLISTITEM* before using these functions.

   All of the following functions work in two modes:
   -- "one-way" lists, i.e. list which can only be accessed through
      the first item on the list: this mode applies when *ppList
      is specified to be NULL;
   -- "two-way" lists, i.e. list which can be accessed both through
      the first item and the last item on the list; in this case,
      you must also specify *ppLast, the pointer of the last item,
      which will also be maintained by these functions. "Two-way"
      lists are much faster when appending many items, because we
      don't have to go through the whole list to find the last item.

   All of the following functions support mutex semaphores to make
   them thread-safe. You can pass an existing mutex semaphore (of
   which you created one for each maintained list) in hmtx and
   specify the timeout in ulTimeout; if your list will only be used
   by one thread, simply pass 0, 0, and no semaphores will be used.  */

/*
 * lstAppendItem:
 *      this will create an item for a linked list starting at
 *      *ppFirst; this must be a pointer to a pointer to the
 *      first list item; if *ppFirst is NULL (= list is empty),
 *      *ppFirst will be adjusted so that it points to the new
 *      item. If ppLast == NULL, the list is assumed to be
 *      a "one-way" list, otherwise it must point to a pointer
 *      for the last list item, which will be adjusted.
 *      This function returns the address of the new list item.
 */

PLISTITEM lstAppendItem(PLISTITEM *ppFirst, // first list item
            PLISTITEM *ppLast,              // last list item (or NULL)
            PLISTITEM pNewItem,             // item to insert
            HMTX hmtx,                      // mutex semaphore for this list
            ULONG ulTimeout)                // mutex timeout in ms
{
    if (pNewItem)
    {
        if (hmtx)
            if (DosRequestMutexSem(hmtx, ulTimeout) != NO_ERROR)
                return (NULL);

        pNewItem->pNext = NULL;

        if (*ppFirst == NULL) {
            // list is empty: set first pointer to new list item
            *ppFirst = (PVOID)pNewItem;
            pNewItem->pPrevious = NULL;
        } else {
            BOOL    Search = TRUE;
            // list is not empty: find end of list and append the
            // new item

            if (ppLast)
                if (*ppLast) {
                    // last item given: use this
                    (*ppLast)->pNext = pNewItem;
                    pNewItem->pPrevious = *ppLast;
                    Search = FALSE;
                }

            if (Search)
            {
                // last item was not given: find last item
                PLISTITEM pItem = (PLISTITEM)*ppFirst;
                while (pItem->pNext)
                    pItem = pItem->pNext;
                pItem->pNext = pNewItem;
                pNewItem->pPrevious = pItem;
            }
        }
        if (ppLast)
            *ppLast = pNewItem;
    }

    if (hmtx)
        DosReleaseMutexSem(hmtx);

    return (pNewItem);
}

/*
 * lstInsertItemAt:
 *      does just the same, but does not append the new item
 *      at the end of the list, but BEFORE the specified index
 *      lIndex. If lIndex is larger than the number of items
 *      on the list or == -1, the new item will be appended.
 */

PLISTITEM lstInsertItemAt(PLISTITEM *ppFirst, PLISTITEM *ppLast,
            PLISTITEM pNewItem, LONG lIndex, HMTX hmtx, ULONG ulTimeout)
{
    if (pNewItem) {
        if (lIndex == -1)
            lstAppendItem(ppFirst, ppLast, pNewItem, hmtx, ulTimeout);
        else {
            if (hmtx)
                if (DosRequestMutexSem(hmtx, ulTimeout) != NO_ERROR)
                    return (NULL);

            if (lIndex == 0) {
                // insert at beginning:
                pNewItem->pNext = (*ppFirst);
                pNewItem->pPrevious = NULL;
                (*ppFirst)->pPrevious = pNewItem;
                *ppFirst = pNewItem;
            } else {
                // insert at a later position:
                PLISTITEM pItemInsertAfter = lstItemFromIndex(
                                *ppFirst, (lIndex-1),
                                0, 0); // we already have the semaphore

                if (pItemInsertAfter) {
                    pNewItem->pNext = pItemInsertAfter->pNext;
                    pItemInsertAfter->pNext = pNewItem;
                    pNewItem->pPrevious = pItemInsertAfter;
                    if (pNewItem->pNext)
                        pNewItem->pNext->pPrevious = pNewItem;
                } else {
                    // item index too large: append instead
                    lstAppendItem(ppFirst, ppLast, pNewItem,
                            0, 0); // we already have the semaphore
                }
            }

            if (hmtx)
                DosReleaseMutexSem(hmtx);
        }
        return (pNewItem);
    } else
        return (NULL);
}

/*
 * lstRemoveItem:
 *      reversely, this will remove a given pRemoveItem from
 *      the linked list beginning at *ppFirst. Returns TRUE if
 *      successful, FALSE if pRemoveItem was not found.
 */

BOOL lstRemoveItem(PLISTITEM *ppFirst, PLISTITEM *ppLast, PLISTITEM pRemoveItem, HMTX hmtx, ULONG ulTimeout)
{
    BOOL        Found = FALSE;

    if (hmtx)
        if (DosRequestMutexSem(hmtx, ulTimeout) != NO_ERROR)
            return (FALSE);

    if (ppFirst) {
        if ((*ppFirst) && (pRemoveItem))
        {
            if (*ppFirst == pRemoveItem) {
                // pRemoveItem is the first item on the list:
                // set *ppFirst to the second item on the list.
                // This also works if pRemoveItem is the only item */
                *ppFirst = pRemoveItem->pNext;
                if (*ppFirst)
                    (*ppFirst)->pPrevious = NULL;
                Found = TRUE;
            }
            else {
                // otherwise begin searching list:
                PLISTITEM pLastItem = *ppFirst;
                PLISTITEM pItem = pLastItem->pNext;

                while (pItem)
                {
                    if (pItem == pRemoveItem) {
                        // item found: set pNext of previous item to next item
                        // (i.e. leave out pRemoveItem) */
                        pLastItem->pNext = pRemoveItem->pNext;
                        if (pRemoveItem->pNext)
                            pRemoveItem->pNext->pPrevious = pLastItem;
                        Found = TRUE;
                        break;
                    }
                    else {
                        pLastItem = pItem;
                        pItem = pItem->pNext;
                    }
                }
            }
        }
        if (Found) {
            if (ppLast)
                /* if (*ppLast == pRemoveItem)
                    *ppLast = pRemoveItem->pPrevious;*/
                *ppLast = NULL;

            free(pRemoveItem);
        }
    }

    if (hmtx)
        DosReleaseMutexSem(hmtx);

    return (Found);
}

/*
 * lstSwapItems:
 *      this will swap the items ppItem1 and ppItem2 in the linked
 *      list starting at *ppFirst; useful for sorting.
 *      Note that this function will not only adjust the pNext
 *      and pPrevious pointers of the LISTITEMs, but also
 *      swap the *ppItem1 and *ppItem2 pointers.
 */

BOOL lstSwapItems(PLISTITEM *ppFirst, PLISTITEM *ppLast,
            PLISTITEM *ppItem1, PLISTITEM *ppItem2,
            HMTX hmtx, ULONG ulTimeout)
{
    BOOL brc = FALSE;
    // PLISTITEM   pPrevious = NULL;

    if (hmtx)
        if (DosRequestMutexSem(hmtx, ulTimeout) != NO_ERROR)
            return (FALSE);

    if ( (ppFirst) && (ppItem1) && (ppItem2) )
        if ((*ppFirst) && ((*ppItem1)) && ((*ppItem2)))
            if (*ppItem1 != *ppItem2)
            {
                PLISTITEM pBefore1 = (*ppItem1)->pPrevious,
                          pAfter1 = (*ppItem1)->pNext,
                          pBefore2 = (*ppItem2)->pPrevious,
                          pAfter2 = (*ppItem2)->pNext,
                          pItemTemp;

                if ((*ppItem1)->pNext == (*ppItem2)) {
                    (*ppItem1)->pNext = pAfter2;
                    if (pAfter2)
                        pAfter2->pPrevious = (*ppItem1);

                    (*ppItem2)->pNext = (*ppItem1);
                    (*ppItem1)->pPrevious = (*ppItem2);

                    if (pBefore1) {
                        pBefore1->pNext = (*ppItem2);
                        (*ppItem2)->pPrevious = pBefore1;
                    } else {
                        *ppFirst = (*ppItem2);
                        (*ppItem2)->pPrevious = NULL;
                    }
                } else {
                    if (pBefore1) {
                        pBefore1->pNext = (*ppItem2);
                        (*ppItem2)->pPrevious = pBefore1;
                    } else {
                        *ppFirst = (*ppItem2);
                        (*ppItem2)->pPrevious = NULL;
                    }
                    (*ppItem2)->pNext = pAfter1;
                    if (pAfter1)
                        pAfter1->pPrevious = (*ppItem2);

                    if (pBefore2) {
                        pBefore2->pNext = (*ppItem1);
                        (*ppItem1)->pPrevious = pBefore2;
                    } else {
                        *ppFirst = (*ppItem1);
                        (*ppItem1)->pPrevious = NULL;
                    }
                    (*ppItem1)->pNext = pAfter2;
                    if (pAfter2)
                        pAfter2->pPrevious = (*ppItem1);
                }

                // now swap the pointers also
                pItemTemp = *ppItem1;
                *ppItem1 = *ppItem2;
                *ppItem2 = pItemTemp;

                if (ppLast)
                    *ppLast = NULL;
            }

    if (hmtx)
        DosReleaseMutexSem(hmtx);

    return (brc);
}

/*
 * lstQuickSort2:
 *      helper routine for recursive quicksort (below)
 */

VOID lstQuickSort2(PLISTITEM *ppFirst, PLISTITEM *ppLast, PFNSORTLIST pfnSort, PVOID pStorage,
            LONG lLeft, LONG lRight)
{
    LONG ll = lLeft, lr = lRight-1,
         lPivot = lRight;

    if (lRight > lLeft) {
        PLISTITEM   pItemLeft = lstItemFromIndex(*ppFirst, ll, 0, 0),
                    pItemRight = lstItemFromIndex(*ppFirst, lr, 0, 0),
                    pItemPivot = lstItemFromIndex(*ppFirst, lPivot, 0, 0);

        while (TRUE) {
            // compare left item to pivot
            while ( (*pfnSort)(pItemLeft, pItemPivot, pStorage) < 0 )
            {
                ll++;
                pItemLeft = pItemLeft->pNext;
            }

            // compare right item to pivot
            while (     ( (*pfnSort)(pItemRight, pItemPivot, pStorage) >= 0 )
                    && (lr > ll)
                  )
            {
                lr--;
                pItemRight = pItemRight->pPrevious;
            }

            if (lr <= ll)
                break;

            // swap pItemLeft and pItemRight
            lstSwapItems(ppFirst, ppLast, &pItemLeft, &pItemRight, 0, 0);
        }

        // swap pItemLeft and pItemPivot
        lstSwapItems(ppFirst, ppLast, &pItemLeft, &pItemPivot, 0, 0);

        lstQuickSort2(ppFirst, ppLast, pfnSort, pStorage, lLeft, ll-1);
        lstQuickSort2(ppFirst, ppLast, pfnSort, pStorage, ll+1, lRight);
    }
}

/*
 * lstQuickSort:
 *      this will sort the list starting at *ppFirst using the
 *      sort function pfnSort, which works similar to those of the
 *      container sorts, i.e. it must be declared as a
 *          SHORT EXPENTRY fnSort(PVOID pItem1, PVOID pItem1, PVOID pStorage)
 *      These sort functions need to return the following:
 *          0   pItem1 == pItem2
 *         -1   pItem1 <  pItem2
 *         +1   pItem1 >  pItem2
 *      This function uses the "quick sort" algorithm, which is one of
 *      the fastest sort algorithms available. However, this is an
 *      "instable" sort algorithm, meaning that list items considered
 *      equal by pfnSort do _not_ retain their position, but might be
 *      changed. If you need these to remain where they are, use
 *      lstBubbleSort.
 */

BOOL lstQuickSort(PLISTITEM *ppFirst, PLISTITEM *ppLast,
            PFNSORTLIST pfnSort, PVOID pStorage,
            HMTX hmtx, ULONG ulTimeout)
{
    if (hmtx)
        if (DosRequestMutexSem(hmtx, ulTimeout) != NO_ERROR)
            return (FALSE);

    if ((pfnSort) && (ppFirst))
    {
        LONG lRight = lstCountItems(*ppFirst, 0, 0)-1;

        lstQuickSort2(ppFirst, ppLast, pfnSort, pStorage,
                    0,          // lLeft
                    lRight);
    }

    if (hmtx)
        DosReleaseMutexSem(hmtx);

    return (TRUE);
}

/*
 * lstBubbleSort:
 *      this will sort the list starting at *ppFirst using the
 *      sort function pfnSort, which works similar to those of the
 *      container sorts, i.e. it must be declared as a
 *          SHORT EXPENTRY fnSort(PVOID pItem1, PVOID pItem1, PVOID pStorage)
 *      These sort functions need to return the following:
 *          0   pItem1 == pItem2
 *         -1   pItem1 <  pItem2
 *         +1   pItem1 >  pItem2
 *      This function uses the "bubble sort" algorithm, which will cause a
 *      lot of calls to pfnSort and which is really ineffective for lists
 *      with more than 100 items.
 */

BOOL lstBubbleSort(PLISTITEM *ppFirst, PLISTITEM *ppLast,
            PFNSORTLIST pfnSort, PVOID pStorage,
            HMTX hmtx, ULONG ulTimeout)
{
    if (hmtx)
        if (DosRequestMutexSem(hmtx, ulTimeout) != NO_ERROR)
            return (FALSE);

    if ((pfnSort) && (ppFirst))
    {
        LONG lRight = lstCountItems(*ppFirst, 0, 0),
             lSorted, x;

        do {
            lRight--;
            lSorted = 0;
            for (x = 0; x < lRight; x++) {
                // ulComp++;
                PLISTITEM pItem1 = lstItemFromIndex(*ppFirst, x, 0, 0),
                          pItem2 = pItem1->pNext;
                if ((pItem1) && (pItem2))
                    if ( (*pfnSort)(pItem1, pItem2, pStorage) > 0 ) {
                        lstSwapItems(ppFirst, ppLast, &pItem1, &pItem2, 0, 0);
                        lSorted++;
                    }
            }
        } while ( lSorted && (lRight > 1) );
    }

    if (hmtx)
        DosReleaseMutexSem(hmtx);

    return (TRUE);
}

/*
 * lstCountItems:
 *      this will count the number of items from pFirst;
 *      returns 0 if error occured.
 */

ULONG lstCountItems(PLISTITEM pFirst, HMTX hmtx, ULONG ulTimeout)
{
    ULONG ulCount = 0;
    PLISTITEM pItem = pFirst;

    if (hmtx)
        if (DosRequestMutexSem(hmtx, ulTimeout) != NO_ERROR)
            return (0);

    while (pItem)
    {
        ulCount++;
        pItem = pItem->pNext;
    }

    if (hmtx)
        DosReleaseMutexSem(hmtx);

    return (ulCount);
}

/*
 * lstItemFromIndex:
 *      returns the list item with a given ulIndex
 *      (counting from 0), or NULL if ulIndex is too large
 */

PLISTITEM lstItemFromIndex(PLISTITEM pFirst, ULONG ulIndex, HMTX hmtx, ULONG ulTimeout)
{
    ULONG           ulCount = 0;
    PLISTITEM pItem = NULL;

    if (hmtx)
        if (DosRequestMutexSem(hmtx, ulTimeout) != NO_ERROR)
            return (NULL);

    if (pFirst) {
        pItem = pFirst;
        for (ulCount = 0; ((pItem) && (ulCount < ulIndex)); ulCount++)
            if (pItem->pNext)
                pItem = pItem->pNext;
            else
                pItem = NULL;
    }

    if (hmtx)
        DosReleaseMutexSem(hmtx);

    return (pItem);
}

/*
 * lstClear:
 *      this will free the whole list starting
 *      at *ppFirst and set *ppFirst/*ppLast to NULL
 */

void lstClear(PLISTITEM *ppFirst, PLISTITEM *ppLast, HMTX hmtx, ULONG ulTimeout)
{
    PLISTITEM  pItem, pItem2;

    if (hmtx)
        if (DosRequestMutexSem(hmtx, ulTimeout) != NO_ERROR)
            return;

    pItem = (PLISTITEM)*ppFirst;
    while (pItem) {
        pItem2 = pItem->pNext;
        free(pItem);
        pItem = pItem2;
    }; // while (pItem);

    *ppFirst = NULL;
    if (ppLast)
        *ppLast = NULL;

    if (hmtx)
        DosReleaseMutexSem(hmtx);
}

/********************************************************************
 *                                                                  *
 *   Functions for synchronizing threads                            *
 *                                                                  *
 ********************************************************************/

/*
 * thrCreate:
 *      this function creates a THREADINFO structure in *ppti;
 *      you must pass the thread function in pfn, which will
 *      then be executed; specification of hwndOwner is mandatory
 */

BOOL thrCreate(PTHREADINFO *ppti, PFNTHREAD pfn, HWND hwndOwner)
{
    BOOL            rc = FALSE;
    PTHREADINFO     pti = NULL;

    if (ppti)
    {
        if (*ppti == NULL)
            pti = malloc(sizeof(THREADINFO));
        else
            if ((*ppti)->tid == NULLHANDLE)
                pti = *ppti;

        if (pti) {
            /* we arrive here if *ppti was NULL or (*ppti->tid == NULLHANDLE),
               i.e. the thread is not already running */
            DosCreateThread(&(pti->tid),
                            pfn,
                            (ULONG)pti,
                            CREATE_READY |
                            STACK_COMMITTED,
                            65536);
            pti->cb = sizeof(THREADINFO);
            pti->hwndOwner = hwndOwner;
            pti->fExit = FALSE;

            pti->hab = NULLHANDLE;
            pti->fExitComplete = FALSE;
            pti->ulResult = 0;
            pti->ulFuncInfo = 0;
            rc = TRUE;
        }
        *ppti = pti;
    }
    return (rc);
}

/*
 * thrClose:
 *      this functions sets a thread's "Close" flag to
 *      TRUE; the thread should monitor this flag
 *      periodically and then terminate itself
 */

BOOL thrClose(PTHREADINFO pti)
{
    if (pti) {
        pti->fExit = TRUE;
        return (TRUE);
    }
    return (FALSE);
}

/*
 * thrWait:
 *      this function waits for a thread to end
 */

BOOL thrWait(PTHREADINFO pti)
{
    if (pti) {
        DosWaitThread(&(pti->tid), DCWW_WAIT);
        pti->tid = NULLHANDLE;
        return (TRUE);
    }
    return (FALSE);
}

/*
 * thrFree:
 *      this is a combination of thrClose and
 *      thrWait; the THREADINFO block is then freed
 *      and *ppti set to NULL
 */

BOOL thrFree(PTHREADINFO *ppti)
{
    if (ppti)
        if (*ppti) {
            if ((*ppti)->tid) {
                thrClose(*ppti);
                thrWait(*ppti);
            }
            free(*ppti);
            *ppti = NULL;
            return (TRUE);
        }

    return (FALSE);
}

/*
 * thrKill:
 *      just like thrFree, but the thread is
 *      brutally killed
 */

BOOL thrKill(PTHREADINFO *ppti)
{
    if (ppti)
        if (*ppti) {
            if ((*ppti)->tid) {
                DosKillThread((*ppti)->tid);
            }
            free(*ppti);
            *ppti = NULL;
            return (TRUE);
        }
    return (FALSE);
}

/*
 * thrQueryID:
 *      returns thread ID
 */

TID thrQueryID(PTHREADINFO pti)
{
    if (pti)
        if (!(pti->fExitComplete))
            return (pti->tid);

    return NULLHANDLE;
}

/********************************************************************
 *                                                                  *
 *   String functions                                               *
 *                                                                  *
 ********************************************************************/

/*
 * strhexcat:
 *      this converts a ULONG into a hex string
 *      and appends the result to the buffer pszTarget
 */

void strhexcat(PSZ pszTarget, ULONG ul)
{
/*    CHAR    szBuf[] = "00000000";
    int     i;

    UL2H(szBuf, ul);
    for (i = 0; ((i < 7) && (szBuf[i] == '0')); i++)
        szBuf[i] = ' ';

    strcat(pszTarget, "0x");
    strcat(pszTarget, szBuf+i); */
    sprintf(pszTarget+strlen(pszTarget), "0x%lX", ul);
}

/*
 * strrpl:
 *      replace oldStr by newStr in str.
 *
 *      str should have enough allocated space for the replacement, no check
 *      is made for this. str and OldStr/NewStr should not overlap.
 *      The empty string ("") is found at the beginning of every string.
 *
 *      Returns: pointer to first location behind where NewStr was inserted
 *      or NULL if OldStr was not found.
 *      This is useful for multiple replacements also.
 *      (be careful not to replace the empty string this way !)
 *
 *      Author:     Gilles Kohl
 *      Started:    09.06.1992   12:16:47
 *      Modified:   09.06.1992   12:41:41
 *      Subject:    Replace one string by another in a given buffer.
 *                  This code is public domain. Use freely.
 */

char* strrpl(char* str, char* oldStr, char* newStr)
{
      int OldLen, NewLen;
      char *p, *q;

      if (NULL == (p = strstr(str, oldStr)))
            return p;
      OldLen = strlen(oldStr);
      NewLen = strlen(newStr);
      memmove(q = p+NewLen, p+OldLen, strlen(p+OldLen)+1);
      memcpy(p, newStr, NewLen);
      return (q);
}

/*
 * strinsert:
 *      inserts to_insert into in_str at position place and
 *      copies the result to out_str
 *
 *      Author:     Gilles Kohl
 *      Started:    09.06.1992   12:16:47
 *      Modified:   09.06.1992   12:41:41
 *      Subject:    Replace one string by another in a given buffer.
 *                  This code is public domain. Use freely.
 */

char* strinsert(char* out_str, char* in_str, char* to_insert, int place)
{
    char *out_ptr = out_str, buffer[255], *ptr = buffer;

    if (place <= 0)
        return ptr;

    if ( in_str == out_str )
    {
        while( -- place )
            if ( ! *out_str++ ) return(out_ptr);
            while ( *ptr++ = *in_str++ )
                ; /* strcpy : used if in_str = out_str */
            ptr = buffer;
    }
    else {
        while( -- place )
            if ( ! (*out_str++ = *in_str++) )
                return(out_ptr);

        ptr = in_str;
    }

    while( *out_str++ = *to_insert++ )
        ;
    out_str --;

    while ( *out_str++ = *ptr++ )
        ;

    return out_ptr;
}

/*
 * stristr:
 *      like strstr, but case-insensitive.
 */

PSZ stristr(PSZ string1, PSZ string2)
{
    PSZ prc = NULL;

    if ((string1) && (string2)) {
        PSZ pszSrchIn = strdup(string1);
        PSZ pszSrchFor = strdup(string2);
        strupr(pszSrchIn);
        strupr(pszSrchFor);

        if ((pszSrchIn) && (pszSrchFor)) {
            prc = strstr(pszSrchIn, pszSrchFor);
            if (prc) {
                // prc now has the first occurence of the string,
                // but in pszSrchIn; we need to map this
                // return value to the original string
                prc = (prc-pszSrchIn) // offset in pszSrchIn
                      + string1;
            }
        }
        if (pszSrchFor)
            free(pszSrchFor);
        if (pszSrchIn)
            free(pszSrchIn);
    }
    return (prc);
}

/*
 * strths:
 *      converts a ULONG into a decimal string, while
 *      inserting thousands separators into it. Specify
 *      the separator char in cThousands.
 *      Returns pszTarget so you can use it directly
 *      with sprintf and the "%s" flag.
 */

CHAR* strths(PSZ pszTarget, ULONG ul, CHAR cThousands)
{
    USHORT ust, uss, usc;
    CHAR   szTemp[40];
    sprintf(szTemp, "%d", ul);

    ust = 0;
    usc = strlen(szTemp);
    for (uss = 0; uss < usc; uss++)
    {
        if (uss)
            if (((usc - uss) % 3) == 0)
            {
                pszTarget[ust] = cThousands;
                ust++;
            }
        pszTarget[ust] = szTemp[uss];
        ust++;
    }
    pszTarget[ust] = '\0';

    return (pszTarget);
}

/*
 * strFileDate:
 *      converts FILEFINDBUF4 date data to a string (to pszBuf).
 *      ulDateFormat can be queried using
 *              PrfQueryProfileInt(HINI_USER, "PM_National", "iDate", 0);
 *      meaning:
 *          0   mm.dd.yyyy  (English)
 *          1   dd.mm.yyyy  (German)
 *          2   yyyy.mm.dd  (Japanese)
 *          3   yyyy.dd.mm
 *      cDateSep is used as a date separator (e.g. '.').
 *              PrfQueryProfileString(HINI_USER, "PM_National", "sDate", "/",
 *                  szDateSep, sizeof(szDateSep)-1);
 */

VOID strFileDate(PSZ pszBuf, FILEFINDBUF4 *pffb4, ULONG ulDateFormat, CHAR cDateSep)
{
    switch (ulDateFormat) {
        case 0:  // MDY
            sprintf(pszBuf, "%02d%c%02d%c%04d",
                pffb4->fdateLastWrite.month, cDateSep,
                pffb4->fdateLastWrite.day, cDateSep,
                ((pffb4->fdateLastWrite.year)+1980));
        break;
        case 1:  // DMY
            sprintf(pszBuf, "%02d%c%02d%c%04d",
                pffb4->fdateLastWrite.day, cDateSep,
                pffb4->fdateLastWrite.month, cDateSep,
                ((pffb4->fdateLastWrite.year)+1980));
        break;
        case 2: // YMD
            sprintf(pszBuf, "%04d%c%02d%c%02d",
                ((pffb4->fdateLastWrite.year)+1980), cDateSep,
                pffb4->fdateLastWrite.month, cDateSep,
                pffb4->fdateLastWrite.day);
        break;
        default: // YDM -- not documented!
            sprintf(pszBuf, "%04d%c%02d%c%02d",
                ((pffb4->fdateLastWrite.year)+1980), cDateSep,
                pffb4->fdateLastWrite.day, cDateSep,
                pffb4->fdateLastWrite.month);
        break;
    }
}

/*
 * strFileTime:
 *      converts FILEFINDBUF4 time data to a string (to pszBuf).
 *      ulTimeFormat can be queried using
 *              PrfQueryProfileInt(HINI_USER, "PM_National", "iTime", 0);
 *      meaning:
 *          0   12-hour clock
 *          >0  24-hour clock
 *      cDateSep is used as a date separator (e.g. ':').
 *              PrfQueryProfileString(HINI_USER, "PM_National", "sTime", ":",
 *                  szTimeSep, sizeof(szTimeSep)-1);
 */

VOID strFileTime(PSZ pszBuf, FILEFINDBUF4 *pffb4, ULONG ulTimeFormat, CHAR cTimeSep)
{
    if (ulTimeFormat == 0) { // 12-hour clock
        CHAR szAMPM[10];
        PrfQueryProfileString(HINI_USER, "PM_National",
            (pffb4->ftimeLastWrite.hours > 12) ? "s2359" : "s1159",
            (pffb4->ftimeLastWrite.hours > 12) ? "PM" : "AM",
            szAMPM, sizeof(szAMPM)-1);

        sprintf(pszBuf, "%02d%c%02d%c%02d %s",
            (pffb4->ftimeLastWrite.hours > 12)  // leave 12 == 12 (not 0)
              ? (pffb4->ftimeLastWrite.hours % 12)
              :  pffb4->ftimeLastWrite.hours,
                cTimeSep,
            pffb4->ftimeLastWrite.minutes, cTimeSep,
            (pffb4->ftimeLastWrite.twosecs*2),
            szAMPM);
    }
    else // 24-hour clock
        sprintf(pszBuf, "%02d%c%02d%c%02d",
            pffb4->ftimeLastWrite.hours, cTimeSep,
            pffb4->ftimeLastWrite.minutes, cTimeSep,
            (pffb4->ftimeLastWrite.twosecs*2));
}

/*
 * strFindKey:
 *      finds pszKey in pszSearchIn; similar to stristr,
 *      but this one makes sure the key is at the beginning
 *      of a line. Spaces before the key are tolerated.
 *      Used by strGet/SetParameter
 */

PSZ strFindKey(PSZ pszSearchIn, PSZ pszKey)
{
    PSZ     p = NULL;
    BOOL    fFound = FALSE;

    p = pszSearchIn;
    do {
        p = stristr(p, pszKey);

        if ((p) && (p > pszSearchIn)) {
            // make sure the key is at the beginning of a line
            // by going backwards until we find a char != " "
            PSZ p2 = p-1;
            while ( (*p2 == ' ') && (p2 > pszSearchIn) )
                p2--;

            // if this char is an EOL sign, go on
            if (    (*p2 == '\r')
                 || (*p2 == '\n')
                 || (p2 == pszSearchIn)
               )
            {
                fFound = TRUE; // go on, p contains found key
            } // else search next key
            else
                p++; // search on after this key
        }
    } while ((!fFound) && (p != NULL) && (p != pszSearchIn));

    return (p);
}

/*
 * strGetParameter:
 *      searches pszSearchIn for the key pszKey; if found, it
 *      returns a pointer to the following characters in pszSearchIn
 *      and, if pszCopyTo != NULL, copies the rest of the line to
 *      that buffer, of which cbCopyTo specified the size.
 *      If the key is not found, NULL is returned.
 *      This is useful for querying CONFIG.SYS settings.
 */

PSZ strGetParameter(PSZ pszSearchIn, PSZ pszKey, PSZ pszCopyTo, ULONG cbCopyTo)
{
    PSZ p = strFindKey(pszSearchIn, pszKey),
        prc = NULL;
    // copy to pszCopyTo
    if (p) {
        prc = p + strlen(pszKey);
        if (pszCopyTo) {
            PSZ pEOL = strstr(prc, "\r\n");
            if (pEOL) {
                ULONG cb = (pEOL-prc);
                if (cb > cbCopyTo)
                    cb = cbCopyTo-1;
                strncpy(pszCopyTo, prc, cb);
                pszCopyTo[cbCopyTo-1] = '\0';
            }
        }
    }

    return (prc);
}

/*
 * strSetParameter:
 *      searches pszSearchIn for the key pszKey; if found, it
 *      replaces the characters following this key up to the
 *      end of the line by pszParam. If pszKey is not found in
 *      pszSearchIn, it is appended in a new line. You must
 *      ensure that *pszSearchin is large enough for the
 *      manipulation, i.e. it must be at least of the following
 *      size:
 *          strlen(pszSearchIn) + strlen(pszNewParam)
 *      in case the key was not found in the text.
 *      This function searches w/out case sensitivity.
 *      This function returns a pointer to the new parameter.
 */

PSZ strSetParameter(PSZ pszSearchIn, PSZ pszKey, PSZ pszNewParam)
{
    PSZ pKey = strFindKey(pszSearchIn, pszKey),
        prc = NULL;

    /* printf("Going for %s%s\n", pszKey, pszNewParam);
    printf("  Old file length: %d\n", strlen(pszSearchIn)); */

    if (pKey) {
        // key found in file:
        // replace existing parameter
        PSZ pOldParam = pKey + strlen(pszKey);
        prc = pOldParam;
        // pOldParam now has the old parameter, which we
        // will overwrite now

        // printf("  found %s @ %d\n", pszKey, (pKey-pszSearchIn));
        if (pOldParam) {
            PSZ pEOL = strstr(pOldParam, "\r\n");
            // pEOL now has first end-of-line after the
            // parameter

            // printf("  old param is @ %d\n", pOldParam-pszSearchIn);
            if (pEOL) {
                ULONG ulOldParam = (pEOL-pOldParam);
                PSZ pszOldCopy = malloc(ulOldParam+1);
                strncpy(pszOldCopy, pOldParam, ulOldParam);
                pszOldCopy[ulOldParam] = '\0';
                // printf("  found EOL @ %d\n", pEOL-pszSearchIn);

                strrpl(pOldParam, pszOldCopy, pszNewParam);

                free(pszOldCopy);
            }
        }
    } else {
        // key not found: append to end of file
        strcat(pszSearchIn, "\r\n");
        strcat(pszSearchIn, pszKey);
        strcat(pszSearchIn, pszNewParam);
        strcat(pszSearchIn, "\r\n");
        // printf("  Appended %s%s to end of file\n", pszKey, pszNewParam);
    }

    return (prc);
}

/*
 * strDeleteLine:
 *      this deletes the line in pszSearchIn which starts with
 *      the key pszKey. Returns TRUE if the line was found and
 *      deleted.
 */

BOOL strDeleteLine(PSZ pszSearchIn, PSZ pszKey)
{
    PSZ pKey = strFindKey(pszSearchIn, pszKey);
    BOOL brc = FALSE;

    if (pKey) {
        PSZ pEOL = strstr(pKey, "\r\n");
        // pEOL now has first end-of-line after the key
        if (pEOL) {
            // delete line by overwriting it with
            // the next line
            strcpy(pKey, pEOL+2);
            /* PSZ pFrom = pEOL+1,
                pTo = pKey-1;
            do {
                pTo++;
                pFrom++;
                *pTo = *pFrom;
            } while (*pFrom != '\0'); */
        } else {
            // EOL not found: we must be at the end of the file
            *pKey = '\0';
        }
        brc = TRUE;
    }

    return (brc);
}

/*
 * strBeautifyTitle:
 *      replaces line breaks (0xd, 0xa) with spaces
 */

BOOL strBeautifyTitle(PSZ psz)
{
    BOOL rc = FALSE;
    CHAR *p;
    while (p = strchr(psz, 0xa))
    {
        *p = ' ';
        rc = TRUE;
    }
    while (p = strchr(psz, 0xd))
    {
        *p = ' ';
        rc = TRUE;
    }
    return (rc);
}

/********************************************************************
 *                                                                  *
 *   Common helper routines                                         *
 *                                                                  *
 ********************************************************************/

/*
 * UL2H:
 *      converts a ULONG into an 8-digit hex string
 */

void UL2H(PSZ szMsg, ULONG ul)
{
    CHAR    digits[] = "0123456789ABCDEF";
    int     i;
    ULONG   test = ul;
    for (i = 0; i < 8; i++)
    {
        szMsg[(7-i)] = digits[(test & 0xF)];
        test = test >> 4;
    }
}

/*
 * printfRtl:
 *      printf's rectangle information
 */

VOID printfRtl(PSZ psz, PRECTL prcl)
{
    _Pmpf(("%s xLeft: %d, yBottom: %d, xRight: %d, yTop: %d\n",
        psz,
        prcl->xLeft, prcl->yBottom, prcl->xRight, prcl->yTop));
}

/*
 * DosGetULongTime:
 *      this returns the current time as a ULONG value (in milliseconds).
 *      Useful for stopping how much time the machine has spent in
 *      a certain function.
 */

ULONG DosGetULongTime(VOID) {
    DATETIME    dt;
    DosGetDateTime(&dt);
    return (10*(dt.hundredths + 100*(dt.seconds + 60*(dt.minutes + 60*(dt.hours)))));
}

/*
 * DosQueryThreadPriority:
 *          returns the priority of the calling thread.
 *          The low byte of the low word is a hexadecimal value
 *          representing a rank (value 0 to 31) within a priority class.
 *          Class values, found in the high byte of the low word, are as follows:
 *              0x01  idle
 *              0x02  regular
 *              0x03  time-critical.
 *              0x04  server
 */

ULONG DosQueryThreadPriority(VOID)
{
    PTIB    ptib;
    PPIB    ppib;
    if (DosGetInfoBlocks(&ptib, &ppib) == NO_ERROR)
        if (ptib)
            if (ptib->tib_ptib2)
                return (ptib->tib_ptib2->tib2_ulpri);
    return (0);
}

/*
 * DosBrutallyKillProcess:
 *      this kills a process by using the new OS/2 "kill, kill"
 *      interface of Warp 3 FP17 or Warp 4; requires that
 *      XF86SUP.SYS be installed (from XFREE86/OS2)
 *
 *      (W) (C) Holger Veit.
 *      Taken from the XFree86/OS/2 packages.
 *      Public domain (I guess).
 */

APIRET DosBrutallyKillProcess(PID pid)
{
    APIRET rc;
    HFILE hfd = 0;
    ULONG action,plen;
    USHORT param;

    rc = DosOpen((PSZ)"/dev/fastio$", (PHFILE)&hfd, (PULONG)&action,
        (ULONG)0, FILE_SYSTEM, FILE_OPEN,
        OPEN_SHARE_DENYNONE|OPEN_FLAGS_NOINHERIT|OPEN_ACCESS_READONLY,
        (ULONG)0);
    if (rc == NO_ERROR) {
        param = pid;

        rc = DosDevIOCtl(hfd, (ULONG)0x76, (ULONG)0x65,
            (PULONG*)&param,sizeof(USHORT),&plen,
            NULL, 0, NULL);

        DosClose(hfd);
    }
    return (rc);
}

/*
 * DosQueryShiftState:
 *      returns TRUE if any of the SHIFT keys are
 *      currently pressed
 */

BOOL DosQueryShiftState(VOID)
{
    // BYTE abBuf[512];
    HFILE hfKbd;
    ULONG ulAction; //, cbRead, cbWritten;
    SHIFTSTATE      ShiftState;
    ULONG           cbDataLen = sizeof(ShiftState);

    DosOpen("KBD$", &hfKbd, &ulAction, 0,
             FILE_NORMAL, FILE_OPEN,
             OPEN_ACCESS_READONLY | OPEN_SHARE_DENYWRITE,
             (PEAOP2)NULL);

    DosDevIOCtl(hfKbd, IOCTL_KEYBOARD, KBD_GETSHIFTSTATE,
                      NULL, 0, NULL,      // no parameters
                      &ShiftState, cbDataLen, &cbDataLen);

    DosClose(hfKbd);

    return ((ShiftState.fsState & 3) != 0);
}


/*
 * DosIsWarp4:
 *      returns TRUE if at least OS/2 Warp 4 is running, FALSE otherwise
 */

BOOL DosIsWarp4(VOID)
{
    ULONG       aulBuf[3];

    DosQuerySysInfo(QSV_VERSION_MAJOR, QSV_VERSION_MINOR,
                    &aulBuf, sizeof(aulBuf));
    return ((aulBuf[0] != 20) || (aulBuf[1] > 30));
}

/*
 * DosQueryBootDrive:
 *      returns the letter of the boot drive
 */

CHAR DosQueryBootDrive(VOID)
{
    ULONG ulBootDrive;
    CHAR    cBootDrives[27] = " ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    DosQuerySysInfo(5, 5, &ulBootDrive, sizeof(ulBootDrive));
    return (cBootDrives[ulBootDrive]);
}

/*
 * DosQueryFileSize:
 *      returns the size of an already opened file
 */

ULONG DosQueryFileSize(HFILE hFile)
{
    FILESTATUS3 fs3;
    if (DosQueryFileInfo(hFile, FIL_STANDARD, &fs3, sizeof(fs3)))
        return (0);
    else
        return (fs3.cbFile);
}

/*
 * DosQueryPathSize:
 *      returns the size of anyfile
 */

ULONG DosQueryPathSize(PSZ pszFile)
{
    FILESTATUS3 fs3;
    if (DosQueryPathInfo(pszFile, FIL_STANDARD, &fs3, sizeof(fs3)))
        return (0);
    else
        return (fs3.cbFile);
}

/*
 * DosReadTextFile:
 *      reads a text file from disk, allocates memory
 *      via malloc() and returns a pointer to this
 *      buffer (or NULL upon errors). Specify in
 *      ulExtraMemory how much extra memory (in addition
 *      to the file's size) should be allocated to
 *      allow text manipulation.
 *      Returns NULL if an error occured.
 */

PSZ DosReadTextFile(PSZ pszFile, ULONG ulExtraMemory)
{
    ULONG   ulSize,
            ulBytesRead = 0,
            ulAction, ulLocal;
    HFILE   hFile;
    PSZ     pszContent = NULL;
    APIRET rc = DosOpen(pszFile,
                 &hFile,
                 &ulAction,                      // action taken
                 5000L,                          // primary allocation size
                 FILE_ARCHIVED | FILE_NORMAL,    // file attribute
                 OPEN_ACTION_CREATE_IF_NEW |
                 OPEN_ACTION_OPEN_IF_EXISTS,     // open flags
                 OPEN_FLAGS_NOINHERIT |
                 OPEN_SHARE_DENYNONE  |
                 OPEN_ACCESS_READONLY,               // read-write mode
                 NULL);                          // no EAs

    if (rc == NO_ERROR) {
        ulSize = DosQueryFileSize(hFile);
        pszContent = malloc(ulSize+10+ulExtraMemory);
        rc = DosSetFilePtr(hFile,
                            0L,
                            FILE_BEGIN,
                            &ulLocal);
        rc = DosRead(hFile,
                      pszContent,
                      ulSize,
                      &ulBytesRead);
        DosClose(hFile);
        *(pszContent+ulBytesRead) = '\0';
    }
    return (pszContent);
}

/*
 * DosCreateBackupFileName:
 *      creates a new PSZ (malloc) with a proper
 *      backup filename which does not exist yet.
 */

PSZ DosCreateBackupFileName(PSZ pszExisting)
{
    // CHAR    szTemp[CCHMAXPATH];
    PSZ     pszNew = strdup(pszExisting),
            pszLastDot = strrchr(pszNew, '.');
    ULONG   ulCount = 1;
    CHAR    szCount[5];
    do {
        sprintf(szCount, ".%03d", ulCount);
        strcpy(pszLastDot, szCount);
        ulCount++;
    } while (DosQueryPathSize(pszNew) > 0);
    return (pszNew);
}

/*
 * DosWriteTextFile:
 *      writes a text file to disk; pszFile must contain the
 *      whole path and filename.
 *      An existing file will be backed up if (fBackup).
 *      Returns the number of bytes written.
 */

ULONG DosWriteTextFile(PSZ pszFile, PSZ pszContent, BOOL fBackup)
{
    ULONG   ulWritten = 0,
            ulAction, ulLocal;
    HFILE   hFile;
    APIRET rc;

    if (fBackup) {
        PSZ     pszBackup = DosCreateBackupFileName(pszFile);
        DosCopy(pszFile, pszBackup, DCPY_EXISTING);
        free(pszBackup);
    }

    rc = DosOpen(pszFile,
                 &hFile,
                 &ulAction,                      // action taken
                 5000L,                          // primary allocation size
                 FILE_ARCHIVED | FILE_NORMAL,    // file attribute
                 OPEN_ACTION_CREATE_IF_NEW |
                 OPEN_ACTION_REPLACE_IF_EXISTS,     // open flags
                 OPEN_FLAGS_NOINHERIT |
                 OPEN_SHARE_DENYNONE  |
                 OPEN_ACCESS_WRITEONLY,          // read-write mode
                 NULL);                          // no EAs

    if (rc == NO_ERROR) {
        rc = DosSetFilePtr(hFile,
                            0L,
                            FILE_BEGIN,
                            &ulLocal);
        rc = DosWrite(hFile,
                      pszContent,
                      strlen(pszContent),  // exclude 0 byte
                      &ulWritten);
        DosClose(hFile);
    }
    return (ulWritten);
}

/*
 * DosOpenLogFile:
 *      this opens a log file in the root directory of
 *      the boot drive; it is titled pszFilename, and
 *      the file handle is returned
 */

HFILE DosOpenLogFile(PSZ pszFileName)
{
    APIRET  rc;
    CHAR    szFileName[CCHMAXPATH];
    HFILE   hfLog;
    ULONG   ulAction;
    ULONG   ibActual;

    sprintf(szFileName, "%c:\\%s", DosQueryBootDrive(), pszFileName);
    rc = DosOpen(szFileName, &hfLog, &ulAction, 0,
                 FILE_NORMAL,
                 OPEN_ACTION_CREATE_IF_NEW | OPEN_ACTION_OPEN_IF_EXISTS,
                 OPEN_ACCESS_WRITEONLY | OPEN_SHARE_DENYWRITE,
                 (PEAOP2)NULL);
    if (rc == NO_ERROR) {
        DosSetFilePtr(hfLog, 0, FILE_END, &ibActual);
        return (hfLog);
    } else
        return (0);
}

/*
 * DosWriteToLogFile
 *      writes a string to a log file
 */

APIRET DosWriteToLogFile(HFILE hfLog, PSZ psz)
{
    if (hfLog) {
        DATETIME dt;
        CHAR szTemp[2000];
        ULONG   cbWritten;
        DosGetDateTime(&dt);
        sprintf(szTemp, "Time: %02d:%02d:%02d %s",
            dt.hours, dt.minutes, dt.seconds,
            psz);
        return (DosWrite(hfLog, psz, strlen(psz), &cbWritten));
    }
    else return (ERROR_INVALID_HANDLE);
}

/*
 * DosQuickStartSession:
 *      shortcut to DosStartSession w/out having to deal
 *      with all the parameters; this one needs the following:
 *          PSZ pszPath     the executable (e.g. "F:\OS2\CMD.EXE")
 *          PSZ pszParams   parameters to pass (e.g. "/c blah.cmd")
 *          BOOL Visible    make program visible or not
 *          BOOL Wait       if TRUE: this function does not return until
 *                          the program has finished; DO NOT use this
 *                          on a PM main thread, or everything will block
 *          PULONG pulSID   session ID to be returned by DosStartSession
 *          PPID   ppid     process ID to be returned by DosStartSession
 */

APIRET DosQuickStartSession(PSZ pszPath, PSZ pszParams, BOOL Visible, BOOL Wait,
            PULONG pulSID, PPID ppid)
{
    APIRET rc;
    REQUESTDATA rqdata;
    ULONG DataLength; PULONG DataAddress; BYTE elpri;
    int er;
    // Queue Stuff
    HQUEUE hq;
    PID qpid=0;
    STARTDATA   SData;
    CHAR        szObjBuf[CCHMAXPATH];

    if (Wait) {
        if (    (er=DosCreateQueue(&hq, QUE_FIFO|QUE_CONVERT_ADDRESS, "\\queues\\kfgstart.que"))
             != NO_ERROR)
        {
            char str[80];
            sprintf(str, "Create %ld\n", er);
            DebugBox("DosQuickStartSession:", str);
        }

        if ((er=DosOpenQueue(&qpid, &hq, "\\queues\\kfgstart.que")) != NO_ERROR)
        {  char str[80];
           sprintf(str, "Open %ld\n", er);
           DebugBox("DosQuickStartSession:", str);
        }
    }

    SData.Length  = sizeof(STARTDATA);
    SData.Related = SSF_RELATED_CHILD; //INDEPENDENT;
    SData.FgBg    = SSF_FGBG_FORE;
    SData.TraceOpt = SSF_TRACEOPT_NONE;

    SData.PgmTitle = pszPath;       // title for window
    SData.PgmName = pszPath;
    SData.PgmInputs = pszParams;

    SData.TermQ = (Wait) ? "\\queues\\kfgstart.que" : NULL;
    SData.Environment = 0;
    SData.InheritOpt = SSF_INHERTOPT_SHELL;
    SData.SessionType = SSF_TYPE_DEFAULT;
    SData.IconFile = 0;
    SData.PgmHandle = 0;

    SData.PgmControl = ((Visible) ? SSF_CONTROL_VISIBLE : SSF_CONTROL_INVISIBLE);
    SData.InitXPos  = 30;
    SData.InitYPos  = 40;
    SData.InitXSize = 200;
    SData.InitYSize = 140;
    SData.Reserved = 0;
    SData.ObjectBuffer  = (CHAR*)&szObjBuf;
    SData.ObjectBuffLen = (ULONG)sizeof(szObjBuf);

    rc = DosStartSession(&SData, pulSID, ppid);

    if (Wait) {
        rqdata.pid=qpid;
        DosReadQueue(hq, &rqdata, &DataLength, (PVOID*)&DataAddress, 0, 0, &elpri, 0);
        DosCloseQueue(hq);
    }
    return (rc);
}

/*
 * DosSetPathAttr:
 *      sets the file attributes of pszFile,
 *      which must be fully qualified.
 *      fAttr can be:
 *              FILE_ARCHIVED
 *              FILE_READONLY
 *              FILE_SYSTEM
 *              FILE_HIDDEN
 */

APIRET DosSetPathAttr(PSZ pszFile, ULONG fAttr)
{
    FILESTATUS3 fs3;
    APIRET rc = DosQueryPathInfo(pszFile,
            FIL_STANDARD,
            &fs3,
            sizeof(fs3));

    if (rc == NO_ERROR) {
        fs3.attrFile = fAttr;
        rc = DosSetPathInfo(pszFile,
                FIL_STANDARD,
                &fs3,
                sizeof(fs3),
                DSPI_WRTTHRU);
    }
    return (rc);
}

/********************************************************************
 *                                                                  *
 *   Extended Attribute handling                                    *
 *                                                                  *
 ********************************************************************/

/* All the following code dealing with Extended Attributes is
   written (w) by Chris Hanson (cph@zurich.ai.mit.edu)
   Copyright (c) 1995 Massachusetts Institute of Technology
   Modified 1998 Ulrich Mller

   This material was developed by the Scheme project at the Massachusetts
   Institute of Technology, Department of Electrical Engineering and
   Computer Science.  Permission to copy this software, to redistribute
   it, and to use it for any purpose is granted, subject to the following
   restrictions and understandings.

   1. Any copy made of this software must include this copyright notice
   in full.

   2. Users of this software agree to make their best efforts (a) to
   return to the MIT Scheme project any improvements or extensions that
   they make, so that these may be included in future releases; and (b)
   to inform MIT of noteworthy uses of this software.

   3. All materials developed as a consequence of the use of this
   software shall duly acknowledge such use, in accordance with the usual
   standards of acknowledging credit in academic research.

   4. MIT has made no warrantee or representation that the operation of
   this software will be error-free, and MIT is under no obligation to
   provide any services, by way of maintenance, update, or otherwise.

   5. In conjunction with products arising from the use of this material,
   there shall be no use of the name of the Massachusetts Institute of
   Technology nor of any adaptation thereof in any advertising,
   promotional, or sales literature without prior written consent from
   MIT in each case. */

#define EA_BINDING_FLAGS(binding) ((binding)->bFlags)
#define EA_BINDING_NAME_LENGTH(binding) ((binding)->bNameLength)
#define EA_BINDING_VALUE_LENGTH(binding) ((binding)->usValueLength)
#define EA_BINDING_NAME(binding) ((binding)->pszName)
#define EA_BINDING_VALUE(binding) ((binding)->pszValue)

#define EA_LIST_BINDING(list) ((list)->binding)
#define EA_LIST_NEXT(list) ((list)->next)

static PEALIST      read_eas (ULONG, PVOID);
static EABINDING *  read_ea_by_index (ULONG, PVOID, ULONG);
static EABINDING *  read_ea_by_name (ULONG, PVOID, PSZ);
static PDENA2       read_dena_by_index (ULONG, PVOID, ULONG);
static PEABINDING   get_ea_value (ULONG, PVOID, PDENA2);
static void         setup_query_ea_info (PDENA2, PEAOP2);
static PEABINDING   convert_feal_to_binding (PFEA2LIST);
static void         write_eas (ULONG, PVOID, PEALIST);
static void         write_ea (ULONG, PVOID, PEABINDING );
static PFEA2LIST    convert_binding_to_feal (PEABINDING );
static void         fatal_api_call_error (APIRET, const char *);

#define API_CALL(proc, args) do                     \
{                                   \
  APIRET API_CALL_rc = (proc args);                 \
  if (API_CALL_rc != NO_ERROR)                      \
    fatal_api_call_error (API_CALL_rc, #proc);              \
} while (0)

#define FREE(ptr) free ((void *) (ptr))

/* The following procedures deallocate the EA binding memory
   structures that are generated by the `read_...' procedures
   described above.  These procedures assume that `malloc' is used for
   allocation of all storage, and that the `name' and `value' fields
   of each binding are also allocated with `malloc'.  The `value'
   field may also be 0.  */

/*
 * EAFreeBinding:
 *
 */

void EAFreeBinding(PEABINDING binding)
{
  FREE (EA_BINDING_NAME(binding));
  if ((EA_BINDING_VALUE(binding)) != 0)
    FREE (EA_BINDING_VALUE(binding));
  FREE (binding);
}

/*
 * EAFreeList:
 *
 */

void EAFreeList(PEALIST list)
{
  while (list != 0)
    {
      PEALIST next = (EA_LIST_NEXT (list));
      EAFreeBinding (EA_LIST_BINDING (list));
      FREE (list);
      list = next;
    }
}

/*
 * EAHFReadAll:
 * Read all of the extended attributes, and return them as a list.
 * Returns 0 if there are no attributes.  The returned list of
 * attributes should be freed by calling `EAFreeList'.
 */

PEALIST EAHFileReadAll(HFILE file_handle)
{
  return (read_eas (ENUMEA_REFTYPE_FHANDLE, (& file_handle)));
}

PEALIST EAPathReadAll(PSZ path)
{
  return (read_eas (ENUMEA_REFTYPE_PATH, path));
}

/*
 * EAHFReadOneByIndex:
 * Return the extended attribute specified by a given index.  The
 * index of the first attribute is 1.  Returns 0 if the index is
 * larger than the number of attributes attached to the file; in
 * particular, returns 0 for index 1 if there are no attributes.  The
 * returned binding should be freed by calling `EAFreeBinding'.
 */

PEABINDING EAHFileReadOneByIndex(HFILE file_handle, ULONG index)
{
  return (read_ea_by_index (ENUMEA_REFTYPE_FHANDLE, (& file_handle), index));
}

PEABINDING EAPathReadOneByIndex(PSZ path, ULONG index)
{
  return (read_ea_by_index (ENUMEA_REFTYPE_PATH, path, index));
}

/*
 * EAPReadOneByName:
 * Return the extended attribute specified by the given name.  If
 * there is no such attribute, returns 0.  The returned binding should
 * be freed by calling `EAFreeBinding'.
 */

PEABINDING EAHFileReadOneByName(HFILE file_handle, PSZ name)
{
  return (read_ea_by_name (ENUMEA_REFTYPE_FHANDLE, (& file_handle), name));
}

PEABINDING EAPathReadOneByName(PSZ path, PSZ name)
{
  return (read_ea_by_name (ENUMEA_REFTYPE_PATH, path, name));
}

/* The following procedures write extended attributes to a given file.
   A given attribute is deleted if its `usValueLength' is 0; in that
   case the `value' field is also permitted to be 0.  */

/*
 * EAHFWriteAll:
 * Write a list of extended attributes to a given file.  These
 * attributes add to any attributes already stored on the file.
 */

void EAHFileWriteAll(HFILE file_handle, PEALIST list)
{
  write_eas (ENUMEA_REFTYPE_FHANDLE, (& file_handle), list);
}

void EAPathWriteAll(PSZ path, PEALIST list)
{
  write_eas (ENUMEA_REFTYPE_PATH, path, list);
}

/*
 * EAHFWriteOne:
 * Write a single extended attribute to a given file.  This attribute
 * is added to any attributes already stored on the file.
 */

void EAHFileWriteOne(HFILE file_handle, PEABINDING binding)
{
  write_ea (ENUMEA_REFTYPE_FHANDLE, (& file_handle), binding);
}

void EAPathWriteOne(PSZ path, PEABINDING binding)
{
  write_ea (ENUMEA_REFTYPE_PATH, path, binding);
}

static PEALIST read_eas (ULONG type, PVOID pfile)
{
  ULONG index = 1;
  PEALIST head = 0;
  PEALIST tail = 0;
  while (1)
    {
      PEABINDING binding = (read_ea_by_index (type, pfile, index));
      if (binding == 0)
    break;
      {
    PEALIST list = (malloc(sizeof (EALIST)));
    (EA_LIST_BINDING (list)) = binding;
    (EA_LIST_NEXT (list)) = 0;
    if (head == 0)
      head = list;
    else
      (EA_LIST_NEXT (tail)) = list;
    tail = list;
      }
      index += 1;
    }
  return (head);
}

static PEABINDING read_ea_by_index (ULONG type, PVOID pfile, ULONG index)
{
  PDENA2 dena = (read_dena_by_index (type, pfile, index));
  return ((dena == 0) ? 0 : (get_ea_value (type, pfile, dena)));
}

static PEABINDING read_ea_by_name (ULONG type, PVOID pfile, PSZ name)
{
    ULONG index = 1;
    while (1)
    {
        PDENA2 dena = read_dena_by_index (type, pfile, index);
        if (dena == 0)
            return (NULL);
        if ((strcmp(name, (dena->szName))) == 0)
            return (get_ea_value (type, pfile, dena));
        FREE (dena);
        index += 1;
    }
}

static PDENA2 read_dena_by_index (ULONG type, PVOID pfile, ULONG index)
{
  ULONG count = 1;
  PDENA2 dena = (malloc(500)); // 500 is magic -- IBM doesn't explain.
  API_CALL (DosEnumAttribute, (type, pfile, index, dena, 500, (& count),
                   ENUMEA_LEVEL_NO_VALUE));
  if (count == 0)
    {
      FREE (dena);
      return (0);
    }
  else
    return (dena);
}

static PEABINDING get_ea_value (ULONG type, PVOID pfile, PDENA2 dena)
{
  ULONG level = FIL_QUERYEASFROMLIST;
  EAOP2 eaop;
  ULONG size = (sizeof (eaop));
  setup_query_ea_info (dena, (& eaop));
  if (type == ENUMEA_REFTYPE_FHANDLE)
    API_CALL (DosQueryFileInfo, ((* ((PHFILE) pfile)), level, (& eaop), size));
  else
    API_CALL (DosQueryPathInfo, (pfile, level, (& eaop), size));
  FREE (eaop . fpGEA2List);
  return (convert_feal_to_binding (eaop . fpFEA2List));
}

static void setup_query_ea_info (PDENA2 dena, PEAOP2 eaop)
{
  unsigned int geal_size = ((sizeof (GEA2LIST)) + (dena->cbName));
  unsigned int feal_size
    = ((sizeof (FEA2LIST)) + (dena->cbName) + (dena->cbValue));
  (eaop->fpGEA2List) = (malloc(geal_size));
  ((eaop->fpGEA2List)->cbList) = geal_size;
  (eaop->fpFEA2List) = (malloc(feal_size));
  ((eaop->fpFEA2List)->cbList) = feal_size;
  (eaop->oError) = 0;
  {
    PGEA2 gea = (& (((eaop->fpGEA2List)->list) [0]));
    (gea->oNextEntryOffset) = 0;
    (gea->cbName) = (dena->cbName);
    strcpy ((gea->szName), (dena->szName));
  }
  FREE (dena);
}

static PEABINDING convert_feal_to_binding (PFEA2LIST feal)
{
  PFEA2 fea = (& ((feal->list) [0]));
  PEABINDING binding = (malloc(sizeof (EABINDING )));
  (EA_BINDING_FLAGS (binding)) = (fea->fEA);
  (EA_BINDING_NAME_LENGTH (binding)) = (fea->cbName);
  (EA_BINDING_VALUE_LENGTH (binding)) = (fea->cbValue);
  (EA_BINDING_NAME (binding)) = (malloc((fea->cbName) + 1));
  strcpy ((EA_BINDING_NAME (binding)), (fea->szName));
  (EA_BINDING_VALUE (binding)) = (malloc(fea->cbValue));
  memcpy ((EA_BINDING_VALUE (binding)),
      (& ((fea->szName) [(fea->cbName) + 1])),
      (fea->cbValue));
  FREE (feal);
  return (binding);
}

static void write_eas (ULONG type, PVOID pfile, PEALIST list)
{
  while (list != 0)
    {
      write_ea (type, pfile, (EA_LIST_BINDING (list)));
      list = (EA_LIST_NEXT (list));
    }
}

static void write_ea (ULONG type, PVOID pfile, PEABINDING binding)
{
  ULONG level = FIL_QUERYEASIZE;
  EAOP2 eaop;
  ULONG size = (sizeof (eaop));
  (eaop . fpGEA2List) = 0;
  (eaop . fpFEA2List) = (convert_binding_to_feal (binding));
  (eaop . oError) = 0;
  if (type == ENUMEA_REFTYPE_FHANDLE)
    API_CALL (DosSetFileInfo, ((* ((PHFILE) pfile)), level, (& eaop), size));
  else
    API_CALL (DosSetPathInfo, (pfile, level, (& eaop), size, DSPI_WRTTHRU));
  FREE (eaop . fpFEA2List);
}

static PFEA2LIST convert_binding_to_feal (PEABINDING binding)
{
  unsigned int feal_size
    = ((sizeof (FEA2LIST))
       + (EA_BINDING_NAME_LENGTH (binding))
       + (EA_BINDING_VALUE_LENGTH (binding)));
  PFEA2LIST feal = (malloc(feal_size));
  PFEA2 fea = (& ((feal->list) [0]));
  (feal->cbList) = feal_size;
  (fea->oNextEntryOffset) = 0;
  (fea->fEA) = (EA_BINDING_FLAGS (binding));
  (fea->cbName) = (EA_BINDING_NAME_LENGTH (binding));
  (fea->cbValue) = (EA_BINDING_VALUE_LENGTH (binding));
  strcpy ((fea->szName), (EA_BINDING_NAME (binding)));
  if ((EA_BINDING_VALUE (binding)) != 0)
    memcpy ((& ((fea->szName) [(fea->cbName) + 1])),
        (EA_BINDING_VALUE (binding)),
        (fea->cbValue));
  return (feal);
}

void (* ea_error_hook) (APIRET, const char *);

static void fatal_api_call_error (APIRET rc, const char * name)
{
  if (ea_error_hook != 0)
    (* ea_error_hook) (rc, name);
  else
    {
      fprintf (stderr, "Fatal error %d in procedure %s.\n", rc, name);
      fflush (stderr);
    }
  exit (EXIT_FAILURE);
}

/********************************************************************
 *                                                                  *
 *   Miscellaneous                                                  *
 *                                                                  *
 ********************************************************************/

/*
 * DumbDlgProc:
 *      this is what the name says: a really dumb dlg proc
 */

MRESULT EXPENTRY DumbDlgProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
    return (WinDefDlgProc(hwnd, msg, mp1, mp2));
}

/*
 * cmnINIError:
 *      this is called whenever an error occurs saving a
 *      profile. This will compose an error string and
 *      all the callback func fncbError.
 */

ULONG cmnINIError(ULONG ulOptions, HAB hab, HFILE hfLog, PFNWP fncbError, PSZ pszErrorString)
{
    ULONG ulrc;
    CHAR szError2[2000];

    if (fncbError) {
        if (ulOptions == MB_ABORTRETRYIGNORE)
            sprintf(szError2, "%s"
                              "\nWinGetLastError: 0x%lX"
                              "\nPress 'Abort' to abort saving the INI files. "
                              "\nPress 'Retry' to attempt saving this INI file again. "
                              "\nPress 'Ignore' to ignore this error. This might help, but "
                              "also lead to more errors.",
                              pszErrorString,
                              WinGetLastError(hab));
        else
            sprintf(szError2, "%s\nPress 'Cancel' to abort shutdown. ", pszErrorString);
        ulrc = ( (ULONG)(*fncbError)(0, 0, szError2, (MPARAM)ulOptions) );

    } else
        ulrc = (MBID_ABORT);

    if (hfLog) {
        CHAR szLog[2000];
        sprintf(szLog, "    Error occured: %s\n    Return code: %s (%d)\n    WinGetLastError: %lX\n",
                pszErrorString,
                    (ulrc == MBID_ABORT) ? "MBID_ABORT"
                    : (ulrc == MBID_IGNORE) ? "MBID_IGNORE"
                    : (ulrc == MBID_RETRY) ? "MBID_RETRY"
                    : "unknown",
                ulrc,
                WinGetLastError(hab));
        DosWriteToLogFile(hfLog, szLog);
    }
    return (ulrc);
}

/*
 * cmnINIError2:
 *      like the previous func, but with additional info
 *      for the error string.
 */

ULONG cmnINIError2(ULONG ulOptions, PSZ pszINI, HAB hab, HFILE hfLog, PFNWP fncbError, PSZ pszErrorString)
{
    CHAR szError2[2000];
    sprintf(szError2, "An error occured copying the profile %s: \n%s",
            pszINI, pszErrorString);
    return (cmnINIError(ulOptions, hab, hfLog, fncbError, szError2));
}

/*
 * PrfCopyProfile:
 *      this function copies a given profile into a
 *      new file. hOld is the handle, pszOld the filename
 *      of the profile to be copied (e.g. HINI_USERPROFILE
 *      and "?:\OS2\OS2.INI"; pszNew must point to a
 *      buffer which will contain the new filename ending
 *      in ".XFL" after PrfCopyProfile returns
 *      (e.g. "?:\OS2\OS2.XFL").
 *      You may specify a Callback procedure which gets
 *      called upon every copied application in the INI
 *      file; this way, you can provide a progress bar.
 *      fncbError gets called upon errors.
 */

BOOL PrfCopyProfile(HAB hab,
                    HFILE hfLog,            // logfile or 0
                    HINI hOld,              // profile to be copied
                    PSZ pszOld,             // its filename
                    PSZ pszNew,             // new filename
                    PFNWP fncbUpdate,       // update callback
                    HWND hwnd, ULONG msg, ULONG ulCount, ULONG ulMax,
                                            // params for fncbUpdate
                    PFNWP fncbError,        // error callback
                    PULONG pulFunc2)
{
    // BOOL    rc = TRUE;
    HINI    hNew;
    PSZ     pApps, pKeys, pData,
            pApp2, pKey2;
    ULONG   ulAppsSize, ulKeysSize, ulDataSize;
    CHAR    szLog[1000];
    ULONG   ulErrorOccured = MBID_IGNORE;
    PSZ     p_;
    CHAR    szError2[2000];

    *pulFunc2 = 20000;
    if (hfLog) {
        sprintf(szLog, "    Entering PrfCopyProfile, params: "
                    "hiniOld 0x%lX, pszOld %s, pfnwpCallback 0x%lX, "
                    "hwnd 0x%lX, msg 0x%lX, ulCount 0x%lX, ulMax 0x%lX\n",
                    hOld, pszOld, fncbUpdate, hwnd, msg, ulCount, ulMax);
        DosWriteToLogFile(hfLog, szLog);
    }

    *pulFunc2 = 20010;
    if ( (!pszOld) || (!pszNew) )
        ulErrorOccured = cmnINIError2(MB_CANCEL, pszOld, hab, hfLog, fncbError,
                "Invalid name buffers given.");

    *pulFunc2 = 20020;
    if (ulErrorOccured == MBID_IGNORE) {
        strupr(pszOld);
        strcpy(pszNew, pszOld);
        p_ = strstr(pszNew, ".INI");
        if (!p_)
            ulErrorOccured = cmnINIError2(MB_CANCEL, pszNew, hab, hfLog, fncbError,
                    "Error composing new filename.");
    }

    *pulFunc2 = 20030;
    if (ulErrorOccured == MBID_IGNORE) {
        strcpy(p_, ".XFL");
        if (hfLog) {
            sprintf(szLog, "    DEL %s\n", pszNew);
            DosWriteToLogFile(hfLog, szLog);
        }
        DosDelete(pszNew);

        if (hfLog) {
            sprintf(szLog, "    PrfOpenProfile %s\n", pszNew);
            DosWriteToLogFile(hfLog, szLog);
        }

        *pulFunc2 = 20040;
        hNew = PrfOpenProfile(hab, pszNew);
        if (!hNew)
            ulErrorOccured = cmnINIError2(MB_CANCEL, pszNew, hab, hfLog, fncbError,
                    "Error creating temporary profile.");
    }

    *pulFunc2 = 20100;
    if (ulErrorOccured == MBID_IGNORE) {
        if (!PrfQueryProfileSize(hOld, NULL, NULL, &ulAppsSize))
            ulErrorOccured = cmnINIError2(MB_CANCEL, pszNew, hab, hfLog, fncbError,
                    "Error creating temporary profile.");
        else
        if (ulAppsSize == 0)
            ulErrorOccured = cmnINIError2(MB_CANCEL, pszNew, hab, hfLog, fncbError,
                "Size of application buffer cannot be 0.");
    }

    if (ulErrorOccured == MBID_IGNORE) {

        do {
            ulErrorOccured = MBID_IGNORE; // if we have a retry in the do-while loop

            *pulFunc2 = 20110;
            if (ulErrorOccured == MBID_IGNORE) {
                pApps = malloc(ulAppsSize);
                pApp2 = pApps;
                if (!PrfQueryProfileData(hOld, NULL, NULL, pApps, &ulAppsSize))
                    ulErrorOccured = cmnINIError2(MB_CANCEL, pszNew, hab, hfLog, fncbError,
                                "Could not query application list.");
            }

            // applications loop

            *pulFunc2 = 20200;
            while (   (*pApp2 != 0)
                   && (ulErrorOccured == MBID_IGNORE)
                  )
            {
                *pulFunc2 = 20210;
                if (hfLog) {
                    sprintf(szLog, "    Copying app %s... ", pApp2);
                    DosWriteToLogFile(hfLog, szLog);
                }

                // get size of keys list for pApp2
                *pulFunc2 = 20220;
                if (!PrfQueryProfileSize(hOld, pApp2, NULL, &ulKeysSize)) {
                    sprintf(szError2, "Error querying size of keys list for %s.",
                            pApp2);
                    ulErrorOccured = cmnINIError2(MB_ABORTRETRYIGNORE, pszNew, hab, hfLog, fncbError,
                            szError2);
                }
                else if (ulAppsSize == 0) {
                    sprintf(szError2, "Size of keys list for %s should not be 0.",
                            pApp2);
                    ulErrorOccured = cmnINIError2(MB_ABORTRETRYIGNORE, pszNew, hab, hfLog, fncbError,
                                szError2);
                }

                if (ulErrorOccured == MBID_IGNORE) {
                    *pulFunc2 = 20250;
                    pKeys = malloc(ulKeysSize);
                    pKey2 = pKeys;
                    if (!PrfQueryProfileData(hOld, pApp2, NULL, pKeys, &ulKeysSize)) {
                        sprintf(szError2, "Error querying keys list for %s (size: %d bytes).",
                                pApp2, ulKeysSize);
                        ulErrorOccured = cmnINIError2(MB_ABORTRETRYIGNORE, pszNew, hab, hfLog, fncbError,
                                szError2);
                    }

                    // keys loop

                    *pulFunc2 = 20300;
                    while (   (*pKey2 != 0)
                           && (ulErrorOccured == MBID_IGNORE)
                          )
                    {
                        *pulFunc2 = 20310;
                        if (!PrfQueryProfileSize(hOld, pApp2, pKey2, &ulDataSize)) {
                            sprintf(szError2, "Error querying data size of %s/%s.",
                                    pApp2, pKey2);
                            ulErrorOccured = cmnINIError2(MB_ABORTRETRYIGNORE, pszNew, hab, hfLog, fncbError,
                                    szError2);
                        }
                        else if (ulDataSize == 0) {
                            sprintf(szError2, "Size of %s/%s should not be 0.",
                                    pApp2, pKey2);
                            ulErrorOccured = cmnINIError2(MB_ABORTRETRYIGNORE, pszNew, hab, hfLog, fncbError,
                                    szError2);
                        }

                        if (ulErrorOccured == MBID_IGNORE) {

                            // finally, copy data

                            *pulFunc2 = 20400;
                            pData = malloc(ulDataSize);
                            if (!pData) {
                                sprintf(szError2, "Error allocating memory for data from key %s (app %s, size %d bytes).",
                                        pKey2, pApp2, ulDataSize);
                                ulErrorOccured = cmnINIError2(MB_ABORTRETRYIGNORE, pszNew, hab, hfLog, fncbError,
                                        szError2);
                            }

                            *pulFunc2 = 20410;
                            if (    (!PrfQueryProfileData(hOld, pApp2, pKey2, pData, &ulDataSize))
                                 // || (TRUE) // *** debug!!!
                               )
                            {
                                sprintf(szError2, "Error reading data from key %s (app %s, size %d bytes).",
                                        pKey2, pApp2, ulDataSize);
                                ulErrorOccured = cmnINIError2(MB_ABORTRETRYIGNORE, pszNew, hab, hfLog, fncbError,
                                        szError2);
                            }

                            if (ulErrorOccured == MBID_IGNORE) {
                                *pulFunc2 = 20420;
                                if (!PrfWriteProfileData(hNew, pApp2, pKey2, pData, ulDataSize)) {
                                    sprintf(szError2, "Error writing data to %s/%s (size %d bytes).",
                                            pApp2, pKey2, ulDataSize);
                                    ulErrorOccured = cmnINIError2(MB_ABORTRETRYIGNORE, pszNew, hab, hfLog, fncbError,
                                            szError2);
                                }
                            }

                            *pulFunc2 = 20430;
                            if (pData)
                                free(pData);
                        }

                        // go for next key
                        *pulFunc2 = 20490;
                        pKey2 += strlen(pKey2)+1;
                        *pulFunc2 = 20491;
                    } // end while (*pKey2 != 0)

                    *pulFunc2 = 20500;
                    if (pKeys)
                        free(pKeys);

                    if (ulErrorOccured == MBID_IGNORE) {
                        *pulFunc2 = 20510;
                        if (hfLog)
                            DosWriteToLogFile(hfLog, "OK\n");
                    }

                } // end if (ulErrorOccured == MBID_IGNORE)

                *pulFunc2 = 20520;
                if (ulErrorOccured == MBID_IGNORE)
                    if (fncbUpdate) {
                        ULONG ulNow2, ulMax2;
                        ulNow2 = ((1000*(pApp2-pApps))/ulAppsSize) + (ulCount*1000);
                        ulMax2 = (ulMax+1)*1000; // ulAppsSize;
                        (*fncbUpdate)(hwnd, msg, (MPARAM)ulNow2, (MPARAM)ulMax2);
                    }

                *pulFunc2 = 20530;
                // go for next app
                pApp2 += strlen(pApp2)+1;
                *pulFunc2 = 20531;
            } // end while (*pApp2 != 0)

            if (ulErrorOccured == MBID_IGNORE) {
                *pulFunc2 = 20600;
                if (hfLog) {
                    sprintf(szLog, "    Done copying apps\n");
                    DosWriteToLogFile(hfLog, szLog);
                }
            }

            *pulFunc2 = 20610;
            if (pApps)
                free(pApps);

            if (hfLog) {
                sprintf(szLog, "    Done copying apps\n");
                DosWriteToLogFile(hfLog, szLog);
            }
        } while (ulErrorOccured == MBID_RETRY);
    } // end if (ulErrorOccured == MBID_IGNORE)

    *pulFunc2 = 20620;
    if (hfLog) {
        sprintf(szLog, "    PrfCloseProfile %s\n", pszNew);
        DosWriteToLogFile(hfLog, szLog);
    }

    *pulFunc2 = 20630;
    PrfCloseProfile(hNew);
    *pulFunc2 = 20640;
    if (fncbUpdate)
        (*fncbUpdate)(hwnd, msg, (MPARAM)(ulCount+1), (MPARAM)(ulMax+1));

    *pulFunc2 = 20650;
    if (hfLog)
        DosWriteToLogFile(hfLog, "    Exiting PrfCopyProfile.\n");

    *pulFunc2 = 20660;

    return (ulErrorOccured == MBID_IGNORE); // FALSE if error occured
}

/*
 * PrfSaveINIs:
 *      this will enforce saving of OS2.INI and OS2SYS.INI
 *      by calling PrfCopyProfile (above) on them; the therefrom
 *      resulting ".XFL" files will replace the original
 *      ".INI" files.
 *      Specify fncbUpdate and fncbError like in PrfCopyProfile.
 */

APIRET PrfSaveINIs(HAB hab, HFILE hfLog,
        PFNWP fncbUpdate, HWND hwnd, ULONG msg,
        PFNWP fncbError, PULONG pulFunc2)
{
    PRFPROFILE Profiles;
    APIRET  arc = NO_ERROR, arc2;
    BOOL    brc = TRUE;

    // FILESTATUS3 fs3;
    CHAR    szSysNew[CCHMAXPATH],
            szUserNew[CCHMAXPATH],
            szSysBackup[CCHMAXPATH],
            szUserBackup[CCHMAXPATH],
            szLog[1000];

    // PSZ     p;

    // the following flag may be one of the following:
    //        MBID_IGNORE --- everything's fine, continue
    //                        (or ignore error)
    //        MBID_RETRY  --- error occured, but retry
    //        MBID_ABORT  --- error occured, abort saving
    ULONG   ulErrorOccured = MBID_IGNORE;

    // DosEnterCritSec();

    *pulFunc2 = 10000;

    if (hfLog)
        DosWriteToLogFile(hfLog, "  Entering PrfSaveINIs...\n");

    Profiles.cchUserName = Profiles.cchSysName = 0;
    brc = PrfQueryProfile(hab, &Profiles);

    *pulFunc2 = 10010;
    if (!brc) {
        ulErrorOccured = cmnINIError(MB_CANCEL, hab, hfLog, fncbError,
                "Error querying system profiles size.");
    }

    *pulFunc2 = 10020;
    if (ulErrorOccured == MBID_IGNORE) {
        Profiles.pszUserName  = malloc(Profiles.cchUserName);
        Profiles.pszSysName  = malloc(Profiles.cchSysName);
        if (    (Profiles.pszSysName == NULL)
             || (Profiles.pszUserName == NULL)
           )
            ulErrorOccured = cmnINIError(MB_CANCEL, hab, hfLog, fncbError,
                    "Error allocating memory (1).");
    }

    *pulFunc2 = 10030;
    if (ulErrorOccured == MBID_IGNORE) {
        if (!PrfQueryProfile(hab, &Profiles))
            ulErrorOccured = cmnINIError(MB_CANCEL, hab, hfLog, fncbError,
                    "Error querying profiles (2).");
    }

    *pulFunc2 = 10040;
    if (ulErrorOccured == MBID_IGNORE) {
        if (hfLog) {
            sprintf(szLog, "  System profiles are %s, %s\n",
                    Profiles.pszUserName, Profiles.pszSysName);
            DosWriteToLogFile(hfLog, szLog);
        }

        // create filenames
        strcpy(szSysBackup, Profiles.pszSysName);
        strcpy(strstr(szSysBackup, ".INI"), ".BAK");
        strcpy(szUserBackup, Profiles.pszUserName);
        strcpy(strstr(szUserBackup, ".INI"), ".BAK");

        *pulFunc2 = 10050;
        // create OS2SYS.XFL
        if (hfLog) {
            sprintf(szLog, "  Storing %s in *.XFL\n",
                    Profiles.pszSysName);
            DosWriteToLogFile(hfLog, szLog);
        }
        if (!PrfCopyProfile(hab, hfLog, HINI_SYSTEMPROFILE,
                Profiles.pszSysName, szSysNew,
                fncbUpdate, hwnd, msg, 0, 1,
                fncbError, pulFunc2))
        {
            // abort, since PrfCopyProfile already has error handling
            ulErrorOccured = MBID_ABORT;
        }
    }

    *pulFunc2 = 10060;
    if (ulErrorOccured == MBID_IGNORE) {
        // create OS2.XFL
        if (hfLog) {
            sprintf(szLog, "  Storing %s in *.XFL\n",
                    Profiles.pszUserName);
            DosWriteToLogFile(hfLog, szLog);
        }
        if (!PrfCopyProfile(hab, hfLog,
                HINI_USERPROFILE, Profiles.pszUserName, szUserNew,
                fncbUpdate, hwnd, msg, 1, 1,
                fncbError, pulFunc2))
        {
            // abort, since PrfCopyProfile already has error handling
            ulErrorOccured = MBID_ABORT;
        }
    }

    /*
     * renaming stuff for OS2SYS.INI
     *
     */

    if (ulErrorOccured == MBID_IGNORE) {
        do {
            ulErrorOccured = MBID_IGNORE; // if we have a retry in the do-while loop

            *pulFunc2 = 10070;
            if (ulErrorOccured == MBID_IGNORE) {
                // attrib -r -s -h -a OS2SYS.BAK
                if (hfLog) {
                    sprintf(szLog, "  ATTRIB -R -S -H -A %s\n",
                            szSysBackup);
                    DosWriteToLogFile(hfLog, szLog);
                }
                arc2 = DosSetPathAttr(szSysBackup, FILE_NORMAL);
                if (hfLog) {
                    sprintf(szLog, "    rc2: %d\n", arc2);
                    DosWriteToLogFile(hfLog, szLog);
                }

                // delete OS2SYS.BAK
                if (hfLog) {
                    sprintf(szLog, "  DEL %s\n",
                            szSysBackup);
                    DosWriteToLogFile(hfLog, szLog);
                }
                arc2 = DosDelete(szSysBackup);
                if (hfLog) {
                    sprintf(szLog, "    rc2: %d\n", arc2);
                    DosWriteToLogFile(hfLog, szLog);
                }

                // attrib -r -s -h -a OS2SYS.INI
                if (hfLog) {
                    sprintf(szLog, "  ATTRIB -R -S -H -A %s\n",
                            Profiles.pszSysName);
                    DosWriteToLogFile(hfLog, szLog);
                }
                arc2 = DosSetPathAttr(Profiles.pszSysName, FILE_NORMAL);
                if (hfLog) {
                    sprintf(szLog, "    rc2: %d\n", arc2);
                    DosWriteToLogFile(hfLog, szLog);
                }

                // move OS2SYS.INI OS2SYS.BAK
                if (hfLog) {
                    sprintf(szLog, "  MOVE %s %s\n",
                            Profiles.pszSysName, szSysBackup);
                    DosWriteToLogFile(hfLog, szLog);
                }
                arc = DosMove(Profiles.pszSysName, szSysBackup);
                if (hfLog) {
                    sprintf(szLog, "    rc: %d\n", arc);
                    DosWriteToLogFile(hfLog, szLog);
                }

                if (arc)
                    ulErrorOccured = cmnINIError(MB_ABORTRETRYIGNORE, hab, hfLog, fncbError, "Error moving original system profile to backup.");
            }

            *pulFunc2 = 10200;
            if (ulErrorOccured == MBID_IGNORE) {
                if (hfLog) {
                    sprintf(szLog, "  MOVE %s %s\n",
                            szSysNew, Profiles.pszSysName);
                    DosWriteToLogFile(hfLog, szLog);
                }
                *pulFunc2 = 10210;
                arc = DosMove(szSysNew, Profiles.pszSysName);
                if (hfLog) {
                    sprintf(szLog, "    rc: %d\n", arc);
                    DosWriteToLogFile(hfLog, szLog);
                }
                if (arc)
                    ulErrorOccured = cmnINIError(MB_ABORTRETRYIGNORE, hab, hfLog, fncbError, "Error moving newly created profile to system profile.");
            }
        } while (ulErrorOccured == MBID_RETRY);
    }

    /*
     * renaming stuff for OS2.INI
     *
     */

    if (ulErrorOccured == MBID_IGNORE) {
        do {
            ulErrorOccured = MBID_IGNORE; // if we have a retry in the do-while loop

            *pulFunc2 = 10300;
            if (ulErrorOccured == MBID_IGNORE) {
                // attrib -r -s -h -a OS2.BAK
                if (hfLog) {
                    sprintf(szLog, "  ATTRIB -R -S -H -A %s\n",
                            szUserBackup);
                    DosWriteToLogFile(hfLog, szLog);
                }
                arc2 = DosSetPathAttr(szUserBackup, FILE_NORMAL);
                if (hfLog) {
                    sprintf(szLog, "    rc2: %d\n", arc2);
                    DosWriteToLogFile(hfLog, szLog);
                }

                // delete OS2.BAK
                *pulFunc2 = 10310;
                if (hfLog) {
                    sprintf(szLog, "  DEL %s\n",
                            szUserBackup);
                    DosWriteToLogFile(hfLog, szLog);
                }
                arc2 = DosDelete(szUserBackup);
                if (hfLog) {
                    sprintf(szLog, "    rc2: %d\n", arc2);
                    DosWriteToLogFile(hfLog, szLog);
                }

                // attrib -r -s -h -a OS2.INI
                if (hfLog) {
                    sprintf(szLog, "  ATTRIB -R -S -H -A %s\n",
                            Profiles.pszUserName);
                    DosWriteToLogFile(hfLog, szLog);
                }
                arc2 = DosSetPathAttr(Profiles.pszUserName, FILE_NORMAL);
                if (hfLog) {
                    sprintf(szLog, "    rc2: %d\n", arc2);
                    DosWriteToLogFile(hfLog, szLog);
                }

                // move OS2.INI OS2.BAK
                if (hfLog) {
                    sprintf(szLog, "  MOVE %s %s\n",
                            Profiles.pszUserName, szUserBackup);
                    DosWriteToLogFile(hfLog, szLog);
                }
                arc = DosMove(Profiles.pszUserName, szUserBackup);
                if (hfLog) {
                    sprintf(szLog, "    rc: %d\n", arc);
                    DosWriteToLogFile(hfLog, szLog);
                }

                if (arc)
                    ulErrorOccured = cmnINIError(MB_ABORTRETRYIGNORE, hab, hfLog, fncbError, "Error moving original user profile to backup.");
            }

            *pulFunc2 = 10350;
            if (ulErrorOccured == MBID_IGNORE) {
                // move OS2.XFL OS2.INI
                if (hfLog) {
                    sprintf(szLog, "  MOVE %s %s\n",
                            szUserNew, Profiles.pszUserName);
                    DosWriteToLogFile(hfLog, szLog);
                }
                arc = DosMove(szUserNew, Profiles.pszUserName);
                if (hfLog) {
                    sprintf(szLog, "    rc: %d\n", arc);
                    DosWriteToLogFile(hfLog, szLog);
                }

                if (arc)
                    ulErrorOccured = cmnINIError(MB_ABORTRETRYIGNORE, hab, hfLog, fncbError, "Error moving newly created profile to user profile.");
            }
        } while (ulErrorOccured == MBID_RETRY);
    }

    // DosExitCritSec();

    *pulFunc2 = 10400;
    if (ulErrorOccured != MBID_IGNORE) {
        DosMove(szSysBackup, Profiles.pszSysName);
        DosMove(szUserBackup, Profiles.pszUserName);
    }

    *pulFunc2 = 10410;
    if (    (Profiles.pszSysName)
         && (Profiles.pszUserName)
       )
    {
        *pulFunc2 = 10420;
        free(Profiles.pszSysName);
        *pulFunc2 = 10430;
        free(Profiles.pszUserName);
    }

    *pulFunc2 = 10430;
    if (hfLog) {
        DosWriteToLogFile(hfLog, "  Done with PrfSaveINIs\n");
    }

    if (ulErrorOccured != MBID_IGNORE)
        return (999);
    else
        return (NO_ERROR);
}

/*
 * PrfQueryColor:
 *      returns a system color in OS2.INI's PM_Colors as a LONG
 */

LONG PrfQueryColor(PSZ pszKeyName, PSZ pszDefault)
{
    CHAR szColor[30];
    LONG r, g, b;
    PrfQueryProfileString(
                HINI_USER,
                "PM_Colors",
                pszKeyName,
                pszDefault,
                szColor,
                sizeof(szColor)-1);
    sscanf(szColor, "%d %d %d ", &r, &g, &b);
    return (r*0x10000 + g*0x100 + b);
}

/*
 * WinInsertMenuItem:
 *      this inserts one one menu item into a given menu; parameters:
 *          HWND hwndMenu       the menu to add the submenu to
 *          ULONG iPosition     the index at which to insert
 *          SHORT sItemId       menu ID of the new item
 *          PSZ pszItemTitle    title of this item
 */

SHORT WinInsertMenuItem(HWND hwndMenu, SHORT iPosition,
            SHORT sItemId, PSZ pszItemTitle,
            SHORT afStyle, SHORT afAttr)
{
    MENUITEM mi;
    SHORT    src = MIT_ERROR;

    mi.iPosition = iPosition;
    mi.afStyle = afStyle;
    mi.afAttribute = afAttr;
    mi.id = sItemId;
    mi.hwndSubMenu = 0;
    mi.hItem = 0;
    src = (SHORT)WinSendMsg(hwndMenu, MM_INSERTITEM, (MPARAM)&mi, (MPARAM)pszItemTitle);
    return (src);
}

/*
 * WinInsertSubmenu:
 *      this inserts a submenu into a given menu and inserts one item
 *      into this new submenu; parameters:
 *          HWND hwndMenu       the menu to add the submenu to
 *          ULONG iPosition     the index at which to insert
 *          SHORT sMenuId       menu ID of the new menu
 *          PSZ pszSubmenuTitle title of the submenu
 *          USHORT afMenuStyle  style flags of the submenu; MIS_SUBMENU will
 *                              always be set by this func
 *          USHORT sItemId      menu ID of the first item in the new submenu
 *          PSZ pszItemTitle    title of this item
 *          USHORT afItemStyle  style flags of this item, e.g. MIS_TEXT
 *          USHORT afAttribute  attributes of this item, e.g. MIA_DISABLED
 *      Returns the HWND of the new submenu.
 */

HWND WinInsertSubmenu(HWND hwndMenu, ULONG iPosition,
                 SHORT sMenuId, PSZ pszSubmenuTitle, USHORT afMenuStyle,
                 SHORT sItemId, PSZ pszItemTitle, USHORT afItemStyle, USHORT afAttribute)
{
    MENUITEM mi; //, miItem;
    SHORT    src = MIT_ERROR;
    HWND     hwndNewMenu;

    hwndNewMenu = WinCreateMenu(hwndMenu, NULL);
    if (hwndNewMenu)
    {
        mi.iPosition = iPosition;
        mi.afStyle = afMenuStyle | MIS_SUBMENU;
        mi.afAttribute = 0;
        mi.id = sMenuId;
        mi.hwndSubMenu = hwndNewMenu;
        mi.hItem = 0;
        src = (SHORT)WinSendMsg(hwndMenu, MM_INSERTITEM, (MPARAM)&mi, (MPARAM)pszSubmenuTitle);
        if (    (src != MIT_MEMERROR)
            &&  (src != MIT_ERROR)
           )
        {
            WinSetWindowUShort(hwndNewMenu, QWS_ID, sMenuId);

            mi.iPosition = 0;
            mi.afStyle = afItemStyle;
            mi.afAttribute = afAttribute;
            mi.id = sItemId;
            mi.hwndSubMenu = 0;
            mi.hItem = 0;
            WinSendMsg(hwndNewMenu, MM_INSERTITEM, (MPARAM)&mi, (MPARAM)pszItemTitle);
        }
    }
    return (hwndNewMenu);
}

/*
 * WinInsertMenuSeparator:
 *      this inserts a separator into a given menu at
 *      the given position (which may be MIT_END);
 *      returns the position at which the item was
 *      inserted
 */

SHORT WinInsertMenuSeparator(HWND hMenu, SHORT iPosition, SHORT sId)
{
    MENUITEM mi;
    mi.iPosition = iPosition;
    mi.afStyle = MIS_SEPARATOR;             // append separator
    mi.afAttribute = 0;
    mi.id = sId;
    mi.hwndSubMenu = 0;
    mi.hItem = 0;

    return ((SHORT)WinSendMsg(hMenu,
                    MM_INSERTITEM,
                    (MPARAM)&mi,
                    (MPARAM)""));
}

/*
 * WinMenuRemoveEllipse:
 *      removes a "..." substring from a menu item
 *      title, if found
 */

VOID WinMenuRemoveEllipse(HWND hwndMenu, USHORT usItemId)
{
    CHAR szBuf[255];
    CHAR *p;
    WinSendMsg(hwndMenu, MM_QUERYITEMTEXT,
            MPFROM2SHORT(usItemId, sizeof(szBuf)-1),
            (MPARAM)&szBuf);
    if (p = strstr(szBuf, "..."))
        strcpy(p, p+3);
    WinSendMsg(hwndMenu, MM_SETITEMTEXT,
            MPFROMSHORT(usItemId),
            (MPARAM)&szBuf);
}

/*
 * WinSetDlgItemSpinData:
 *          sets a spin button's data within a dialog window
 */

VOID WinSetDlgItemSpinData(HWND hwndDlg,
                           ULONG idSpinButton,
                           ULONG min,       // minimum allowed value
                           ULONG max,       // maximum allowed value
                           ULONG current)   // current value
{
    HWND hwndSpinButton = WinWindowFromID(hwndDlg, idSpinButton);
    if (hwndSpinButton) {
        WinSendMsg(hwndSpinButton,
                      SPBM_SETLIMITS,          // Set limits message
                      (MPARAM)max,             // Spin Button maximum setting
                      (MPARAM)min);             // Spin Button minimum setting

        WinSendMsg(hwndSpinButton,
                      SPBM_SETCURRENTVALUE,    // Set current value message
                      (MPARAM)current,
                      (MPARAM)NULL);
    }
}

/*
 * WinCnrScrollToRecord:
 *      scrolls a given container control to make a given
 *      record visible; returns TRUE if actual scrolling was
 *      performed, FALSE if not neccessary or on errors.
 *      Parameters:
 *          HWND hwndCnr     container window
 *          PRECORDCORE precc the record to scroll to; this can also
 *                           be a PMINIRECORDCORE, we don't care
 *          ULONG fsExtent   determines the record core coordinates to
 *                           take into account (ORed):
 *                           CMA_ICON    icon rectangle
 *                           CMA_TEXT    text rectangle
 *                           CMA_TREEICON "plus" sign in tree view
 *          BOOL KeepParent  for tree views only: keep parent record of
 *                           the record visible; in case pRec is not
 *                           currently visible, we will scroll so that
 *                           the parent record is at the top of the
 *                           visible records (Win95 style)
 *      Returns:
 *          0           scrolled
 *          >0          not scrolled for the following reason:
 *                  1       record rectangle query failed (error)
 *                  2       cnr viewport query failed (error)
 *                  3       record is already visible
 *                  4       cnrinfo query failed (error)
 *                  5       parent record rectangle query failed (error)
 *      This function an improved version of the one (W)(C) Dan Libby, found at
 *      http://zebra.asta.fh-weingarten.de/os2/Snippets/Howt6364.HTML
 *      Improvements (C) 1998 Ulrich Mller.
 */

ULONG WinCnrScrollToRecord(HWND hwndCnr, PRECORDCORE pRec, ULONG fsExtent, BOOL KeepParent)
{
    QUERYRECORDRECT qRect, qRect2;
    RECTL           rclRecord, rclParentRecord, rclCnr, rclCnr2;
    POINTL          ptlRecord, ptlParentRecord;
    CNRINFO         CnrInfo;
    HAB             hab = WinQueryAnchorBlock(hwndCnr);
    BOOL            KeepParent2;
    LONG            lYOfs;

    qRect.cb = sizeof(qRect);
    qRect.pRecord = (PRECORDCORE)pRec;
    qRect.fsExtent = fsExtent;

    // query record location and size of container
    if (!WinSendMsg(hwndCnr, CM_QUERYRECORDRECT, &rclRecord, &qRect))
        return 1;
    if (!WinSendMsg(hwndCnr, CM_QUERYVIEWPORTRECT, &rclCnr, MPFROM2SHORT(CMA_WINDOW, FALSE)) )
        return 2;

    // check if left bottom point of pRec is currently visible in container
    ptlRecord.x = (rclRecord.xLeft);
    ptlRecord.y = (rclRecord.yBottom);
    // ptlRecord.x = (rclRecord.xLeft + rclRecord.xRight) / 2;
    // ptlRecord.y = (rclRecord.yBottom + rclRecord.yTop) / 2;
    if (WinPtInRect(hab, &rclCnr, &ptlRecord))
         return 3;

    if (KeepParent)
        if (!WinSendMsg(hwndCnr, CM_QUERYCNRINFO, (MPARAM)&CnrInfo, (MPARAM)sizeof(CnrInfo)))
            return 4;
        else KeepParent2 = (CnrInfo.flWindowAttr & CV_TREE);
    else KeepParent2 = FALSE;

    // calculate offset to scroll to make pRec visible
    lYOfs = (rclCnr.yBottom - rclRecord.yBottom)    // this would suffice
          + (rclRecord.yTop - rclRecord.yBottom);  // but we make the next rcl visible too

    if (KeepParent2) {
        qRect2.cb = sizeof(qRect2);
        qRect2.pRecord = (PRECORDCORE)WinSendMsg(hwndCnr, CM_QUERYRECORD,
                (MPARAM)pRec, MPFROM2SHORT(CMA_PARENT, CMA_ITEMORDER));
        qRect2.fsExtent = fsExtent;

        // now query PARENT record location and size of container
        if (!WinSendMsg(hwndCnr, CM_QUERYRECORDRECT, &rclParentRecord, &qRect2))
            return 5;

        ptlParentRecord.x = (rclParentRecord.xLeft);
        ptlParentRecord.y = (rclParentRecord.yTop);
        // ptlParentRecord.x = (rclParentRecord.xLeft + rclParentRecord.xRight) / 2;
        // ptlParentRecord.y = (rclParentRecord.yBottom + rclParentRecord.yTop) / 2;
        rclCnr2 = rclCnr;
        WinOffsetRect(hab, &rclCnr2, 0, -lYOfs);
        // if ( (rclParentRecord.yBottom - rclRecord.yBottom) > (rclCnr.yTop - rclCnr.yBottom) )
        if (!(WinPtInRect(hab, &rclCnr2, &ptlParentRecord)))
        {
            lYOfs = (rclCnr.yTop - rclParentRecord.yTop) // this would suffice
                  - (rclRecord.yTop - rclRecord.yBottom);  // but we make the previous rcl visible too
        }
    }

    if (!KeepParent2)
        WinPostMsg(hwndCnr, CM_SCROLLWINDOW, (MPARAM)CMA_HORIZONTAL,
            (MPARAM)(rclRecord.xLeft - rclCnr.xLeft));
    WinPostMsg(hwndCnr, CM_SCROLLWINDOW,
        (MPARAM)CMA_VERTICAL, (MPARAM)lYOfs);

    return 0;
}

/*
 * WinCnrAllocRecord:
 *      this is a shortcut to allocating memory for
 *      a container record. Returns the recc.
 *          ULONG cbrecc    total size of your record core
 *                          or sizeof(RECORDCORE), if you have
 *                          no additional recc data
 */

PRECORDCORE WinCnrAllocRecord(HWND hwndCnr, ULONG cbrecc)
{
    PRECORDCORE      precc;
    precc = (PRECORDCORE)WinSendMsg(hwndCnr, CM_ALLOCRECORD,
            (MPARAM)(cbrecc-sizeof(RECORDCORE)),
            (MPARAM)1);
    precc->cb = cbrecc;
    return (precc);
}

/*
 * WinCnrInsertRecord:
 *      shortcut to inserting a previously allocated recc
 *      into a container. This func returns the recordcore.
 *      Params:
 *          HWND hwndCnr    the container window
 *          PRECORDCORE preccParent
 *                          for tree views: parent record
 *          PSZ pszText     text for record
 *          ULONG flRecordAttr
 *                  recc flags, as explained in pmstddlg.h:
 *                          CRA_SELECTED       record is selected
 *                          CRA_TARGET         record has target emphasis
 *                          CRA_CURSORED       cursor is on the record
 *                          CRA_INUSE          record has in-use emphasis
 *                          CRA_FILTERED       record has been filtered
 *                          CRA_DROPONABLE     record can be dropped on
 *                          CRA_RECORDREADONLY record is read-only
 *                          CRA_EXPANDED       record is expanded
 *                          CRA_COLLAPSED      record is collapsed
 *                          CRA_PICKED         record picked (Lazy Drag)
 *                  plus the following undocumented:
 *                          CRA_IGNORE
 *                          CRA_SOURCE
 *                          CRA_DISABLED
 *                          CRA_OWNERFREE
 *                          CRA_OWNERDRAW
 */

PRECORDCORE WinCnrInsertRecord(HWND hwndCnr, PRECORDCORE preccParent, PRECORDCORE precc,
            PSZ pszText, ULONG flRecordAttr)
{
    RECORDINSERT     ri;

    if (precc) {
        // RECORDCORE stuff
        precc->flRecordAttr = flRecordAttr;
        precc->preccNextRecord = NULL;
        precc->pszIcon = pszText;
        precc->pszText = pszText;
        precc->pszName = pszText;
        precc->pszTree = pszText;
        precc->hptrIcon = NULLHANDLE;

        // setup RECORDINSERT struct
        ri.cb = sizeof(RECORDINSERT);
        ri.pRecordOrder = (PRECORDCORE)CMA_FIRST;
        ri.pRecordParent = (PRECORDCORE)preccParent;
        ri.zOrder = CMA_TOP;
        ri.fInvalidateRecord = TRUE;
        ri.cRecordsInsert = 1;

        // printf("  Sending CM_INSERTRECORD\n");
        WinSendMsg(hwndCnr, CM_INSERTRECORD, (MPARAM)precc, (MPARAM)&ri);
    }
    return (precc);
}

/*
 * WinCenterWindow:
 *      centers a window within its parent window.
 *      The window should not be visible to avoid flickering.
 */

void WinCenterWindow(HWND hwnd)
{
   RECTL rclParent;
   RECTL rclWindow;

   WinQueryWindowRect(hwnd, &rclWindow);
   WinQueryWindowRect(WinQueryWindow(hwnd, QW_PARENT), &rclParent);

   rclWindow.xLeft   = (rclParent.xRight - rclWindow.xRight) / 2;
   rclWindow.yBottom = (rclParent.yTop   - rclWindow.yTop  ) / 2;

   WinSetWindowPos(hwnd, NULLHANDLE, rclWindow.xLeft, rclWindow.yBottom,
                    0, 0, SWP_MOVE);
}

/*
 * WinCenteredDlgBox:
 *      just like WinDlgBox, but the dlg box is centered on the screen;
 *      you should mark the dlg template as not visible in the dlg
 *      editor, or display will flicker
 */

ULONG WinCenteredDlgBox(HWND hwndParent, HWND hwndOwner,
              PFNWP pfnDlgProc, HMODULE hmod, ULONG idDlg, PVOID pCreateParams)
{
    ULONG   ulReply;
    HWND    hwndDlg = WinLoadDlg(hwndParent, hwndOwner, pfnDlgProc,
                                hmod, idDlg, pCreateParams);
    WinCenterWindow(hwndDlg);
    ulReply = WinProcessDlg(hwndDlg);
    WinDestroyWindow(hwndDlg);
    return (ulReply);
}

/*
 * WinQueryPresColor:
 *      returns the specified color. This is queried in the
 *      following order:
 *      1)  hwnd's pres params are searched for ulPP;
 *      2)  if this fails, PM_Colors in OS2.INI is searched
 *          for pszKeyName;
 *      3)  if this fails also, pszDefault is evaluated.
 *      The return value is always an RGB LONG.
 *      The following ulPP / pszKeyName pairs are useful:
 *          PP_FOREGROUNDCOLOR             "WindowText"
 *             Foreground color
 *          PP_BACKGROUNDCOLOR             "Window" or "DialogBackground"
 *             Background color
 *          PP_HILITEFOREGROUNDCOLOR       "HiliteForeground" or "MenuHiliteText"
 *             Highlighted foreground color, for example for selected menu
 *          PP_HILITEBACKGROUNDCOLOR       "HiliteBackground" or "MenuHilite"
 *             Highlighted background color
 *          PP_DISABLEDFOREGROUNDCOLOR
 *             Disabled foreground color
 *          PP_DISABLEDBACKGROUNDCOLOR
 *             Disabled background color
 *          PP_BORDERCOLOR                 "WindowFrame"
 *             Border color
 *          PP_ACTIVECOLOR
 *             Active color
 *          PP_INACTIVECOLOR
 *             Inactive color
 *          PP_ACTIVETEXTFGNDCOLOR
 *             Active text foreground color
 *          PP_ACTIVETEXTBGNDCOLOR
 *             Active text background color
 *          PP_INACTIVETEXTFGNDCOLOR
 *             Inactive text foreground color
 *          PP_INACTIVETEXTBGNDCOLOR
 *             Inactive text background color
 */

LONG WinQueryPresColor(HWND hwnd, ULONG ulPP, PSZ pszKeyName, PSZ pszDefault)
{
    ULONG ul, attrFound, abValue[32];
    if (ulPP != (ULONG)-1)
        if (ul = WinQueryPresParam(hwnd,
                    ulPP,
                    0,
                    &attrFound,
                    (ULONG)sizeof(abValue),
                    (PVOID)&abValue,
                    0))
            return (abValue[0]);

    return (PrfQueryColor(pszKeyName, pszDefault));
}

/*
 * WinCreateFakeDesktop:
 *      this routine creates and displays a frameless window over
 *      the whole screen to fool the user that all windows
 *      have been closed (which in fact might not be the
 *      case); the window background color is set to the Desktop's
 *      (PM's one, not the WPS's one); returns the hwnd of this window
 */

HWND WinCreateFakeDesktop(HWND hwndSibling)
{
    typedef struct _BACKGROUND {
        ULONG   cb;         //  Length of the aparam parameter, in bytes.
        ULONG   id;     //  Attribute type identity.
        ULONG   cb2;     //  Byte count of the ab parameter.
        RGB     rgb;    //  Attribute value.
     } BACKGROUND;
    RECTL       R;
    BACKGROUND  background;
    ULONG       flStyle;          // window style
    LONG        lDesktopColor;

    /* create fake desktop window = empty window with
       the size of full screen */
    lDesktopColor = WinQuerySysColor(HWND_DESKTOP,
        SYSCLR_BACKGROUND, 0);
    flStyle = WS_VISIBLE;
    R.xLeft      = 0;
    R.yBottom    = 0;
    R.xRight     = WinQuerySysValue(HWND_DESKTOP,SV_CXSCREEN);
    R.yTop       = WinQuerySysValue(HWND_DESKTOP,SV_CYSCREEN);
    background.cb = sizeof(background.id)
                  + sizeof(background.cb)
                  + sizeof(background.rgb);
    background.id = PP_BACKGROUNDCOLOR;
    background.cb2 = sizeof(RGB);
    background.rgb.bBlue = (CHAR1FROMMP(lDesktopColor));
    background.rgb.bGreen= (CHAR2FROMMP(lDesktopColor));
    background.rgb.bRed  = (CHAR3FROMMP(lDesktopColor));

    return (WinCreateWindow(HWND_DESKTOP,  // parent window
                           WC_FRAME,      // class name
                           "",            // window text
                           flStyle,       // window style
                           0, 0,          // position (x,y)
                           R.xRight,
                           R.yTop,
                           NULLHANDLE,    // owner window
                           hwndSibling,    // sibling window
                           1,             // window id
                           NULL,          // control data
                           &background)); // presentation parms
}

/*
 * WinQueryCnrFromFrame:
 *         find the handle of a frame window's container; we do this
 *         by enumerating all the frame's child windows and comparing
 *         their window classes to the WC_CONTAINER code;
 *         returns NULLHANDLE if not found
 */

HWND WinQueryCnrFromFrame(HWND hwndFrame)
{
    HENUM               henum;
    CHAR                szClassName[256];
    HWND                hwndCnr = NULLHANDLE,
                        hwndTemp = NULLHANDLE;

    if (hwndFrame) {
        henum = WinBeginEnumWindows(hwndFrame);
        if (henum) {
            do {
                hwndTemp = WinGetNextWindow(henum);
                if (hwndTemp) {
                    if (WinQueryClassName(hwndTemp, 250, szClassName))
                        if (strcmp(szClassName, "#37") == 0)
                            // unbelievable, this is PM's code for WC_CONTAINER...
                            hwndCnr = hwndTemp;
                }
            } while (hwndTemp);
        }
        WinEndEnumWindows(henum);
    }

    return hwndCnr;
}

/*
 * WinQueryItemUnderMouse:
 *      this queries the menu item which corresponds
 *      to the given mouse coordinates
 */

SHORT WinQueryItemUnderMouse(HWND hwndMenu, POINTL *pptlMouse, RECTL *prtlItem)
{
    SHORT   s, sItemId, sItemCount;
    HAB     habDesktop = WinQueryAnchorBlock(HWND_DESKTOP);

    sItemCount = (SHORT)WinSendMsg(hwndMenu, MM_QUERYITEMCOUNT, MPNULL, MPNULL);

    for (s = 0;
         s <= sItemCount;
         s++)
    {
        sItemId = (SHORT)WinSendMsg(hwndMenu, MM_ITEMIDFROMPOSITION, (MPARAM)s, MPNULL);
        WinSendMsg(hwndMenu, MM_QUERYITEMRECT,
            MPFROM2SHORT(sItemId, FALSE),
            (MPARAM)prtlItem);
        if (WinPtInRect(habDesktop, prtlItem, pptlMouse))
            return (sItemId);
    }
    /* sItemId = (SHORT)WinSendMsg(hwndMenu, MM_ITEMIDFROMPOSITION, (MPARAM)(sItemCount-1), MPNULL);
    return (sItemId); */
    return (-1); // error: no valid menu item
}

/*
 * WinDrawFormattedText:
 *      this func takes a rectangle and draws pszText into
 *      it, breaking the words as neccessary. The line spacing
 *      is determined from the currently selected font.
 *      Parameters:
 *          HPS hps         the presentation space, whose settings are
 *                          not altered;
 *          PRECTL prcl     the rectangle to draw text in; the y coords
 *                          are modified and return the space actually
 *                          used for drawing
 *          PSZ pszText     the text to draw;
 *          flCmd0          may be the ORed flags like in WinDrawText. Tested
 *                          are: DT_TOP, DT_LEFT.
 *                          DT_WORDBREAK | DT_TEXTATTRS are automatically
 *                          set.
 *                          If you specify DT_QUERYEXTENT, no drawing is performed,
 *                          but only the above prcl calculation.
 *      Returns the number of lines drawn.
 */

ULONG WinDrawFormattedText(HPS hps, PRECTL prcl, PSZ pszText, ULONG flCmd)
{
    PSZ     p = pszText;
    LONG    lDrawn = 1,
            lTotalDrawn = 0,
            lLineCount = 0,
            lOrigYTop = prcl->yTop;
    ULONG   ulTextLen = strlen(pszText),
            ulCharHeight,
            flCmd2;
    // POINTL  aptlText[TXTBOX_COUNT];
    RECTL   rcl2;

    // printf("Entering WinDrawFormattedText\n");

    flCmd2 = flCmd | DT_WORDBREAK | DT_TEXTATTRS;

    ulCharHeight = GpiQueryLineSpacing(hps, pszText);
    // _Pmpf(("  ulCharHeight: %d\n", ulCharHeight));

    while (   (lDrawn)
           && (lTotalDrawn < ulTextLen)
          )
    {
        memcpy(&rcl2, prcl, sizeof(rcl2));
        lDrawn = WinDrawText(hps,
                ulTextLen-lTotalDrawn,
                p,
                &rcl2,
                0, 0,                       // colors
                flCmd2
            );
        p += lDrawn;
        lTotalDrawn += lDrawn;
        prcl->yTop -= ulCharHeight;
        lLineCount++;
        // printf("  lDrawn: %d, lTotalDrawn: %d\n", lDrawn, lTotalDrawn);
    }
    // printf("  LineCount: %d\n", lLineCount);
    prcl->yBottom = prcl->yTop;
    prcl->yTop = lOrigYTop;

    return (lLineCount);
}

/*
 * GpiQueryLineSpacing:
 *      this returns the optimal line spacing for text
 *      output with the current HPS; this is computed
 *      by evaluating those incredible FONTMETRICS
 */

LONG GpiQueryLineSpacing(HPS hps, PSZ pszText)
{
    FONTMETRICS fm;
    POINTL aptlText[TXTBOX_COUNT];
    GpiQueryTextBox(hps, strlen(pszText), pszText,
             TXTBOX_COUNT, (PPOINTL)&aptlText);

    if (GpiQueryFontMetrics(hps, sizeof(FONTMETRICS), &fm))
        return ( (  fm.lMaxBaselineExt     // max vertical font space
                   +fm.lExternalLeading)    // space advised by font designer
                 // * 12 / 10
               );
    else
        return (15);
}

/*
 * GpiDraw3DFrame:
 *      this draws a rectl in 3D style with a given line width;
 *      if (fRaised), the 3D line is given a "raised" look,
 *      "sunken" otherwise
 */

VOID GpiDraw3DFrame(HPS hps, PRECTL prcl, USHORT usWidth, BOOL fRaised)
{
    RECTL rcl2 = *prcl;
    LONG lColorLeft, lColorRight;
    USHORT us;
    POINTL ptl1;

    if (fRaised) {
        lColorLeft = WinQuerySysColor(HWND_DESKTOP, SYSCLR_BUTTONLIGHT, 0);
        lColorRight = WinQuerySysColor(HWND_DESKTOP, SYSCLR_BUTTONDARK, 0);
    } else {
        lColorLeft = WinQuerySysColor(HWND_DESKTOP, SYSCLR_BUTTONDARK, 0);
        lColorRight = WinQuerySysColor(HWND_DESKTOP, SYSCLR_BUTTONLIGHT, 0);
    }

    for (us = 0; us < usWidth; us++)
    {
        GpiSetColor(hps, lColorLeft);
        ptl1.x = rcl2.xLeft;
        ptl1.y = rcl2.yBottom;
        GpiMove(hps, &ptl1);
        ptl1.y = rcl2.yTop-1;
        GpiLine(hps, &ptl1);
        ptl1.x = rcl2.xRight-1;
        GpiLine(hps, &ptl1);
        GpiSetColor(hps, lColorRight);
        ptl1.y = rcl2.yBottom;
        GpiLine(hps, &ptl1);
        ptl1.x = rcl2.xLeft;
        GpiLine(hps, &ptl1);

        rcl2.xLeft++;
        rcl2.yBottom++;
        rcl2.xRight--;
        rcl2.yTop--;
    }
}

/*
 * GpiCreateBmpFromPS:
 *      this creates a new bitmap, which is compatible with the
 *      device associated with hpsScreen, and then copies the
 *      rectangle prcl into it; returns the handle of the bitmap,
 *      which can then be used for WinDrawBitmap and such
 */

HBITMAP GpiCreateBmpFromPS(HAB hab, HPS hpsScreen, PRECTL prcl)
{

    /* To copy an image from a display screen to a bit map:
      1. Associate the memory device context with a presentation space.
      2. Create a bit map.
      3. Select the bit map into the memory device context by calling GpiSetBitmap.
      4. Determine the location (in device coordinates) of the image.
      5. Call GpiBitBlt and copy the image to the bit map. */

    HDC hdcMem;
    PSZ pszData[4] = { "Display", NULL, NULL, NULL };
    HPS hpsMem;
    SIZEL sizlPage = {0, 0};
    BITMAPINFOHEADER2 bmp;
    PBITMAPINFO2 pbmi = NULL;
    HBITMAP hbm = NULLHANDLE;
    POINTL aptl[3];
    LONG alData[2];

    // create the memory device context and presentation space so they
    // are compatible with the screen device context and presentation space.

    if (hdcMem = DevOpenDC(hab, OD_MEMORY, "*", 4, (PDEVOPENDATA)pszData, NULLHANDLE))
    {
        if (hpsMem = GpiCreatePS(hab, hdcMem, &sizlPage, PU_PELS | GPIA_ASSOC | GPIT_MICRO))
        {
            // determine the device's plane/bit-count format.
            if (GpiQueryDeviceBitmapFormats(hpsMem, 2, alData))
            {
                // set up the BITMAPINFOHEADER2 and BITMAPINFO2 structures
                bmp.cbFix = (ULONG)sizeof(BITMAPINFOHEADER2);
                bmp.cx = (prcl->xRight - prcl->xLeft);
                bmp.cy = (prcl->yTop - prcl->yBottom);
                bmp.cPlanes = alData[0];
                bmp.cBitCount = alData[1];
                bmp.ulCompression = BCA_UNCOMP;
                bmp.cbImage = (((bmp.cx *
                    (1 << bmp.cPlanes) * (1 << bmp.cBitCount)) + 31) / 32) * bmp.cy;
                bmp.cxResolution = 70;
                bmp.cyResolution = 70;
                bmp.cclrUsed = 2;
                bmp.cclrImportant = 0;
                bmp.usUnits = BRU_METRIC;
                bmp.usReserved = 0;
                bmp.usRecording = BRA_BOTTOMUP;
                bmp.usRendering = BRH_NOTHALFTONED;
                bmp.cSize1 = 0;
                bmp.cSize2 = 0;
                bmp.ulColorEncoding = BCE_RGB;
                bmp.ulIdentifier = 0;

                if (DosAllocMem((PPVOID)&pbmi, sizeof(BITMAPINFO2) +
                    (sizeof(RGB2) * (1 << bmp.cPlanes) * (1 << bmp.cBitCount)),
                    PAG_COMMIT | PAG_READ | PAG_WRITE) == NO_ERROR)
                {
                    pbmi->cbFix = bmp.cbFix;
                    pbmi->cx = bmp.cx;
                    pbmi->cy = bmp.cy;
                    pbmi->cPlanes = bmp.cPlanes;
                    pbmi->cBitCount = bmp.cBitCount;
                    pbmi->ulCompression = BCA_UNCOMP;
                    pbmi->cbImage = ((bmp.cx+31)/32) * bmp.cy;
                    pbmi->cxResolution = 70;
                    pbmi->cyResolution = 70;
                    pbmi->cclrUsed = 2;
                    pbmi->cclrImportant = 0;
                    pbmi->usUnits = BRU_METRIC;
                    pbmi->usReserved = 0;
                    pbmi->usRecording = BRA_BOTTOMUP;
                    pbmi->usRendering = BRH_NOTHALFTONED;
                    pbmi->cSize1 = 0;
                    pbmi->cSize2 = 0;
                    pbmi->ulColorEncoding = BCE_RGB;
                    pbmi->ulIdentifier = 0;

                    // Create a bit map that is compatible with the display.
                    if (hbm = GpiCreateBitmap(hpsMem, &bmp, FALSE, NULL, pbmi))
                    {
                        // Associate the bit map and the memory presentation space.
                        if (GpiSetBitmap(hpsMem, hbm) != HBM_ERROR) {
                            // Copy the screen to the bit map.
                            aptl[0].x = 0;              // lower-left corner of destination rectangle
                            aptl[0].y = 0;
                            aptl[1].x = prcl->xRight;   // upper-right corner for both
                            aptl[1].y = prcl->yTop;
                            aptl[2].x = prcl->xLeft;    // lower-left corner of source rectangle
                            aptl[2].y = prcl->yBottom;

                            // hps = WinGetScreenPS(HWND_DESKTOP);

                            GpiBitBlt(hpsMem, hpsScreen,
                                sizeof(aptl) / sizeof(POINTL), // Number of points in aptl
                                aptl, ROP_SRCCOPY, BBO_IGNORE);
                        }
                    } // end if (hbm = )
                    // WinReleasePS(hps);
                } // end if (DosAllocMem((
            } // end if if (GpiQueryDeviceBitmapFormats(

            GpiDestroyPS(hpsMem);
        } // end if (hpsMem = GpiCreatePS(
        DevCloseDC(hdcMem);
    } // end if (hdcMem = DevOpenDC(

    return (hbm);
}

/*
 * GpiPowerOffAnim:
 *      displays an animation that looks like a
 *      monitor being turned off; hps must have
 *      been acquired using WinGetScreenPS,
 *      ulSteps should be around 40-50
 */

#define LAST_WIDTH 2
#define LAST_STEPS 50
#define WAIT_TIME  10

VOID GpiPowerOffAnim(HPS hps, ULONG ulSteps)
{
    RECTL       rclScreen, rclNow, rclLast, rclDraw;
    ULONG       ul = ulSteps;
    ULONG       ulPhase = 1;

    WinQueryWindowRect(HWND_DESKTOP, &rclScreen);

    WinShowPointer(HWND_DESKTOP, FALSE);

    memcpy(&rclLast, &rclScreen, sizeof(RECTL));
    // WinFillRect(hps, &rclScreen, CLR_BLUE);
    ul = 0;
    ulPhase = 1;
    do {
        ULONG ulFromTime = DosGetULongTime();
        if (ulPhase == 1)
        {
            rclNow.xLeft = ((rclScreen.yTop / 2) * ul / ulSteps );
            rclNow.xRight = (rclScreen.xRight) - rclNow.xLeft;
            rclNow.yBottom = ((rclScreen.yTop / 2) * ul / ulSteps );
            rclNow.yTop = (rclScreen.yTop) - rclNow.yBottom;

            if (rclNow.yBottom > (rclNow.yTop - LAST_WIDTH) ) {
                rclNow.yBottom = (rclScreen.yTop / 2) - LAST_WIDTH;
                rclNow.yTop = (rclScreen.yTop / 2) + LAST_WIDTH;
            }

            // top part
            rclDraw.xLeft = rclLast.xLeft;
            rclDraw.xRight = rclLast.xRight;
            rclDraw.yBottom = rclNow.yTop;
            rclDraw.yTop = rclLast.yTop;
            WinFillRect(hps, &rclDraw, CLR_BLACK);

            // left part
            rclDraw.xLeft = rclLast.xLeft;
            rclDraw.xRight = rclNow.xLeft;
            rclDraw.yBottom = rclLast.yBottom;
            rclDraw.yTop = rclLast.yTop;
            WinFillRect(hps, &rclDraw, CLR_BLACK);

            // right part
            rclDraw.xLeft = rclNow.xRight;
            rclDraw.xRight = rclLast.xRight;
            rclDraw.yBottom = rclLast.yBottom;
            rclDraw.yTop = rclLast.yTop;
            WinFillRect(hps, &rclDraw, CLR_BLACK);

            // bottom part
            rclDraw.xLeft = rclLast.xLeft;
            rclDraw.xRight = rclLast.xRight;
            rclDraw.yBottom = rclLast.yBottom;
            rclDraw.yTop = rclNow.yBottom;
            WinFillRect(hps, &rclDraw, CLR_BLACK);

            memcpy(&rclLast, &rclNow, sizeof(RECTL));

            if ( rclNow.xRight < ((rclScreen.xRight / 2) + LAST_WIDTH) ) {
                ulPhase = 2; // exit
            }
        } else if (ulPhase == 2) {
            rclDraw.xLeft = (rclScreen.xRight / 2) - LAST_WIDTH;
            rclDraw.xRight = (rclScreen.xRight / 2) + LAST_WIDTH;
            rclDraw.yBottom = (rclScreen.yTop * 1 / 4);
            rclDraw.yTop = (rclScreen.yTop * 3 / 4);
            WinFillRect(hps, &rclDraw, CLR_WHITE);

            ulPhase = 3;
            ul = 0;
        } else if (ulPhase == 3) {
            rclDraw.xLeft = (rclScreen.xRight / 2) - LAST_WIDTH;
            rclDraw.xRight = (rclScreen.xRight / 2) + LAST_WIDTH;
            rclDraw.yTop = (rclScreen.yTop * 3 / 4);
            rclDraw.yBottom = (rclScreen.yTop * 3 / 4) - ((rclScreen.yTop * 1 / 4) * ul / LAST_STEPS);
            if (rclDraw.yBottom < ((rclScreen.yTop / 2) + LAST_WIDTH))
                rclDraw.yBottom = ((rclScreen.yTop / 2) + LAST_WIDTH);
            WinFillRect(hps, &rclDraw, CLR_BLACK);

            rclDraw.xLeft = (rclScreen.xRight / 2) - LAST_WIDTH;
            rclDraw.xRight = (rclScreen.xRight / 2) + LAST_WIDTH;
            rclDraw.yBottom = (rclScreen.yTop * 1 / 4);
            rclDraw.yTop = (rclScreen.yTop * 1 / 4) + ((rclScreen.yTop * 1 / 4) * ul / LAST_STEPS);
            if (rclDraw.yTop > ((rclScreen.yTop / 2) - LAST_WIDTH))
                rclDraw.yBottom = ((rclScreen.yTop / 2) - LAST_WIDTH);

            WinFillRect(hps, &rclDraw, CLR_BLACK);

            ul++;
            if (ul > LAST_STEPS) {
                ulPhase = 99;
            }
        }

        ul++;

        while (DosGetULongTime() < ulFromTime + WAIT_TIME) {
            // PSZ p = NULL; // keep compiler happy
        }
    } while (ulPhase != 99);

    DosSleep(500);
    WinFillRect(hps, &rclScreen, CLR_BLACK);
    DosSleep(500);

    WinShowPointer(HWND_DESKTOP, TRUE);
}

/********************************************************************
 *                                                                  *
 *   Common routines for progress bars                              *
 *                                                                  *
 ********************************************************************/

/*
 * fnwpProgressBar:
 *      this is the window procedure for the progress bar control,
 *      which is a subclassed static rectangle control; we need
 *      to capture WM_PAINT to draw the progress bar according to
 *      shutdown's progress, and we also update the static text field
 *      (percentage) next to it
 */

/* PROGRESSBARDATA:
   structure for progress bar data, saved at QWL_USER window ulong */
typedef struct _PROGRESSBARDATA
{
    ULONG      ulNow, ulMax, ulPaintX, ulOldPaintX;
    ULONG      ulAttr;
    PFNWP      OldStaticProc;
    RECTL      rtlBar;
} PROGRESSBARDATA, *PPROGRESSBARDATA;

MRESULT EXPENTRY fnwpProgressBar(HWND hwndBar, ULONG msg, MPARAM mp1, MPARAM mp2)
{
    HPS     hps;
    POINTL  ptl1, ptlText, aptlText[TXTBOX_COUNT];
    RECTL   rcl, rcl2;

    CHAR    szPercent[10] = "";

    PPROGRESSBARDATA    pData = (PPROGRESSBARDATA)WinQueryWindowULong(hwndBar, QWL_USER);

    PFNWP   OldStaticProc = NULL;

    MRESULT mrc = NULL;

    if (pData) {
        OldStaticProc = pData->OldStaticProc;

        switch(msg) {

            case WM_UPDATEPROGRESSBAR: {
                if (    (pData->ulNow != (ULONG)mp1)
                     || (pData->ulMax != (ULONG)mp2)
                   )
                {
                    pData->ulNow = (ULONG)mp1;
                    pData->ulMax = (ULONG)mp2;
                    // and continue with WM_PAINT
                }
                else break;

                // check validity
                if (pData->ulNow > pData->ulMax)
                    pData->ulNow = pData->ulMax;
                // avoid division by zero
                if (pData->ulMax == 0) {
                    pData->ulMax = 1;
                    pData->ulNow = 0;
                }

                pData->ulPaintX =
                    (ULONG)(
                              (    (ULONG)(pData->rtlBar.xRight-pData->rtlBar.xLeft)
                                 * (ULONG)(pData->ulNow)
                              )
                              /  (ULONG)pData->ulMax
                           );
                if (pData->ulPaintX != pData->ulOldPaintX)
                    goto wmpaint;
                // else continue
            break; }

            case WM_PAINT: {
                wmpaint:
                hps = WinBeginPaint(hwndBar, NULLHANDLE, &rcl);
                WinEndPaint(hps);

                hps = WinGetPS(hwndBar);

                // switch to RGB mode
                GpiCreateLogColorTable(hps, 0, LCOLF_RGB, 0, 0, NULL);

                if (pData->ulPaintX <= pData->ulOldPaintX)
                {
                    /* draw frame and background only if this is either
                       a "real" WM_PAINT (i.e., the window was overlapped
                       and needs repainting) or if ulNow has _de_creased */
                    GpiSetColor(hps, WinQuerySysColor(HWND_DESKTOP, SYSCLR_BUTTONDARK, 0));
                    ptl1.x = 0;
                    ptl1.y = 0;
                    GpiMove(hps, &ptl1);
                    ptl1.y = (pData->rtlBar.yTop);
                    GpiLine(hps, &ptl1);
                    ptl1.x = (pData->rtlBar.xRight);
                    GpiLine(hps, &ptl1);
                    GpiSetColor(hps, WinQuerySysColor(HWND_DESKTOP, SYSCLR_BUTTONLIGHT, 0));
                    ptl1.y = 0;
                    GpiLine(hps, &ptl1);
                    ptl1.x = 0;
                    GpiLine(hps, &ptl1);

                    pData->rtlBar.xLeft = 1;
                    pData->rtlBar.yBottom = 1;
                    WinFillRect(hps, &(pData->rtlBar),
                        WinQuerySysColor(HWND_DESKTOP, SYSCLR_SCROLLBAR, 0));
                }

                if (pData->ulAttr & PBA_PERCENTFLAGS)
                {
                    sprintf(szPercent, "%d %%", ((100 * pData->ulNow) / pData->ulMax) );
                    GpiQueryTextBox(hps, strlen(szPercent), szPercent,
                        TXTBOX_COUNT, (PPOINTL)&aptlText);
                    ptlText.x = pData->rtlBar.xLeft +
                                    (   (   (pData->rtlBar.xRight-pData->rtlBar.xLeft)
                                          - (aptlText[TXTBOX_BOTTOMRIGHT].x-aptlText[TXTBOX_BOTTOMLEFT].x)
                                        )
                                    / 2);
                    ptlText.y = 3 + pData->rtlBar.yBottom +
                                    (   (   (pData->rtlBar.yTop-pData->rtlBar.yBottom)
                                          - (aptlText[TXTBOX_TOPLEFT].y-aptlText[TXTBOX_BOTTOMLEFT].y)
                                        )
                                    / 2);
                    if (    ((ptlText.x + (aptlText[TXTBOX_BOTTOMRIGHT].x-aptlText[TXTBOX_BOTTOMLEFT].x))
                            > pData->ulPaintX)
                        &&  (pData->ulPaintX > pData->ulOldPaintX)
                       )
                    {
                        rcl.xLeft      = ptlText.x;
                        rcl.xRight     = ptlText.x + (aptlText[TXTBOX_BOTTOMRIGHT].x-aptlText[TXTBOX_BOTTOMLEFT].x);
                        rcl.yBottom    = ptlText.y;
                        rcl.yTop       = ptlText.y + (aptlText[TXTBOX_TOPLEFT].y-aptlText[TXTBOX_BOTTOMLEFT].y);
                        WinFillRect(hps, &rcl,
                            WinQuerySysColor(HWND_DESKTOP, SYSCLR_SCROLLBAR, 0));
                    }
                }

                rcl2.xLeft = pData->rtlBar.xLeft;
                rcl2.xRight = (pData->ulPaintX > (rcl2.xLeft + 3))
                        ? pData->ulPaintX
                        : rcl2.xLeft + ((pData->ulAttr & PBA_BUTTONSTYLE)
                                       ? 3 : 1);
                rcl2.yBottom = pData->rtlBar.yBottom;
                rcl2.yTop = pData->rtlBar.yTop-1;

                // now draw the actual progress
                if (pData->ulAttr & PBA_BUTTONSTYLE)
                {
                    RECTL rcl3 = rcl2;
                    // GpiDraw3DFrame(hps, &rclBar, 2, FALSE);
                    // draw "raised" inner rect
                    // WinInflateRect(WinQueryAnchorBlock(hwndBar), &rcl2, -1, -1);
                    rcl3.yTop++;
                    rcl3.xRight++;
                    GpiDraw3DFrame(hps, &rcl3, 2, TRUE);
                    WinInflateRect(WinQueryAnchorBlock(hwndBar), &rcl2, -2, -2);
                }

                if (rcl2.xRight > rcl2.xLeft) {
                    GpiSetColor(hps, WinQuerySysColor(HWND_DESKTOP,
                            // SYSCLR_HILITEBACKGROUND,
                            SYSCLR_BUTTONMIDDLE,
                        0));
                    ptl1.x = rcl2.xLeft;
                    ptl1.y = rcl2.yBottom;
                    GpiMove(hps, &ptl1);
                    ptl1.x = rcl2.xRight;
                    ptl1.y = rcl2.yTop;
                    GpiBox(hps, DRO_FILL | DRO_OUTLINE,
                        &ptl1,
                        0,
                        0);
                }

                // now print the percentage
                if (pData->ulAttr & PBA_PERCENTFLAGS)
                {
                    GpiMove(hps, &ptlText);
                    GpiSetColor(hps, WinQuerySysColor(HWND_DESKTOP,
                            // SYSCLR_HILITEFOREGROUND,
                            SYSCLR_BUTTONDEFAULT,
                        0));
                    GpiCharString(hps, strlen(szPercent), szPercent);
                    /* WinSetDlgItemText(hwndStatus, ID_SPDI_PERCENT,
                            szPercent); */
                }
                WinReleasePS(hps);

                pData->ulOldPaintX = pData->ulPaintX;
            break; }

            default:
                mrc = OldStaticProc(hwndBar, msg, mp1, mp2);
       }
    }
    return (mrc);
}

/*
 * WinProgressBarFromStatic:
 *      this function turns an existing static rectangle control
 *      into a progress bar by subclassing its window procedure;
 *      this way you can easily create a progress bar in the Dialog
 *      Editor; after loading the dlg template, simply call this
 *      function with the hwnd of the static control.
 *      In order to update the progress bar, simply post or send
 *      WM_UPDATEPROGRESSBAR to the static (= progress bar) window;
 *      this message is equal to WM_USER and needs the following
 *      parameters:
 *          mp1     ULONG ulNow, the current progress
 *          mp2     ULONG ulMax, the maximally possible progress
 *                               (= 100%)
 *      The progress bar automatically calculates the current progress
 *      display. For example, if ulMax == 8192 and ulNow == 4096,
 *      a progress of 50% will be shown. It is possible to change
 *      ulMax after the progress bar has started display. If ulMax
 *      is 0, a progress of 0% will be shown (to avoid dividision
 *      by zero traps).
 *
 *      ulAttr accepts of the following:
 *          PBA_NOPERCENTAGE    do not display percentage
 *          PBA_ALIGNLEFT       left-align percentage
 *          PBA_ALIGNRIGHT      right-align percentage
 *          PBA_ALIGNCENTER     center percentage
 *          PBA_BUTTONLOOK      no "flat", but button-like look
 */

BOOL WinProgressBarFromStatic(HWND hwndStatic, ULONG ulAttr)
{
    PFNWP OldStaticProc = WinSubclassWindow(hwndStatic, fnwpProgressBar);
    if (OldStaticProc) {
        PPROGRESSBARDATA pData = malloc(sizeof(PROGRESSBARDATA));
        pData->ulMax = 1;
        pData->ulNow = 0;
        pData->ulPaintX = 0;
        pData->ulAttr = ulAttr;
        pData->OldStaticProc = OldStaticProc;
        WinQueryWindowRect(hwndStatic, &(pData->rtlBar));

        WinSetWindowULong(hwndStatic, QWL_USER, (ULONG)pData);
        return (TRUE);
    }
    else return (FALSE);
}

/*
 * fnwpAnimatedIcon:
 *      subclassed wnd proc for static icons subclassed
 *      with WinAnimatedIconFromStatic (below)
 */

typedef struct _ANIMICONDATA
{
    PFNWP      OldStaticProc;
    HPOINTER   hptr;
    RECTL      rclIcon;
} ANIMICONDATA, *PANIMICONDATA;

MRESULT EXPENTRY fnwpAnimatedIcon(HWND hwndIcon, ULONG msg, MPARAM mp1, MPARAM mp2)
{
    HPS     hps;
    PANIMICONDATA pData = (PANIMICONDATA)WinQueryWindowULong(hwndIcon, QWL_USER);
    PFNWP   OldStaticProc = NULL;
    MRESULT mrc = NULL;

    if (pData) {
        OldStaticProc = pData->OldStaticProc;

        switch(msg) {
            case SM_SETHANDLE: {
                pData->hptr = (HPOINTER)mp1;
                hps = WinGetPS(hwndIcon);
                GpiCreateLogColorTable(hps, 0,
                    LCOLF_RGB,
                    0, 0, NULL);
                WinFillRect(hps, &(pData->rclIcon),
                    WinQuerySysColor(HWND_DESKTOP, SYSCLR_DIALOGBACKGROUND, 0));
                if (pData->hptr) {
                    WinDrawPointer(hps, 0, 0, pData->hptr, DP_NORMAL);
                }
                WinReleasePS(hps);
            break; }

            case WM_PAINT: {
                RECTL rcl;
                hps = WinBeginPaint(hwndIcon, NULLHANDLE, &rcl);
                if (pData->hptr) {
                    WinDrawPointer(hps, 0, 0, pData->hptr, DP_NORMAL);
                }
                WinEndPaint(hps);
            break; }

            default:
                mrc = OldStaticProc(hwndIcon, msg, mp1, mp2);
       }
    }
    return (mrc);
}

/*
 * WinAnimatedIconFromStatic:
 *      turns a static control into one which
 *      properly displays icons when given a
 *      SM_SETHANDLE msg; you should perform this
 *      function on all your animation icons, because
 *      regular static controls do not properly display
 *      transparency
 */

BOOL WinAnimatedIconFromStatic(HWND hwndStatic)
{
    PFNWP OldStaticProc = WinSubclassWindow(hwndStatic, fnwpAnimatedIcon);
    if (OldStaticProc) {
        PANIMICONDATA pData = malloc(sizeof(ANIMICONDATA));
        pData->OldStaticProc = OldStaticProc;
        WinQueryWindowRect(hwndStatic, &(pData->rclIcon));

        WinSetWindowULong(hwndStatic, QWL_USER, (ULONG)pData);
        return (TRUE);
    }
    else return (FALSE);
}

/*
 * DebugNum:
 *      displays a msgbox with a ULONG for debugging
 */

void DebugNum(PSZ pszTitle, ULONG ul)
{
    CHAR    szBuf[20] = "Nr: ";
    strhexcat(szBuf, ul);
    DebugBox(pszTitle, szBuf);
}


