/*
 * objdlgs.c:
 *      contains the logic to display the WPS class tree
 *      in a container control. See objdlgs.h for more
 *      explanations.
 *
 *      Also contains the logic for the notebook "Internals"
 *      page.
 *
 *      Copyright (C) 1997-98 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of the XFolder main distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

#include <malloc.h>

#define INCL_DOSPROCESS
#define INCL_DOSSEMAPHORES
#define INCL_DOSEXCEPTIONS

// pmwin.h
#define INCL_WINWINDOWMGR
#define INCL_WINSYS             // presparams
#define INCL_WINPOINTERS

#define INCL_WINMENUS
#define INCL_WINENTRYFIELDS
#define INCL_WINDIALOGS

// pmstddlg.h
#define INCL_WINSTDCNR          // container
#define INCL_WINSTDFILE         // file dialog

#define INCL_GPILOGCOLORTABLE
#define INCL_GPIPRIMITIVES
#include <os2.h>

#include "xfldr.ih"

#include <wpclsmgr.h>
#include <wpdesk.h>
#include <wpabs.h>

#include <repostry.h>      // SOM interface repository */
#include <containr.h>
#include <somapi.h>

#include "common.h"

#include "xfsort.h"        // XFolder container sort functions
#include "wphandles.h"     // Henk Kelder's object handles stuff
#include "procstat.h"      // Kai Uwe Rommel's DosQProc() stuff

#include "objdlgs.h"

// #define DEBUG_WPSCLASSES

/********************************************************************
 *                                                                  *
 *   "Select class" dialog                                          *
 *                                                                  *
 ********************************************************************/

/*
 * cmnAddClass2Cnr:
 *      this inserts one WPS class record core into a cnr tree by
 *      creating a new record core (CLASSRECORDCORE struct). The
 *      class to be inserted is described by a WPSLISTITEM structure.
 *      SELECTCLASSDATA is a "Select class" dialog description
 *      structure used throughout this file.
 *      All these structs are defined in objdlgs.h.
 *      Returns the newly created and inserted recordcore or NULL
 *      on errors.
 */

PCLASSRECORDCORE cmnAddClass2Cnr(HWND hwndCnr,          // cnr to insert into
                        PCLASSRECORDCORE preccParent,   // tree view parent recc
                        PSZ pszCurrentClass,            //
                        PWPSLISTITEM pwpsMyself,        // class to be inserted
                        PSELECTCLASSDATA pscd)          // dlg description
{
    PCLASSRECORDCORE      preccNew = (PCLASSRECORDCORE)
            WinCnrAllocRecord(hwndCnr, sizeof(CLASSRECORDCORE)); // helpers.c
    // recc attributes
    ULONG            usAttrs = CRA_EXPANDED | CRA_RECORDREADONLY;

    #ifdef DEBUG_WPSCLASSES
        _Pmpf(("  cmnAddClass2Cnr %s\n", pszCurrentClass));
    #endif

    if (preccNew) {
        // store the WPS class info structure into recc
        // set extra recc attribute (WPSLISTITEM)
        preccNew->pwps = pwpsMyself;
        if (pwpsMyself) {
            pwpsMyself->pRecord = (PRECORDCORE)preccNew;

            if (pscd)
                // now call callback function defined in
                // SELECTCLASSDATA to determine the recc
                // attributes; this way, a class can be
                // disabled, expanded etc.
                if (pscd->pfnwpReturnAttrForClass)
                    usAttrs = (USHORT)(*pscd->pfnwpReturnAttrForClass)(
                            hwndCnr, (ULONG)pscd,
                            (MPARAM)pwpsMyself, (MPARAM)preccParent);
                else {
                    // no callback defined: set read-only
                    usAttrs = CRA_RECORDREADONLY;
                    // and expand three tree view levels
                    if (pscd->lRecurse < 3)
                        usAttrs |= CRA_EXPANDED;
                }
        } else
            // pwpsMyself might be NULL for the "Orphans" tree parent item
            usAttrs = CRA_RECORDREADONLY;

        #ifdef DEBUG_WPSCLASSES
            _Pmpf(("    Inserting %s\n", pszCurrentClass));
        #endif
        // insert the record (helpers.c)
        WinCnrInsertRecord(hwndCnr, (PRECORDCORE)preccParent, (PRECORDCORE)preccNew,
                pszCurrentClass, usAttrs);

        // select the new record?
        if (usAttrs & (CRA_INUSE | CRA_CURSORED | CRA_PICKED | CRA_SELECTED)) {
            #ifdef DEBUG_WPSCLASSES
                _Pmpf(("    Selecting %s\n", pszCurrentClass));
            #endif
            WinSendMsg(hwndCnr, CM_SETRECORDEMPHASIS,
                    (MPARAM)preccNew,
                    MPFROM2SHORT(TRUE,
                            (USHORT)(usAttrs & (CRA_INUSE | CRA_CURSORED | CRA_PICKED | CRA_SELECTED)))
            );
        }

        // mark the class as processed (needed by cmnWPSClasses2Cnr)
        if (pwpsMyself)
            pwpsMyself->fProcessed = TRUE;
    }
    return (preccNew);
}

/*
 * cmnWPSClasses2Cnr:
 *      this func inserts a WPS class hierarchy starting with
 *      the class "pszCurrentClass" as the root class into
 *      a given container window. This cnr should be in tree
 *      view to have a meaningful display. This function can
 *      be used with any exising container window.
 *      Parameters:
 *          HWND hwndCnr        guess what this is
 *          PSZ pszCurrentClass the "root" class of the display,
 *                              e.g. "WPObject"; make sure this
 *                              class exists, or nothing occurs
 *          PWPSLISTITEM pliClasses2 must be NULL
 *          PCLASSRECORDCORE preccParent
 *                              the parent record core to insert
 *                              the tree in; if NULL, a new tree
 *                              is created (root level).
 *          PSELECTCLASSDATA pscd
 *                              dlg structure, may be NULL; if
 *                              pfnwpReturnAttrForClass in this
 *                              this structure is != NULL, it will
 *                              be called as a callback for each
 *                              item inserted (see typedef above)
 *      Returns:
 *          PWPSCLASSESINFO         structure which WPS data.
 *
 *      NOTE: This function eats up a lot of memory, depending
 *      on how many WPS classes are installed. This memory is NOT
 *      freed when this function returns, because the container
 *      still needs it. AFTER destroying the container window,
 *      you must issue cmnCleanupWpsClasses (below) with the
 *      return value of whis function. This will free the linked
 *      WPS class item list and free allocated SOM resources.
 */

PWPSCLASSESINFO cmnWpsClasses2Cnr(HWND hwndCnr, PSZ pszCurrentClass,
        PWPSLISTITEM pwpsClasses2, PCLASSRECORDCORE preccParent,
        PSELECTCLASSDATA pscd)
{
    PCLASSRECORDCORE precc;
    BOOL             fBufferAllocated = FALSE;
    BOOL             fRecursiveCall = FALSE;
    PWPSLISTITEM     pwps0, pwpsRoot, pwpsClasses = NULL, pwpsMyself = NULL;
    PWPSCLASSESINFO  pwpsciReturn = malloc(sizeof(WPSCLASSESINFO));
    POBJCLASS        pObjClass2;

    #ifdef DEBUG_WPSCLASSES
        _Pmpf(("Entering cmnWpsClasses2Cnr for %s...\n", pszCurrentClass));
    #endif

    if (pwpsClasses2 == NULL) {
        // pwpsClasses2 == NULL when this func is not called
        // recursively, i.e. this is the first invocation;
        // we will now set up all the neccessary data and
        // pass pwpsClasses2 to subsequent recursive calls
        ULONG ulSize;

        #ifdef DEBUG_WPSCLASSES
            _Pmpf(("  cmnWpsClasses2Cnr: Collecting classes\n"));
        #endif

        // get WPS class list
        WinEnumObjectClasses(NULL, &ulSize);
        pwpsciReturn->pObjClass = malloc(ulSize+1);
        WinEnumObjectClasses(pwpsciReturn->pObjClass, &ulSize);
        fBufferAllocated = TRUE;

        // now go through the buffer we obtained and query
        // more SOM information about the classes contained
        // therein; this information will be stored in a
        // linked list of WPSLISTITEM items
        pObjClass2 = pwpsciReturn->pObjClass;
        while (pObjClass2) {
            pwps0 = malloc(sizeof(WPSLISTITEM));
            memset(pwps0, 0, sizeof(WPSLISTITEM));

            // WPS class name
            pwps0->pszClassName = pObjClass2->pszClassName;

            // DLL containing the class
            pwps0->pszModName   = pObjClass2->pszModName;

            // get SOM id of the class; this will load the
            // class! So the id needs to be freed later
            // (using cmnCleanupWpsClasses)
            pwps0->somid = somIdFromString(pObjClass2->pszClassName);
            if (pwps0->somid) {
                // get pointer to metaclass (e.g. M_WPObject).
                // NOTE: for replaced WPS classes, somFindClass
                // returns a pointer to the replaced class!
                pwps0->pClassObject = _somFindClass(SOMClassMgrObject, pwps0->somid, 0, 0);
                if (pwps0->pClassObject) {

                    // now check if the class has been replaced
                    PSZ pszTrueClass = _somGetName(pwps0->pClassObject);
                    // get parent class; for replaced classes, this
                    // now points to the replaced class
                    SOMClass *pParentClass = _somGetParent(pwps0->pClassObject);
                    // PSZ pszParentClass = _somGetName(pParentClass);
                    if (pParentClass)
                        strcpy(pwps0->szParentClass, _somGetName(pParentClass));

                    if (strcmp(pwps0->pszClassName, pszTrueClass) == 0)
                        // not replaced
                        (pwps0->szReplacedByClass)[0] = '\0';
                    else {
                        // replaced
                        #ifdef DEBUG_WPSCLASSES
                            _Pmpf(("  Class %s has been replaced...\n", pObjClass2->pszClassName));
                        #endif
                        strcpy(pwps0->szReplacedByClass, pszTrueClass);

                        // "repair" class object: for replaced classes, the
                        // parent class points to itself (see above), so
                        // we call somGetParent again
                        pwps0->pClassObject = _somGetParent(pwps0->pClassObject);
                        pParentClass = _somGetParent(pwps0->pClassObject);
                        if (pParentClass) {
                            strcpy(pwps0->szParentClass, _somGetName(pParentClass));
                            #ifdef DEBUG_WPSCLASSES
                                _Pmpf(("      Real parent: %s\n", _somGetName(pParentClass)));
                            #endif
                        }
                    }
                }
            }
            pwps0->szReplacesClass[0] = '\0';
            // mark this list item as not processed; this flag
            // is needed later to avoid excessive recursing
            pwps0->fProcessed = FALSE;
            // append list item to list, even if the class has not been
            // successfully loaded
            lstAppendItem((PLISTITEM*)&pwpsClasses, NULL, (PLISTITEM)pwps0, 0, 0);
            // go for next class in WPS buffer
            pObjClass2 = pObjClass2->pNext;
        }
        pwpsRoot = pwpsClasses;
        pwpsciReturn->pwpsRoot = pwpsClasses;
        pscd->lRecurse = 0;

        // now we know for each class with which class it is replaced;
        // we now go through the list again to add for each class
        // the reverse, i.e. which class it _replaces_

        // printf("  Building reverse replacement list...\n");
        pwps0 = pwpsRoot;
        while (pwps0) {
            if (strlen(pwps0->szReplacedByClass)) {
                PWPSLISTITEM     pwps2 = pwpsRoot;
                /* printf("    Found %s, replaced by %s\n",
                                pwps0->pszClassName, pwps0->szReplacedByClass); */
                while (pwps2) {
                    if (strcmp(pwps2->pszClassName, pwps0->szReplacedByClass) == 0) {
                        strcpy(pwps2->szReplacesClass, pwps0->pszClassName);
                        /* printf("      %s replace %s\n",
                                        pwps2->pszClassName, pwps2->szReplacesClass); */
                        break;
                    }
                    pwps2 = pwps2->pNext;
                }
            }
            pwps0 = pwps0->pNext;
        }
    } else {
        fRecursiveCall = TRUE;
        pwpsRoot = pwpsClasses2;
        (pscd->lRecurse)++;
    }

    #ifdef DEBUG_WPSCLASSES
        _Pmpf(("  Searching myself...\n", pszCurrentClass));
    #endif
    // now find ourselves (pszCurrentClass) in the linked list;
    // for the first invocation, this is specified by the caller,
    // when recursing, this is a subclass of the initially
    // specified class
    pwps0 = pwpsRoot;
    while (pwps0) {
        if (pwps0->pClassObject)
            if ((pwps0->pszClassName) && (pszCurrentClass))
                if (strcmp(pwps0->pszClassName, pszCurrentClass) == 0) {
                    pwpsMyself = pwps0;
                    break;
                }
        pwps0 = pwps0->pNext;
    }

    #ifdef DEBUG_WPSCLASSES
        _Pmpf(("  Processing myself...\n", pszCurrentClass));
    #endif
    if (pwpsMyself)
        if (!(pwpsMyself->fProcessed))
        {
            // list item found for pszCurrentClass
            // and not yet processed:
            // add record core to container
            #ifdef DEBUG_WPSCLASSES
                _Pmpf(("  Adding myself...\n"));
            #endif
            precc = cmnAddClass2Cnr(hwndCnr, preccParent,
                        pszCurrentClass, pwpsMyself,
                        pscd);

            if (pwpsMyself->pClassObject) {
                // now go through the linked list again from
                // the beginnning and call ourselves recursively
                // for all classes which are direct children
                // of our class
                #ifdef DEBUG_WPSCLASSES
                    _Pmpf(("  Looking for my children...\n"));
                #endif
                pwps0 = pwpsRoot;
                while (pwps0) {
                    if (strcmp(pwps0->szParentClass, pszCurrentClass) == 0) {
                        // recurse!
                        cmnWpsClasses2Cnr(hwndCnr, pwps0->pszClassName,
                                pwpsRoot, precc,
                                pscd);
                    }
                    pwps0 = pwps0->pNext;
                }
            }
        }

    if (fBufferAllocated) {
        // only if non-recursive call (root level): go through
        // the list again and add all the classes we
        // haven't added yet ("orphaned" classes w/out parents);
        // this is only done if pscd->pszOrphans has been set
        if (pscd->pszOrphans) {
            PCLASSRECORDCORE preccOrphans = NULL;
            #ifdef DEBUG_WPSCLASSES
                _Pmpf(("  Adding orphans...\n"));
            #endif
            pwps0 = pwpsRoot;
            while (pwps0) {
                if (!pwps0->fProcessed) {
                    if (pwps0->pszClassName)
                    {
                        #ifdef DEBUG_WPSCLASSES
                            _Pmpf(("  Calling cmnAddClass2Cnr...\n"));
                        #endif
                        if (preccOrphans == NULL) {
                            preccOrphans = cmnAddClass2Cnr(hwndCnr, NULL,
                                    pscd->pszOrphans, NULL,
                                    pscd);
                        }
                        cmnAddClass2Cnr(hwndCnr, preccOrphans, pwps0->pszClassName, pwps0,
                                pscd);
                    }
                }
                pwps0 = pwps0->pNext;
            }
        }
    } else
        // recursing: decrease recursion level
        (pscd->lRecurse)--;

    #ifdef DEBUG_WPSCLASSES
        _Pmpf(("Exiting cmnWpsClasses2Cnr.\n"));
    #endif
    return (pwpsciReturn);
}

/*
 * cmnCleanupWpsClasses:
 *      this cleans up the resources allocated by
 *      cmnWpsClasses2Cnr (memory and SOM stuff);
 *      pwpsClasses2 must be the pointer returned
 *      by cmnWpsClasses2Cnr
 */

VOID cmnCleanupWpsClasses(PWPSCLASSESINFO pwpsci)
{
    PWPSLISTITEM pItem, pItem2;

    pItem = pwpsci->pwpsRoot;
    while (pItem) {
        if (pItem->pClassObject)
            SOMFree(pItem->somid);
        pItem2 = pItem->pNext;
        free(pItem);
        pItem = pItem2;
    } while (pItem);

    free(pwpsci->pObjClass);
    free(pwpsci);
}

/*
 * fnwpSelectWPSClass:
 *      dlg proc for "Select class" dialog for selecting a
 *      class in single-object status bar mode
 */

MRESULT EXPENTRY fnwpSelectWPSClass(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2)
{
    MRESULT mrc;
    PSELECTCLASSDATA pscd = (PSELECTCLASSDATA)WinQueryWindowULong(hwndDlg, QWL_USER);

    switch(msg) {

        /*
         * WM_INITDLG:
         *      set up the container data and colors
         */

        case WM_INITDLG: {
            CNRINFO CnrInfo;
            WinSetWindowULong(hwndDlg, QWL_USER, (ULONG)mp2);
            pscd = (PSELECTCLASSDATA)mp2;

            WinSetWindowText(hwndDlg, pscd->pszDlgTitle);
            WinSetDlgItemText(hwndDlg, ID_XSDI_SC_INTROTEXT, pscd->pszIntroText);

            // setup container
            pscd->hwndCnr = WinWindowFromID(hwndDlg, ID_XSDI_SC_CNR);
            WinSendMsg(pscd->hwndCnr, CM_QUERYCNRINFO, &CnrInfo, (MPARAM)sizeof(CnrInfo));
            CnrInfo.pSortRecord = (PVOID)fnCompareName;
            CnrInfo.flWindowAttr =
                    CV_TREE | CA_TREELINE | CV_TEXT
                    | CA_OWNERDRAW;
            CnrInfo.cxTreeIndent = 30;
            WinSendMsg(pscd->hwndCnr, CM_SETCNRINFO,
                    &CnrInfo,
                    (MPARAM)(CMA_PSORTRECORD | CMA_FLWINDOWATTR | CMA_CXTREEINDENT));

            // hide help button if panel was not specified
            if (pscd->ulHelpPanel == 0)
                WinShowDlgItem(hwndDlg, DID_HELP, FALSE);

            // get container colors for owner draw later
            pscd->lBackground =
                    WinQueryPresColor(hwndDlg, PP_BACKGROUNDCOLOR, "Window", "255 255 255");
            pscd->lText =
                    WinQueryPresColor(hwndDlg, PP_FOREGROUNDCOLOR, "WindowText", "0 0 0");
            pscd->lStaticText =
                    WinQueryPresColor(hwndDlg, -1, "ShadowText", "0 0 255");
            pscd->lSelBackground =
                    WinQueryPresColor(hwndDlg, PP_HILITEBACKGROUNDCOLOR, "HiliteBackground", "255 0 0");
            pscd->lSelText =
                    WinQueryPresColor(hwndDlg, PP_HILITEFOREGROUNDCOLOR, "HiliteForeground", "255 255 255");

            // fill container later
            WinPostMsg(hwndDlg, WM_FILLCNR, MPNULL, MPNULL);

            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
        break; }

        /*
         * WM_FILLCNR:
         *      this fills the cnr with all the WPS classes
         */

        case WM_FILLCNR: {
            PRECORDCORE preccSelected;
            HPOINTER hptrOld = WinQueryPointer(HWND_DESKTOP);
            WinSetPointer(HWND_DESKTOP,
                    WinQuerySysPointer(HWND_DESKTOP, SPTR_WAIT, FALSE));
            pscd->pwpsc = cmnWpsClasses2Cnr(pscd->hwndCnr, pscd->pszRootClass,
                    NULL, NULL,
                    pscd);  // also contains callback
            // scroll cnr to the item that is selected
            preccSelected = WinSendMsg(pscd->hwndCnr, CM_QUERYRECORDEMPHASIS,
                    (MPARAM)CMA_FIRST,
                    (MPARAM)CRA_SELECTED);
            WinCnrScrollToRecord(pscd->hwndCnr, preccSelected,
                    CMA_TEXT, TRUE);
            WinSetPointer(HWND_DESKTOP, hptrOld);
        break; }

        /*
         * WM_DRAWITEM:
         *      container control owner draw: the
         *      cnr only allows the same color for
         *      all text items, so we need to draw
         *      the text ourselves
         */

        case WM_DRAWITEM: {
            // get generic DRAWITEM structure
            POWNERITEM poi = (POWNERITEM)mp2;

            // check if we're to draw the text
            // (and not the icon)
            if (poi->idItem == CMA_TEXT) {

                // get container-specific draw-item struct
                PCNRDRAWITEMINFO pcdii = (PCNRDRAWITEMINFO)poi->hItem;
                ULONG flCmd = DT_LEFT | DT_VCENTER | DT_ERASERECT;
                RECTL rcl2;

                // set draw colors
                LONG  lBackground = pscd->lBackground;
                LONG  lForeground = pscd->lText;

                // switch to RGB
                GpiCreateLogColorTable(poi->hps, 0,
                    LCOLF_RGB,
                    0, 0, NULL);

                if (((pcdii->pRecord->flRecordAttr) & CRA_DISABLED) == 0) {
                    // not disabled == valid WPS class
                    if ((pcdii->pRecord->flRecordAttr) & CRA_SELECTED) {
                        lBackground = pscd->lSelBackground;
                        lForeground = pscd->lSelText;
                    }
                } else {
                    if ((pcdii->pRecord->flRecordAttr) & CRA_SELECTED) {
                        lBackground = pscd->lStaticText;
                        lForeground = pscd->lBackground;
                    } else
                        lForeground = pscd->lStaticText;
                }

                memcpy(&rcl2, &(poi->rclItem), sizeof(rcl2));
                WinDrawText(poi->hps,
                    strlen(pcdii->pRecord->pszText),
                    pcdii->pRecord->pszText,
                    &rcl2,
                    lForeground,  // foreground
                    lBackground,
                    flCmd);
                mrc = (MPARAM)TRUE; // tell cnr that we've drawn the item
            } else
                mrc = (MPARAM)FALSE; // tell cnr to draw the item
        break; }

        /*
         * WM_CONTROL:
         *      capture cnr notifications
         */

        case WM_CONTROL: {
            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
            if (SHORT1FROMMP(mp1) == ID_XSDI_SC_CNR) {
                switch (SHORT2FROMMP(mp1)) { // notify code

                    /*
                     * CN_ENTER:
                     *      double-click or "Enter" on
                     *      record core
                     */

                    case CN_ENTER: {
                        if (pscd->fSelectionValid) {
                            WinPostMsg(hwndDlg, WM_COMMAND,
                                    (MPARAM)DID_OK,
                                    MPNULL);
                        }
                    break; }

                    /*
                     * CN_EMPHASIS:
                     *      selection changed: call
                     *      callback func specified in
                     *      SELECTCLASSDATA
                     */

                    case CN_EMPHASIS: {
                        // get cnr notification struct
                        PNOTIFYRECORDEMPHASIS pnre = (PNOTIFYRECORDEMPHASIS)mp2;
                        if (pnre)
                            if (    (pnre->fEmphasisMask & CRA_SELECTED)
                                 && (pnre->pRecord)
                               )
                            {
                                pscd->pRecordSelected = (PCLASSRECORDCORE)(pnre->pRecord);
                                // per definition, this callback func must
                                // return TRUE if the OK button is to be
                                // enabled
                                if (pscd->pfnwpClassSelected) {
                                    pscd->fSelectionValid =
                                        (BOOL)(*(pscd->pfnwpClassSelected))(
                                            pscd->hwndCnr,
                                            pscd->ulUserClassSelected,
                                            // mp1: WPSLISTITEM struct
                                            (MPARAM)( ((PCLASSRECORDCORE)(pnre->pRecord))->pwps ),
                                            // mp2: hwnd of info static text control below cnr
                                            (MPARAM)( WinWindowFromID(hwndDlg, ID_XSDI_SC_TEXT2) )
                                        );
                                    // enable OK button according to
                                    // callback return value
                                    WinEnableDlgItem(hwndDlg, DID_OK, pscd->fSelectionValid);
                                }
                            }
                    break; }
                }
            }
        break; }

        /*
         * WM_COMMAND:
         *      if OK button was pressed (or CN_ENTER posted, see above),
         *      update the SELECTCLASSDATA struct
         */

        case WM_COMMAND: {
            switch (SHORT1FROMMP(mp1)) {
                case DID_OK: {
                    if (pscd->pRecordSelected)
                        strcpy(pscd->szClassSelected,
                                pscd->pRecordSelected->pwps->pszClassName);
                break; }
            }
            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
        break; }

        case WM_HELP: {
            if (pscd->ulHelpPanel) {
                _wpDisplayHelp(_wpclsQueryActiveDesktop(_WPDesktop),
                        pscd->ulHelpPanel,
                        pscd->pszHelpLibrary);
            }
        break; }

        case WM_DESTROY: {
            // cleanup
            cmnCleanupWpsClasses(pscd->pwpsc);
            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
        break; }

        default:
            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
    }
    return (mrc);
}

/*
 * cmnSelectWpsClass:
 *      this is a one-shot func for displaying a
 *      "Select WPS class" dialog; you must fill
 *      the SELECTCLASSDATA struct with the
 *      neccessary data. The class selected by
 *      the user will also be copied to this struct.
 *      This func returns either DID_OK or
 *      DID_CANCEL, in which case you shouldn't
 *      evaluate the return values in SELECTCLASSDATA.
 *      hmod and idDlg specify the dlg template for
 *      this func.
 */

ULONG cmnSelectWpsClass(HWND hwndOwner,
            HMODULE hmod, ULONG idDlg,
            PSELECTCLASSDATA pscd)
{
    return (WinDlgBox(HWND_DESKTOP, hwndOwner,
                            fnwpSelectWPSClass,
                            hmod, idDlg,
                            pscd));
}

/********************************************************************
 *                                                                  *
 *   "Register new class" dlg                                       *
 *                                                                  *
 ********************************************************************/

MRESULT EXPENTRY fnwpOpenFilter(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   return (WinDefFileDlgProc(hwnd, msg, mp1, mp2));
}

MRESULT EXPENTRY fnwpRegisterClass(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2)
{
    MRESULT mrc;
    PREGISTERCLASSDATA prcd = (PREGISTERCLASSDATA)WinQueryWindowULong(hwndDlg, QWL_USER);

    switch(msg) {

        /*
         * WM_INITDLG:
         *      set up the container data and colors
         */

        case WM_INITDLG: {
            // CNRINFO CnrInfo;
            WinSetWindowULong(hwndDlg, QWL_USER, (ULONG)mp2);
            prcd = (PREGISTERCLASSDATA)mp2;
            if (prcd->ulHelpPanel == 0)
                WinShowDlgItem(hwndDlg, DID_HELP, FALSE);
            WinSendDlgItemMsg(hwndDlg, ID_XSDI_SC_CLASSNAME, EM_SETTEXTLIMIT,
                (MPARAM)(255-1), MPNULL);
            WinSendDlgItemMsg(hwndDlg, ID_XSDI_SC_CLASSMODULE, EM_SETTEXTLIMIT,
                (MPARAM)(255-1), MPNULL);
            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
        break; }

        case WM_CONTROL: {
                if ( (     (SHORT1FROMMP(mp1) == ID_XSDI_SC_CLASSNAME)
                        || (SHORT1FROMMP(mp1) == ID_XSDI_SC_CLASSMODULE)
                     )
                     && (SHORT2FROMMP(mp1) == EN_CHANGE)
                   )
                {
                    BOOL fEnable = FALSE;
                    WinQueryDlgItemText(hwndDlg, ID_XSDI_SC_CLASSNAME,
                            sizeof(prcd->szClassName), prcd->szClassName);
                    if (strlen(prcd->szClassName)) {
                        WinQueryDlgItemText(hwndDlg, ID_XSDI_SC_CLASSMODULE,
                                sizeof(prcd->szModName), prcd->szModName);
                        if (strlen(prcd->szClassName))
                            fEnable = TRUE;
                    }
                    WinEnableDlgItem(hwndDlg, DID_OK, fEnable);
                }

        break; }

        case WM_COMMAND: {
            switch (SHORT1FROMMP(mp1)) {
                case DID_OK: {
                    if (prcd) {
                        WinQueryDlgItemText(hwndDlg, ID_XSDI_SC_CLASSNAME,
                                sizeof(prcd->szClassName), prcd->szClassName);
                        WinQueryDlgItemText(hwndDlg, ID_XSDI_SC_CLASSMODULE,
                                sizeof(prcd->szModName), prcd->szModName);
                    }
                    mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
                break; }

                case ID_XSDI_SC_BROWSE: {
                    FILEDLG fd;
                    memset(&fd, 0, sizeof(FILEDLG));
                    fd.cbSize = sizeof(FILEDLG);
                    fd.fl = FDS_OPEN_DIALOG | FDS_CENTER;
                    fd.pfnDlgProc = fnwpOpenFilter;
                    // fd.pszTitle = "Blah";
                    // fd.pszOKButton = "~OK";
                    strcpy(fd.szFullFile, "*.DLL");
                    // fd.pszIDrive = "F:"
                    // fd.ulFQFCount = 1;
                    if (WinFileDlg(HWND_DESKTOP, hwndDlg, &fd) && (fd.lReturn == DID_OK)) {
                        // strcpy(prcd->szClassName, fd.szFullFile);
                        WinSetDlgItemText(hwndDlg, ID_XSDI_SC_CLASSMODULE,
                                    fd.szFullFile);
                    }
                break; }

                default:
                    mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
            }
        break; }

        case WM_HELP: {
            if (prcd->ulHelpPanel) {
                _wpDisplayHelp(_wpclsQueryActiveDesktop(_WPDesktop),
                        prcd->ulHelpPanel,
                        prcd->pszHelpLibrary);
            }
        break; }

        default:
            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
    }
    return (mrc);
}

/********************************************************************
 *                                                                  *
 *   "Object usage" dialog                                          *
 *                                                                  *
 ********************************************************************/

POBJECTUSAGERECORDCORE cmnAddUsage2Cnr(HWND hwndCnr,
    POBJECTUSAGERECORDCORE preccParent, PSZ pszTitle, ULONG flAttrs)
{
    POBJECTUSAGERECORDCORE preccNew = (POBJECTUSAGERECORDCORE)
        WinCnrAllocRecord(hwndCnr, sizeof(OBJECTUSAGERECORDCORE));

    strcpy(preccNew->szText, pszTitle);
    WinCnrInsertRecord(hwndCnr, (PRECORDCORE)preccParent, (PRECORDCORE)preccNew,
            preccNew->szText, flAttrs);
    return (preccNew);
}

#ifdef DEBUG_MEMORY
    /*
     * fncbHeapWalk:
     *      callback func for _heap_walk function used for
     *      object usage (cmnObjectUsage2Cnr)
     */

    LONG    lObjectCount,
            lTotalObjectSize,
            lFreedObjectCount,
            lHeapStatus;

    int fncbHeapWalk(const void *pObject, size_t Size, int useflag, int status,
                          const char *filename, size_t line)
    {
        if (status != _HEAPOK) {
            lHeapStatus = status;
        }
        if (useflag == _USEDENTRY) {
            // object not freed
            lObjectCount++;
            lTotalObjectSize += Size;
        } else
            lFreedObjectCount++;

        return 0;
    }
#endif

/*
 * cmnObjectUsage2Cnr:
 *      adds all the object details into a given
 *      container window
 */

VOID cmnObjectUsage2Cnr(HWND hwndCnr, WPObject *pObject)
{
    POBJECTUSAGERECORDCORE
                     preccRoot, preccLevel2, preccLevel3;
    CHAR             szTemp1[100], szText[300];
    PUSEITEM         pUseItem;

    CHAR             szObjectHandle[20];
    HOBJECT          hObject;
    PSZ              pszObjectID;
    ULONG            ul;

    // printf("Cnr: %d\n", hwndCnr);
    if (pObject) {
        sprintf(szText, "%s (Class: %s)",
                _wpQueryTitle(pObject), _somGetClassName(pObject));
        preccRoot = cmnAddUsage2Cnr(hwndCnr, NULL, szText,
                CRA_RECORDREADONLY | CRA_EXPANDED);

        // object ID
        preccLevel2 = cmnAddUsage2Cnr(hwndCnr, preccRoot, "Object ID",
                CRA_RECORDREADONLY | CRA_EXPANDED);
        pszObjectID = _wpQueryObjectID(pObject);
        if (pszObjectID)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, pszObjectID,
                    (strcmp(pszObjectID, "<WP_DESKTOP") != 0
                        ? 0 // editable!
                        : CRA_RECORDREADONLY)); // for the Desktop
        else
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "none set", 0); // editable!

        // object handle
        preccLevel2 = cmnAddUsage2Cnr(hwndCnr, preccRoot, "Object handle",
                CRA_RECORDREADONLY | CRA_EXPANDED);
        if (_somIsA(pObject, _WPFileSystem)) {
            CHAR    szPath[CCHMAXPATH];
            ULONG   ulSize = sizeof(szPath);
            _wpQueryRealName(pObject, szPath, &ulSize,
                    TRUE);      // fully qualified
            hObject = objQueryHandleFromPath(HINI_USER, HINI_SYSTEM,
                                szPath);
        } else if (_somIsA(pObject, _WPAbstract)) {
            hObject = _wpQueryHandle(pObject);
        }
        if ((LONG)hObject > 0)
            sprintf(szObjectHandle, "0x%lX", hObject);
        else
            sprintf(szObjectHandle, "(none queried)");
        cmnAddUsage2Cnr(hwndCnr, preccLevel2, szObjectHandle, CRA_RECORDREADONLY);

        // object style
        preccLevel2 = cmnAddUsage2Cnr(hwndCnr, preccRoot, "Object style",
                            CRA_RECORDREADONLY);
        ul = _wpQueryStyle(pObject);
        if (ul & OBJSTYLE_CUSTOMICON)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2,
                        "Custom icon (destroy icon when object goes dormant)",
                        CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_NOCOPY)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "no copy",
                                CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_NODELETE)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "no delete",
                                CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_NODRAG)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "no drag",
                                CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_NODROPON)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "no drop-on",
                                CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_NOLINK)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "no link (cannot have shadows)",
                                CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_NOMOVE)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "no move",
                                CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_NOPRINT)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "no print",
                                CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_NORENAME)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "no rename",
                                CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_NOSETTINGS)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "no settings",
                                CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_NOSETTINGS)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "not visible",
                                CRA_RECORDREADONLY);
        if (ul & OBJSTYLE_TEMPLATE)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "template",
                                CRA_RECORDREADONLY);
        /* if (ul & OBJSTYLE_LOCKEDINPLACE)
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, "locked in place",
                                CRA_RECORDREADONLY); */

        // folder data
        if (_somIsA(pObject, _WPFolder)) {
            preccLevel2 = cmnAddUsage2Cnr(hwndCnr, preccRoot, "Folder flags",
                                CRA_RECORDREADONLY);
            ul = _wpQueryFldrFlags(pObject);
            if (ul & FOI_POPULATEDWITHALL)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "Fully populated",
                                    CRA_RECORDREADONLY);
            if (ul & FOI_POPULATEDWITHFOLDERS)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "Populated with folders",
                                    CRA_RECORDREADONLY);
            if (ul & FOI_FIRSTPOPULATE)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "Populated with first objects",
                                    CRA_RECORDREADONLY);
            if (ul & FOI_WORKAREA)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "Work area",
                                    CRA_RECORDREADONLY);
            if (ul & FOI_CHANGEFONT)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "FOI_CHANGEFONT",
                                    CRA_RECORDREADONLY);
            if (ul & FOI_NOREFRESHVIEWS)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "FOI_NOREFRESHVIEWS",
                                    CRA_RECORDREADONLY);
            if (ul & FOI_ASYNCREFRESHONOPEN)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "FOI_ASYNCREFRESHONOPEN",
                                    CRA_RECORDREADONLY);
            if (ul & FOI_REFRESHINPROGRESS)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "FOI_REFRESHINPROGRESS",
                                    CRA_RECORDREADONLY);
            if (ul & FOI_WAMCRINPROGRESS)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "FOI_WAMCRINPROGRESS",
                                    CRA_RECORDREADONLY);
            if (ul & FOI_CNRBKGNDOLDFORMAT)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "FOI_CNRBKGNDOLDFORMAT",
                                    CRA_RECORDREADONLY);
            if (ul & FOI_DELETEINPROGRESS)
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, "FOI_DELETEINPROGRESS",
                                    CRA_RECORDREADONLY);
        }

        // WPS data
        if (pObject == _wpclsQueryActiveDesktop(_WPDesktop)) {
            QPSPROCESS       qpsp;
            PTIB             ptib;
            PPIB             ppib;

            preccLevel2 = cmnAddUsage2Cnr(hwndCnr, preccRoot, "Workplace Shell status",
                                CRA_RECORDREADONLY);

            // we use a conditional compile flag here because
            // _heap_walk adds additional overhead to malloc()
            #ifdef DEBUG_MEMORY
                lObjectCount = 0;
                lTotalObjectSize = 0;
                lFreedObjectCount = 0;
                lHeapStatus = _HEAPOK;

                // get heap info using the callback above
                _heap_walk(fncbHeapWalk);

                strths(szTemp1, lTotalObjectSize, ',');
                sprintf(szText, "XFolder memory consumption: %s bytes\n"
                            "(%d objects used, %d objects freed)",
                            szTemp1,
                            lObjectCount,
                            lFreedObjectCount);
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, szText,
                                    CRA_RECORDREADONLY);

                sprintf(szText, "XFolder memory heap status: %s",
                            (lHeapStatus == _HEAPOK) ? "OK"
                            : (lHeapStatus == _HEAPBADBEGIN) ? "Invalid heap (_HEAPBADBEGIN)"
                            : (lHeapStatus == _HEAPBADNODE) ? "Damaged memory node"
                            : (lHeapStatus == _HEAPEMPTY) ? "Heap not initialized"
                            : "unknown error"
                            );
                cmnAddUsage2Cnr(hwndCnr, preccLevel2, szText,
                                    CRA_RECORDREADONLY);
            #endif

            sprintf(szText, "Currently awake WPS objects: %d",
                        cmnQueryAwakeObjectsCount());
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, szText,
                                CRA_RECORDREADONLY);

            DosGetInfoBlocks(&ptib, &ppib);
            sprintf(szText, "Workplace Shell process ID: 0x%lX", ppib->pib_ulpid);
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, szText,
                                CRA_RECORDREADONLY);

            qpsQueryProcessInfo(ppib->pib_ulpid, &qpsp);
            sprintf(szText, "Workplace Shell thread count: %d", qpsp.usThreads);
            cmnAddUsage2Cnr(hwndCnr, preccLevel2, szText,
                                CRA_RECORDREADONLY);

        }

        // object usage
        preccLevel2 = cmnAddUsage2Cnr(hwndCnr, preccRoot, "Object usage",
                            CRA_RECORDREADONLY);

        preccLevel3 = NULL;
        for (pUseItem = _wpFindUseItem(pObject, USAGE_OPENVIEW, NULL);
            pUseItem;
            pUseItem = _wpFindUseItem(pObject, USAGE_OPENVIEW, pUseItem))
        {
            PVIEWITEM pViewItem = (PVIEWITEM)(pUseItem+1);
            switch (pViewItem->view) {
                case OPEN_SETTINGS: strcpy(szTemp1, "Settings"); break;
                case OPEN_CONTENTS: strcpy(szTemp1, "Icon"); break;
                case OPEN_DETAILS:  strcpy(szTemp1, "Details"); break;
                case OPEN_TREE:     strcpy(szTemp1, "Tree"); break;
                case OPEN_RUNNING:  strcpy(szTemp1, "Program running"); break;
                case OPEN_PROMPTDLG:strcpy(szTemp1, "Prompt dialog"); break;
                case OPEN_PALETTE:  strcpy(szTemp1, "Palette"); break;
                default:            sprintf(szTemp1, "unknown (0x%lX)", pViewItem->view); break;
            }
            sprintf(szText, "%s", szTemp1);
            if (!preccLevel3)
                preccLevel3 = cmnAddUsage2Cnr(hwndCnr, preccLevel2, "Currently open views",
                        CRA_RECORDREADONLY | CRA_EXPANDED);
            cmnAddUsage2Cnr(hwndCnr, preccLevel3, szText, CRA_RECORDREADONLY);
        }

        preccLevel3 = NULL;
        for (pUseItem = _wpFindUseItem(pObject, USAGE_MEMORY, NULL);
            pUseItem;
            pUseItem = _wpFindUseItem(pObject, USAGE_MEMORY, pUseItem))
        {
            PMEMORYITEM pMemoryItem = (PMEMORYITEM)(pUseItem+1);
            sprintf(szText, "Size: %d", pMemoryItem->cbBuffer);
            if (!preccLevel3)
                preccLevel3 = cmnAddUsage2Cnr(hwndCnr, preccLevel2, "Allocated memory",
                        CRA_RECORDREADONLY);
            cmnAddUsage2Cnr(hwndCnr, preccLevel3, szText, CRA_RECORDREADONLY);
        }

        preccLevel3 = NULL;
        for (pUseItem = _wpFindUseItem(pObject, USAGE_LINK, NULL);
            pUseItem;
            pUseItem = _wpFindUseItem(pObject, USAGE_LINK, pUseItem))
        {
            PLINKITEM pLinkItem = (PLINKITEM)(pUseItem+1);
            CHAR      szShadowPath[100];
            ULONG     ulSize = sizeof(szShadowPath);
            if (pLinkItem->LinkObj) {
                _wpQueryRealName(_wpQueryFolder(pLinkItem->LinkObj),
                        szShadowPath, &ulSize, CRA_RECORDREADONLY | CRA_EXPANDED);
                sprintf(szText, "%s in \n%s", _wpQueryTitle(pLinkItem->LinkObj),
                        szShadowPath);
            } else
                strcpy(szText, "broken");

            if (!preccLevel3)
                preccLevel3 = cmnAddUsage2Cnr(hwndCnr, preccLevel2,
                        "Awake shadows of this object",
                        CRA_RECORDREADONLY | CRA_EXPANDED);
            cmnAddUsage2Cnr(hwndCnr, preccLevel3, szText, CRA_RECORDREADONLY);
        }

        preccLevel3 = NULL;
        for (pUseItem = _wpFindUseItem(pObject, USAGE_RECORD, NULL);
            pUseItem;
            pUseItem = _wpFindUseItem(pObject, USAGE_RECORD, pUseItem))
        {
            PRECORDITEM pRecordItem = (PRECORDITEM)(pUseItem+1);
            CHAR szFolderTitle[256];
            WinQueryWindowText(WinQueryWindow(pRecordItem->hwndCnr, QW_PARENT),
                                sizeof(szFolderTitle)-1,
                                szFolderTitle);
            sprintf(szText, "Container window handle: 0x%lX\n(\"%s\")",
                        pRecordItem->hwndCnr,
                        szFolderTitle);
            if (!preccLevel3)
                preccLevel3 = cmnAddUsage2Cnr(hwndCnr, preccLevel2,
                        "Folder windows containing this object",
                        CRA_RECORDREADONLY | CRA_EXPANDED);
            cmnAddUsage2Cnr(hwndCnr, preccLevel3, szText, CRA_RECORDREADONLY);
        }

        preccLevel3 = NULL;
        for (pUseItem = _wpFindUseItem(pObject, USAGE_OPENFILE, NULL);
            pUseItem;
            pUseItem = _wpFindUseItem(pObject, USAGE_OPENFILE, pUseItem))
        {
            PVIEWFILE pViewFile = (PVIEWFILE)(pUseItem+1);
            sprintf(szText, "Open handle: 0x%lX", pViewFile->handle);
            if (!preccLevel3)
                preccLevel3 = cmnAddUsage2Cnr(hwndCnr, preccLevel2,
                        "Applications which opened this object",
                        CRA_RECORDREADONLY | CRA_EXPANDED);
            cmnAddUsage2Cnr(hwndCnr, preccLevel3, szText, CRA_RECORDREADONLY);
        }

        preccLevel3 = NULL;
        for (ul = 0; ul < 100; ul++)
            if (    (ul != USAGE_OPENVIEW)
                 && (ul != USAGE_MEMORY)
                 && (ul != USAGE_LINK)
                 && (ul != USAGE_RECORD)
                 && (ul != USAGE_OPENFILE)
               )
            {
                for (pUseItem = _wpFindUseItem(pObject, ul, NULL);
                    pUseItem;
                    pUseItem = _wpFindUseItem(pObject, ul, pUseItem))
                {
                    sprintf(szText, "Type: 0x%lX", pUseItem->type);
                    if (!preccLevel3)
                        preccLevel3 = cmnAddUsage2Cnr(hwndCnr, preccLevel2,
                                "Undocumented usage types",
                                CRA_RECORDREADONLY | CRA_EXPANDED);
                    cmnAddUsage2Cnr(hwndCnr, preccLevel3, szText, CRA_RECORDREADONLY);
                }
            }
    }
}

/*
 * fnwpObjectUsage:
 *      dlg proc for "Object usage" dlg
 */

MRESULT EXPENTRY fnwpObjectUsage(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2)
{
    MRESULT mrc;
    POBJECTUSAGEDATA poud = (POBJECTUSAGEDATA)WinQueryWindowULong(hwndDlg, QWL_USER);

    switch(msg) {
        case WM_INITDLG: {
            CNRINFO CnrInfo;
            WinSetWindowULong(hwndDlg, QWL_USER, (ULONG)mp2);
            poud = (POBJECTUSAGEDATA)mp2;

            WinSetWindowText(hwndDlg, poud->szDlgTitle);
            WinSetDlgItemText(hwndDlg, ID_XSDI_SC_INTROTEXT, poud->szIntroText);

            // setup container
            poud->hwndCnr = WinWindowFromID(hwndDlg, ID_XSDI_SC_CNR);
            WinSendMsg(poud->hwndCnr, CM_QUERYCNRINFO, &CnrInfo, (MPARAM)sizeof(CnrInfo));
            CnrInfo.pSortRecord = (PVOID)fnCompareName;
            CnrInfo.flWindowAttr = CV_TREE | CV_TEXT | CA_TREELINE;
            CnrInfo.cxTreeIndent = 30;
            WinSendMsg(poud->hwndCnr, CM_SETCNRINFO,
                    &CnrInfo,
                    (MPARAM)(CMA_PSORTRECORD | CMA_FLWINDOWATTR | CMA_CXTREEINDENT));

            if (poud->ulHelpPanel == 0)
                WinShowDlgItem(hwndDlg, DID_HELP, FALSE);

            WinPostMsg(hwndDlg, WM_FILLCNR, MPNULL, MPNULL);
            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
        break; }

        case WM_FILLCNR: {
            HPOINTER hptrOld = WinQueryPointer(HWND_DESKTOP);
            WinSetPointer(HWND_DESKTOP, WinQuerySysPointer(HWND_DESKTOP, SPTR_WAIT, FALSE));
            cmnObjectUsage2Cnr(poud->hwndCnr, poud->pObject);
            WinSetPointer(HWND_DESKTOP, hptrOld);
        break; }

        case WM_CONTROL: {
            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
            if (SHORT1FROMMP(mp1) == ID_XSDI_SC_CNR) {
                switch (SHORT2FROMMP(mp1)) { // notify code
                    case CN_EMPHASIS: {
                        PNOTIFYRECORDEMPHASIS pnre = (PNOTIFYRECORDEMPHASIS)mp2;
                        if (pnre)
                            if (pnre->fEmphasisMask & CRA_SELECTED) {
                            }
                    break; }
                }
            }
        break; }

        case WM_DESTROY: {
            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
        break; }

        default:
            mrc = WinDefDlgProc(hwndDlg, msg, mp1, mp2);
    }
    return (mrc);
}


