/*
 * objdlgs.h:
 *      header file for objdlgs.c, which contains the logic
 *      for displaying the WPS class tree in a container.
 *      In order to use this, you need to do the following:
 *      1) #include objdlgs.h, but make sure you
 *         have #include'd os2.h before
 *      2) set up the SELECTCLASSDATA structure below
 *         and pass this to WinDlgBox. Example:
 *
 *             SELECTCLASSDATA scd;
 *             strcpy(scd.szDlgTitle, "Select Class");
 *             strcpy(scd.szIntroText, "Select the class for which you want to change "
 *                                   "the status bar info.");
 *             strcpy(scd.szRootClass, "WPObject");
 *             strcpy(scd.szClassSelected, "WPObject");
 *             scd.ulHelpPanel = 0;
 *
 *             WinDlgBox(HWND_DESKTOP, hwndDlg,
 *                     fnwpSelectStatusBarClass,
 *                     NLS_MODULE, ID_XSD_SELECTCLASS,
 *                     &scd);
 *
 *      This also contains the logic for the "Internals"
 *      page in each object's settings notebook.
 *
 *      And don't forget to link the dlg resource for this
 *      against your executable.
 *
 *      Copyright (C) 1997-98 Ulrich Mller.
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, in version 2 as it comes in the COPYING
 *      file of the XFolder main distribution.
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 */

/********************************************************************
 *                                                                  *
 *   "Select class" dialog                                          *
 *                                                                  *
 ********************************************************************/

#define WM_FILLCNR   (WM_USER+100)

/* WPS class data linked list struct; one of structs is created
   for each installed WPS class on the system */
typedef struct _WPSLISTITEM {
    struct _WPSLISTITEM *pNext, *pPrevious;
    PSZ                 pszClassName;
                        // class name as returned by WinEnumObjectClasses
    PSZ                 pszModName;
                        // DLL name as returned by WinEnumObjectClasses
    SOMClass            *pClassObject;
                        // pointer to class object (e.g. _WPObject)
    SOMClass            *pMetaclassObject;
                        // pointer to metaclass object (e.g. _M_WPObject)
    somId               somid;
                        // somId for metaclass object
    CHAR                szReplacedByClass[256];
                        // contains replacement class of this class, if applicable
    CHAR                szReplacesClass[256];
                        // contains class which this class replaces, if applicable
    CHAR                szParentClass[256];
                        // contains parent class
    PRECORDCORE         pRecord;
                        // corresponding record core
    BOOL                fProcessed;
                        // anti-recursion flag
} WPSLISTITEM, *PWPSLISTITEM;

/* extended RECORDCORE structure */
typedef struct _CLASSRECORDCORE {
    RECORDCORE     recc;
    PWPSLISTITEM    pwps;
} CLASSRECORDCORE, *PCLASSRECORDCORE;

/* WPSCLASSESINFO structure */
typedef struct _WPSCLASSESINFO {
    PWPSLISTITEM    pwpsRoot;
    POBJCLASS       pObjClass;
} WPSCLASSESINFO, *PWPSCLASSESINFO;

/* SELECTCLASSDATA:
   struct for passing dato the "Select class" dlg */
typedef struct _SELECTCLASSDATA {
    PSZ     pszDlgTitle;            // dialog title
    PSZ     pszIntroText;           // intro text above container
    PSZ     pszRootClass;           // class to begin with
    PSZ     pszOrphans;             // title of "Orphans" tree; if NULL,
                                    // orphaned classes will not be shown
    CHAR    szClassSelected[256];   // in: select this class
                                    // out: class selected by user
    PFNWP   pfnwpReturnAttrForClass;
            // this callback func gets called just before a
            // record core gets inserted into the cnr.
            // Params passed to this callback func are:
            //      HWND hwnd    container wnd handle
            //      ULONG PSELECTCLASSDATA pscd
            //                   pointer to this dlg structure
            //      PWPSLISTITEM mp1  structure with class info
            //      PCLASSRECORDCORE mp2  _parent_ record core in which the
            //                        new recc will be inserted
            // This func must return the CRA_* flags for the new
            // record core which will then be inserted.

    PFNWP   pfnwpClassSelected;
            // this callback func gets called when a a class is selected in the dlg.
            // Params passed to this callback func are:
            //      HWND hwnd    container wnd handle
            //      ULONG msg    user-defined ulUserClassSelected below
            //      PWPSLISTITEM mp1  structure with class info
            //      MPARAM mp2   hwnd of info static text control below cnr
            // If this func returns TRUE, the "OK" button will be enabled.
    ULONG   ulUserClassSelected;

    PSZ     pszHelpLibrary;         // help library
    ULONG   ulHelpPanel;            // help panel; if 0, the "Help"
                                    // button is disabled

    // all the following are for internal use,
    // you don't have to set these
    HWND    hwndCnr;
    PWPSCLASSESINFO pwpsc;
    PCLASSRECORDCORE pRecordSelected;
    BOOL    fSelectionValid;
    LONG    lBackground;
    LONG    lText;
    LONG    lStaticText;
    LONG    lSelBackground;
    LONG    lSelText;
    LONG    lRecurse;
    PRECORDCORE preccExpanded;
} SELECTCLASSDATA, *PSELECTCLASSDATA;

/*
 * cmnWPSClasses2Cnr:
 *      this func inserts a WPS class hierarchy starting with
 *      the class "pszCurrentClass" as the root class into
 *      a given container window. This cnr should be in tree
 *      view to have a meaningful display. This function can
 *      be used with any exising container window.
 *      Parameters:
 *          HWND hwndCnr        guess what this is
 *          PSZ pszCurrentClass the "root" class of the display,
 *                              e.g. "WPObject"; make sure this
 *                              class exists, or nothing occurs
 *          PWPSLISTITEM pliClasses2 must be NULL
 *          PCLASSRECORDCORE preccParent
 *                              the parent record core to insert
 *                              the tree in; if NULL, a new tree
 *                              is created (root level).
 *          PSELECTCLASSDATA pscd
 *                              dlg structure, may be NULL; if
 *                              pfnwpReturnAttrForClass in this
 *                              this structure is != NULL, it will
 *                              be called as a callback for each
 *                              item inserted (see typedef above)
 *      Returns:
 *          PWPSCLASSESINFO         structure which WPS data.
 *
 *      NOTE: This function eats up a lot of memory, depending
 *      on how many WPS classes are installed. This memory is NOT
 *      freed when this function returns, because the container
 *      still needs it. AFTER destroying the container window,
 *      you must issue cmnCleanupWpsClasses (below) with the
 *      return value of whis function. This will free the linked
 *      WPS class item list and free allocated SOM resources.
 */

PWPSCLASSESINFO cmnWpsClasses2Cnr(HWND hwndCnr, PSZ pszCurrentClass,
        PWPSLISTITEM pliClasses2, PCLASSRECORDCORE preccParent,
        PSELECTCLASSDATA pscd);

/*
 * cmnCleanupWpsClasses:
 *      this cleans up the resources allocated by
 *      cmnWpsClasses2Cnr (memory and SOM stuff);
 *      pliClasses2 must be the pointer returned
 *      by cmnWpsClasses2Cnr
 */

VOID cmnCleanupWpsClasses(PWPSCLASSESINFO pwpsd);

/*
 * cmnSelectWpsClass:
 *      this is a one-shot func for displaying a
 *      "Select WPS class" dialog; you must fill
 *      the SELECTCLASSDATA struct with the
 *      neccessary data. The class selected by
 *      the user will also be copied to this struct.
 *      This func returns either DID_OK or
 *      DID_CANCEL, in which case you shouldn't
 *      evaluate the return values in SELECTCLASSDATA.
 *      hmod and idDlg specify the dlg template for
 *      this func.
 */

ULONG cmnSelectWpsClass(HWND hwndOwner,
            HMODULE hmod, ULONG idDlg,
            PSELECTCLASSDATA pscd);

/********************************************************************
 *                                                                  *
 *   "Register new class" dlg                                       *
 *                                                                  *
 ********************************************************************/

typedef struct _REGISTERCLASSDATA {
    CHAR    szModName[CCHMAXPATH];
    CHAR    szClassName[256];
    PSZ     pszHelpLibrary;         // help library
    ULONG   ulHelpPanel;            // help panel; if 0, the "Help"
                                    // button is disabled
} REGISTERCLASSDATA, *PREGISTERCLASSDATA;

MRESULT EXPENTRY fnwpRegisterClass(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);

/********************************************************************
 *                                                                  *
 *   "Object usage" dialog                                          *
 *                                                                  *
 ********************************************************************/

/* extended RECORDCORE structure */
typedef struct _OBJECTUSAGERECORDCORE {
    RECORDCORE     recc;
    CHAR           szText[500];
} OBJECTUSAGERECORDCORE, *POBJECTUSAGERECORDCORE;

typedef struct _OBJECTUSAGEDATA {
    WPObject *pObject;              // object to query
    CHAR    szDlgTitle[256];        // dialog title
    CHAR    szIntroText[256];       // intro text above container
    CHAR    szHelpLibrary[CCHMAXPATH]; // help library
    ULONG   ulHelpPanel;            // help panel; if 0, the "Help"
                                    // button is disabled
    HWND    hwndCnr;                // internal use, don't bother
} OBJECTUSAGEDATA, *POBJECTUSAGEDATA;

/*
 * cmnObjectUsage2Cnr:
 *      adds all the object details into a given
 *      container window
 */

VOID cmnObjectUsage2Cnr(HWND hwndCnr, WPObject *pObject);

MRESULT EXPENTRY fnwpObjectUsage(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2);


